@@ -22,6 +22,7 @@ sig
   val mem_f : 'a eq -> 'a -> 'a list -> bool
   val for_all_i : (int -> 'a -> bool) -> int -> 'a list -> bool
   val for_all2eq : ('a -> 'b -> bool) -> 'a list -> 'b list -> bool
+  val exists_i : (int -> 'a -> bool) -> int -> 'a list -> bool
   val prefix_of : 'a eq -> 'a list -> 'a list -> bool
   val same_length : 'a list -> 'b list -> bool
   val interval : int -> int -> int list
@@ -113,6 +114,7 @@ sig
   val distinct : 'a list -> bool
   val distinct_f : 'a cmp -> 'a list -> bool
   val duplicates : 'a eq -> 'a list -> 'a list
+  val uniquize_key : ('a -> 'b) -> 'a list -> 'a list
   val uniquize : 'a list -> 'a list
   val sort_uniquize : 'a cmp -> 'a list -> 'a list
   val min : 'a cmp -> 'a list -> 'a
@@ -195,6 +197,13 @@ let for_all_i p =
 let for_all2eq f l1 l2 =
   try List.for_all2 f l1 l2 with Invalid_argument _ -> false
 
+let exists_i p =
+  let rec exists_p i = function
+    | [] -> false
+    | a::l -> p i a || exists_p (i+1) l
+  in
+  exists_p
+
 let prefix_of cmp prefl l =
   let rec prefrec = function
     | (h1::t1, h2::t2) -> cmp h1 h2 && prefrec (t1,t2)
@@ -947,18 +956,22 @@ let distinct_f cmp l =
 
 (* FIXME: again, generic hash function *)
 
-let uniquize l =
+let uniquize_key f l =
   let visited = Hashtbl.create 23 in
   let rec aux acc changed = function
-    | h :: t -> if Hashtbl.mem visited h then aux acc true t else
+    | h :: t ->
+        let x = f h in
+        if Hashtbl.mem visited x then aux acc true t else
           begin
-            Hashtbl.add visited h h;
+            Hashtbl.add visited x x;
             aux (h :: acc) changed t
           end
     | [] -> if changed then List.rev acc else l
   in
   aux [] false l
 
+let uniquize l = uniquize_key (fun x -> x) l
+
 (** [sort_uniquize] might be an alternative to the hashtbl-based
     [uniquize], when the order of the elements is irrelevant *)
 
