@@ -7,10 +7,12 @@ ci_dir="$(dirname "$0")"
 
 git_download metacoq
 
+if [ "$DOWNLOAD_ONLY" ]; then exit 0; fi
+
 export COQEXTRAFLAGS='-native-compiler no'
 ( cd "${CI_BUILD_DIR}/metacoq"
   ./configure.sh local
   make .merlin
-  make ci-local
+  make ci-local-noclean
   make install
 )
