@@ -5,11 +5,14 @@ set -e
 ci_dir="$(dirname "$0")"
 . "${ci_dir}/ci-common.sh"
 
-FORCE_GIT=1
+WITH_SUBMODULES=1
 git_download perennial
 
-export COQEXTRAFLAGS='-native-compiler no'
+if [ "$DOWNLOAD_ONLY" ]; then exit 0; fi
+
+ulimit -s
+ulimit -s 65536
+ulimit -s
 ( cd "${CI_BUILD_DIR}/perennial"
-  git submodule update --init --recursive
   make TIMED=false lite
 )
