@@ -21,14 +21,11 @@ open Namegen
 module RelDecl = Context.Rel.Declaration
 module NamedDecl = Context.Named.Declaration
 
-let safe_evar_value sigma ev =
-  let ev = EConstr.of_existential ev in
-  try Some (EConstr.Unsafe.to_constr @@ Evd.existential_value sigma ev)
-  with NotInstantiatedEvar | Not_found -> None
+let create_clos_infos env sigma flags =
+  let open CClosure in
+  let evars ev = Evd.existential_opt_value0 sigma ev in
+  create_clos_infos ~univs:(Evd.universes sigma) ~evars flags env
 
-let new_global evd x =
-  let (evd, c) = Evd.fresh_global (Global.env()) evd x in
-  (evd, c)
 
 (****************************************************)
 (* Expanding/testing/exposing existential variables *)
@@ -36,11 +33,11 @@ let new_global evd x =
 
 let finalize ?abort_on_undefined_evars sigma f =
   let sigma = minimize_universes sigma in
-  let uvars = ref Univ.LSet.empty in
+  let uvars = ref Univ.Level.Set.empty in
   let v = f (fun c ->
       let varsc = EConstr.universes_of_constr sigma c in
       let c = EConstr.to_constr ?abort_on_undefined_evars sigma c in
-      uvars := Univ.LSet.union !uvars varsc;
+      uvars := Univ.Level.Set.union !uvars varsc;
       c)
   in
   let sigma = restrict_universe_context sigma !uvars in
@@ -64,7 +61,7 @@ let flush_and_check_evars sigma c =
 (** Term exploration up to instantiation. *)
 let kind_of_term_upto = EConstr.kind_upto
 
-let nf_evar0 sigma t = EConstr.to_constr ~abort_on_undefined_evars:false sigma (EConstr.of_constr t)
+let nf_evars_universes sigma t = EConstr.to_constr ~abort_on_undefined_evars:false sigma (EConstr.of_constr t)
 let whd_evar = EConstr.whd_evar
 let nf_evar sigma c = EConstr.of_constr (EConstr.to_constr ~abort_on_undefined_evars:false sigma c)
 
@@ -76,12 +73,8 @@ let jv_nf_evar sigma = Array.map (j_nf_evar sigma)
 let tj_nf_evar sigma {utj_val=v;utj_type=t} =
   {utj_val=nf_evar sigma v;utj_type=t}
 
-let nf_evars_universes evm =
-  UnivSubst.nf_evars_and_universes_opt_subst (safe_evar_value evm)
-    (Evd.universe_subst evm)
-
 let nf_named_context_evar sigma ctx =
-  Context.Named.map (nf_evar0 sigma) ctx
+  Context.Named.map (nf_evars_universes sigma) ctx
 
 let nf_rel_context_evar sigma ctx =
   Context.Rel.map (nf_evar sigma) ctx
@@ -363,7 +356,7 @@ let push_rel_decl_to_named_context
   | Some id0 when hypnaming = KeepUserNameAndRenameExistingEvenSectionNames ||
                   (hypnaming = KeepUserNameAndRenameExistingButSectionNames ||
                    hypnaming = ProgramNaming) &&
-                  not (is_section_variable id0) ->
+                  not (is_section_variable (Global.env ()) id0) ->
       (* spiwack: if [id<>id0], rather than introducing a new
           binding named [id], we will keep [id0] (the name given
           by the user) and rename [id0] into [id] in the named
@@ -419,7 +412,7 @@ let new_pure_evar ?(src=default_source) ?(filter = Filter.identity) ?identity
   in
   let identity = match identity with
   | None -> Identity.none ()
-  | Some inst -> Identity.make inst
+  | Some inst -> inst
   in
   let evi = {
     evar_hyps = sign;
@@ -451,7 +444,7 @@ let new_evar ?src ?filter ?abstract_arguments ?candidates ?naming ?typeclass_can
     match filter with
     | None -> instance
     | Some filter -> Filter.filter_list filter instance in
-  let identity = if Int.equal (Environ.nb_rel env) 0 then Some instance else None in
+  let identity = if Int.equal (Environ.nb_rel env) 0 then Some (Identity.make instance) else None in
   let (evd, evk) = new_pure_evar sign evd typ' ?src ?filter ?identity ?abstract_arguments ?candidates ?naming
     ?typeclass_candidate ?principal in
   (evd, EConstr.mkEvar (evk, instance))
@@ -577,7 +570,7 @@ let rec check_and_clear_in_constr env evdref err ids global c =
             let _nconcl =
               try
                 let nids = Id.Map.domain rids in
-                let global = Id.Set.exists is_section_variable nids in
+                let global = Id.Set.exists (fun id -> is_section_variable (Global.env ()) id) nids in
                 let concl = EConstr.Unsafe.to_constr (evar_concl evi) in
                 check_and_clear_in_constr env evdref (EvarTypingBreak ev) nids global concl
               with ClearDependencyError (rid,err,where) ->
@@ -602,7 +595,7 @@ let clear_hyps_in_evi_main env sigma hyps terms ids =
      the contexts of the evars occurring in evi *)
   let evdref = ref sigma in
   let terms = List.map EConstr.Unsafe.to_constr terms in
-  let global = Id.Set.exists is_section_variable ids in
+  let global = Id.Set.exists (fun id -> is_section_variable (Global.env ()) id) ids in
   let terms =
     List.map (check_and_clear_in_constr env evdref (OccurHypInSimpleClause None) ids global) terms in
   let nhyps =
@@ -826,15 +819,15 @@ let subterm_source evk ?where (loc,k) =
    irrelevant positions, unify universes when flexible. *)
 let compare_cumulative_instances cv_pb variances u u' sigma =
   let open UnivProblem in
-  let cstrs = Univ.Constraint.empty in
+  let cstrs = Univ.Constraints.empty in
   let soft = Set.empty in
   let cstrs, soft = Array.fold_left3 (fun (cstrs, soft) v u u' ->
       let open Univ.Variance in
       match v with
       | Irrelevant -> cstrs, Set.add (UWeak (u,u')) soft
       | Covariant when cv_pb == Reduction.CUMUL ->
-        Univ.Constraint.add (u,Univ.Le,u') cstrs, soft
-      | Covariant | Invariant -> Univ.Constraint.add (u,Univ.Eq,u') cstrs, soft)
+        Univ.Constraints.add (u,Univ.Le,u') cstrs, soft
+      | Covariant | Invariant -> Univ.Constraints.add (u,Univ.Eq,u') cstrs, soft)
       (cstrs,soft) variances (Univ.Instance.to_array u) (Univ.Instance.to_array u')
   in
   match Evd.add_constraints sigma cstrs with
