@@ -97,7 +97,11 @@ let is_reserved_type na t =
 (* Turning notations and scopes on and off for printing *)
 module IRuleSet = Set.Make(struct
     type t = interp_rule
-    let compare x y = compare x y
+    let compare x y = match x, y with
+      | SynDefRule kn, SynDefRule kn' -> KerName.compare kn kn'
+      | NotationRule (scope,ntn), NotationRule (scope',ntn') -> compare x y
+      | SynDefRule _, NotationRule _ -> -1
+      | NotationRule _, SynDefRule _ -> 1
   end)
 
 let inactive_notations_table =
@@ -140,7 +144,7 @@ let deactivate_notation nr =
   | NotationRule (inscope, ntn) ->
      let scopt = match inscope with NotationInScope sc -> Some sc | LastLonelyNotation -> None in
      match availability_of_notation (inscope, ntn) (scopt, []) with
-     | None -> user_err ~hdr:"Notation"
+     | None -> user_err
                         (pr_notation ntn ++ spc () ++ str "does not exist"
                          ++ (match inscope with
                              | LastLonelyNotation -> spc () ++ str "in the empty scope."
@@ -240,7 +244,7 @@ let is_record indsp =
 let encode_record r =
   let indsp = Nametab.global_inductive r in
   if not (is_record indsp) then
-    user_err ?loc:r.CAst.loc ~hdr:"encode_record"
+    user_err ?loc:r.CAst.loc
       (str "This type is not a structure type.");
   indsp
 
@@ -303,8 +307,25 @@ let extern_evar n l = CEvar (n,l)
     For instance, in the debugger the tables of global references
     may be inaccurate *)
 
+let rec dirpath_of_modpath = function
+  | MPfile dp -> dp
+  | MPbound mbid -> let (_,id,_) = MBId.repr mbid in DirPath.make [id]
+  | MPdot (t, l) -> Libnames.add_dirpath_suffix (dirpath_of_modpath t) (Label.to_id l)
+
+let path_of_global = function
+  | GlobRef.VarRef id -> Libnames.make_path DirPath.empty id
+  (* We rely on the tacite invariant that the label of a constant is used to build its internal name *)
+  | GlobRef.ConstRef cst -> Libnames.make_path (dirpath_of_modpath (Constant.modpath cst)) (Label.to_id (Constant.label cst))
+  (* We rely on the tacite invariant that an inductive block inherits the name of its first type *)
+  | GlobRef.IndRef (ind,1) -> Libnames.make_path (dirpath_of_modpath (MutInd.modpath ind)) (Label.to_id (MutInd.label ind))
+  (* These are hacks *)
+  | GlobRef.IndRef (ind,n) -> Libnames.make_path (dirpath_of_modpath (MutInd.modpath ind)) (Id.of_string_soft ("<inductive:" ^ Label.to_string (MutInd.label ind) ^ ":" ^ string_of_int n ^ ">"))
+  | GlobRef.ConstructRef ((ind,1),p) -> Libnames.make_path (dirpath_of_modpath (MutInd.modpath ind)) (Id.of_string_soft ("<constructor:" ^ Label.to_string (MutInd.label ind) ^ ":" ^ string_of_int (p+1) ^ ">"))
+  | GlobRef.ConstructRef ((ind,n),p) -> Libnames.make_path (dirpath_of_modpath (MutInd.modpath ind)) (Id.of_string_soft ("<constructor:" ^ Label.to_string (MutInd.label ind) ^ ":" ^ string_of_int n ^ ":" ^ string_of_int (p+1) ^ ">"))
+
 let default_extern_reference ?loc vars r =
-  Nametab.shortest_qualid_of_global ?loc vars r
+  try Nametab.shortest_qualid_of_global ?loc vars r
+  with Not_found when GlobRef.is_bound r -> qualid_of_path (path_of_global r)
 
 let my_extern_reference = ref default_extern_reference
 
@@ -656,7 +677,7 @@ let adjust_implicit_arguments inctx n args impl =
            is_significant_implicit (Lazy.force a))
         in
         if visible then
-          (Lazy.force a,Some (make @@ ExplByName (name_of_implicit imp))) :: tail
+          (Lazy.force a,Some (make @@ explicitation imp)) :: tail
         else
           tail
     | a::args, _::impl -> (Lazy.force a,None) :: exprec (args,impl)
@@ -667,23 +688,17 @@ let adjust_implicit_arguments inctx n args impl =
     | [], _ -> []
   in exprec (args,impl)
 
-let extern_projection (cf,f) args impl =
-  let ip = is_projection (List.length args) cf in
-  match ip with
-    | Some i ->
-      (* Careful: It is possible to have declared implicits ending
-         before the principal argument *)
-      let is_impl =
-        try is_status_implicit (List.nth impl (i-1))
-        with Failure _ -> false
-      in
-      if is_impl
-      then None
-      else
-        let (args1,args2) = List.chop i args in
-        let (impl1,impl2) = try List.chop i impl with Failure _ -> impl, [] in
-        Some (i,(args1,impl1),(args2,impl2))
-    | None -> None
+let extern_projection inctx f nexpectedparams args impl =
+  let (args1,args2) = List.chop (nexpectedparams + 1) args in
+  let nextraargs = List.length args2 in
+  let (impl1,impl2) = impargs_for_proj ~nexpectedparams ~nextraargs impl in
+  let n = nexpectedparams + 1 + nextraargs in
+  let args1 = adjust_implicit_arguments inctx n args1 impl1 in
+  let args2 = adjust_implicit_arguments inctx n args2 impl2 in
+  let (c1,expl), args1 = List.sep_last args1 in
+  assert (expl = None);
+  let c = CProj (false,f,args1,c1) in
+  if args2 = [] then c else CApp (CAst.make c, args2)
 
 let is_start_implicit = function
   | imp :: _ -> is_status_implicit imp && maximal_insertion_of imp
@@ -732,14 +747,13 @@ let extern_record ref args =
 let extern_global impl f us =
   if not !Constrintern.parsing_explicit && is_start_implicit impl
   then
-    CAppExpl ((None, f, us), [])
+    CAppExpl ((f, us), [])
   else
     CRef (f,us)
 
 (* Implicit args indexes are in ascending order *)
 (* inctx is useful only if there is a last argument to be deduced from ctxt *)
 let extern_applied_ref inctx impl (cf,f) us args =
-  let isproj = is_projection (List.length args) cf in
   try
     if not !Constrintern.parsing_explicit &&
        ((!Flags.raw_print ||
@@ -749,34 +763,37 @@ let extern_applied_ref inctx impl (cf,f) us args =
     let impl = if !Constrintern.parsing_explicit then [] else impl in
     let n = List.length args in
     let ref = CRef (f,us) in
-    let f = CAst.make ref in
-    match extern_projection (cf,f) args impl with
-    (* Try a [t.(f args1) args2] projection-style notation *)
-    | Some (i,(args1,impl1),(args2,impl2)) ->
-      let args1 = adjust_implicit_arguments inctx n args1 impl1 in
-      let args2 = adjust_implicit_arguments inctx n args2 impl2 in
-      let ip = Some (List.length args1) in
-      CApp ((ip,f),args1@args2)
-      (* A normal application node with each individual implicit
-         arguments either dropped or made explicit *)
+    let r = CAst.make ref in
+    let ip = is_projection n cf in
+    match ip with
+    | Some i ->
+      (* [t.(f args1) args2] projection-style notation *)
+      extern_projection inctx (f,us) (i-1) args impl
     | None ->
       let args = adjust_implicit_arguments inctx n args impl in
-      if args = [] then ref else CApp ((None, f), args)
+      if args = [] then ref else CApp (r, args)
   with Expl ->
   (* A [@f args] node *)
     let args = List.map Lazy.force args in
-    let isproj = if !print_projections then isproj else None in
-    CAppExpl ((isproj,f,us), args)
+    match is_projection (List.length args) cf with
+    | Some n when !print_projections ->
+       let args = List.map (fun c -> (c,None)) args in
+       let args1, args2 = List.chop n args in
+       let (c1,_), args1 = List.sep_last args1 in
+       let c = CProj (true, (f,us), args1, c1) in
+       if args2 = [] then c else CApp (CAst.make c, args2)
+    | _ ->
+       CAppExpl ((f,us), args)
 
 let extern_applied_syntactic_definition inctx n extraimpl (cf,f) syndefargs extraargs =
   try
     let syndefargs = List.map (fun a -> (a,None)) syndefargs in
     let extraargs = adjust_implicit_arguments inctx n extraargs extraimpl in
     let args = syndefargs @ extraargs in
-    if args = [] then cf else CApp ((None, CAst.make cf), args)
+    if args = [] then cf else CApp (CAst.make cf, args)
   with Expl ->
     let args = syndefargs @ List.map Lazy.force extraargs in
-    CAppExpl ((None,f,None), args)
+    CAppExpl ((f,None), args)
 
 let mkFlattenedCApp (head,args) =
   match head.CAst.v with
@@ -785,7 +802,7 @@ let mkFlattenedCApp (head,args) =
     (* or after removal of a coercion to funclass *)
     CApp (g,args'@args)
   | _ ->
-    CApp ((None, head), args)
+    CApp (head, args)
 
 let extern_applied_notation inctx n impl f args =
   if List.is_empty args then
@@ -908,20 +925,12 @@ let q_infinity () = qualid_of_ref "num.float.infinity"
 let q_neg_infinity () = qualid_of_ref "num.float.neg_infinity"
 let q_nan () = qualid_of_ref "num.float.nan"
 
-let get_printing_float = Goptions.declare_bool_option_and_ref
-    ~depr:false
-    ~key:["Printing";"Float"]
-    ~value:true
-
 let extern_float f scopes =
   if Float64.is_nan f then CRef(q_nan (), None)
   else if Float64.is_infinity f then CRef(q_infinity (), None)
   else if Float64.is_neg_infinity f then CRef(q_neg_infinity (), None)
   else
-    let s =
-      let hex = !Flags.raw_print || not (get_printing_float ()) in
-      if hex then Float64.to_hex_string f else Float64.to_string f in
-    let n = NumTok.Signed.of_string s in
+    let n = NumTok.Signed.of_string (Float64.to_hex_string f) in
     extern_prim_token_delimiter_if_required (Number n)
       "float" "float_scope" scopes
 
@@ -977,6 +986,9 @@ let rec extern inctx ?impargs scopes vars r =
     | GInt i when Coqlib.has_ref "num.int63.wrap_int" ->
        let wrap = Coqlib.lib_ref "num.int63.wrap_int" in
        DAst.make (GApp (DAst.make (GRef (wrap, None)), [r]))
+    | GFloat f when Coqlib.has_ref "num.float.wrap_float" ->
+       let wrap = Coqlib.lib_ref "num.float.wrap_float" in
+       DAst.make (GApp (DAst.make (GRef (wrap, None)), [r]))
     | _ -> r in
 
   try extern_notations inctx scopes vars None r'
@@ -1029,11 +1041,16 @@ let rec extern inctx ?impargs scopes vars r =
                extern_applied_ref inctx
                  (select_stronger_impargs (implicits_of_global ref))
                  (ref,extern_reference ?loc (fst vars) ref) (extern_instance (snd vars) us) args)
+         | GProj (f,params,c) ->
+             extern_applied_proj inctx scopes vars f params c args
          | _ ->
              let args = List.map (fun c -> (sub_extern true scopes vars c,None)) args in
              let head = sub_extern false scopes vars f in
              mkFlattenedCApp (head,args))
 
+  | GProj (f,params,c) ->
+      extern_applied_proj inctx scopes vars f params c []
+
   | GLetIn (na,b,t,c) ->
       CLetIn (make ?loc na,sub_extern (Option.has_some t) scopes vars b,
               Option.map (extern_typ scopes vars) t,
@@ -1127,9 +1144,8 @@ let rec extern inctx ?impargs scopes vars r =
 
   | GHole (e,naming,_) -> CHole (Some e, naming, None) (* TODO: extern tactics. *)
 
-  | GCast (c, c') ->
-      CCast (sub_extern true scopes vars c,
-             map_cast_type (extern_typ scopes vars) c')
+  | GCast (c, k, c') ->
+    CCast (sub_extern true scopes vars c, k, extern_typ scopes vars c')
 
   | GInt i ->
      extern_prim_token_delimiter_if_required
@@ -1381,6 +1397,18 @@ and extern_notation inctx (custom,scopes as allscopes) vars t rules =
       with
           No_match -> extern_notation inctx allscopes vars t rules
 
+and extern_applied_proj inctx scopes vars (cst,us) params c extraargs =
+  let ref = GlobRef.ConstRef cst in
+  let subscopes = find_arguments_scope ref in
+  let nparams = List.length params in
+  let args = params @ c :: extraargs in
+  let args = fill_arg_scopes args subscopes scopes in
+  let args = extern_args (extern true) vars args in
+  let imps = select_stronger_impargs (implicits_of_global ref) in
+  let f = extern_reference (fst vars) ref in
+  let us = extern_instance (snd vars) us in
+  extern_projection inctx (f,us) nparams args imps
+
 let extern_glob_constr vars c =
   extern false (InConstrEntrySomeLevel,(None,[])) vars c
 
@@ -1459,16 +1487,16 @@ let rec glob_of_pat avoid env sigma pat = DAst.make @@ match pat with
       GApp (DAst.make @@ GPatVar (Evar_kinds.SecondOrderPatVar n),
         List.map (glob_of_pat avoid env sigma) args)
   | PProd (na,t,c) ->
-      let na',avoid' = compute_displayed_name_in_pattern sigma avoid na c in
+      let na',avoid' = compute_displayed_name_in_pattern (Global.env ()) sigma avoid na c in
       let env' = Termops.add_name na' env in
       GProd (na',Explicit,glob_of_pat avoid env sigma t,glob_of_pat avoid' env' sigma c)
   | PLetIn (na,b,t,c) ->
-      let na',avoid' = Namegen.compute_displayed_let_name_in sigma Namegen.RenamingForGoal avoid na c in
+      let na',avoid' = Namegen.compute_displayed_let_name_in (Global.env ()) sigma Namegen.RenamingForGoal avoid na c in
       let env' = Termops.add_name na' env in
       GLetIn (na',glob_of_pat avoid env sigma b, Option.map (glob_of_pat avoid env sigma) t,
               glob_of_pat avoid' env' sigma c)
   | PLambda (na,t,c) ->
-      let na',avoid' = compute_displayed_name_in_pattern sigma avoid na c in
+      let na',avoid' = compute_displayed_name_in_pattern (Global.env ()) sigma avoid na c in
       let env' = Termops.add_name na' env in
       GLambda (na',Explicit,glob_of_pat avoid env sigma t, glob_of_pat avoid' env' sigma c)
   | PIf (c,b1,b2) ->
@@ -1546,7 +1574,7 @@ let rec glob_of_pat avoid env sigma pat = DAst.make @@ match pat with
 
 and glob_of_pat_under_context avoid env sigma (nas, pat) =
   let fold (avoid, env, nas, epat) na =
-    let na, avoid = compute_displayed_name_in_pattern sigma avoid na epat in
+    let na, avoid = compute_displayed_name_in_pattern (Global.env ()) sigma avoid na epat in
     let env = Termops.add_name na env in
     let epat = match epat with PLambda (_, _, p) -> p | _ -> assert false in
     (avoid, env, na :: nas, epat)
