@@ -32,12 +32,6 @@ let rec alpha_var id1 id2 = function
   | _::idl -> alpha_var id1 id2 idl
   | [] -> Id.equal id1 id2
 
-let cast_type_iter2 f t1 t2 = match t1, t2 with
-  | CastConv t1, CastConv t2 -> f t1 t2
-  | CastVM t1, CastVM t2 -> f t1 t2
-  | CastNative t1, CastNative t2 -> f t1 t2
-  | (CastConv _ | CastVM _ | CastNative _), _ -> raise Exit
-
 (* used to update the notation variable with the local variables used
    in NList and NBinderList, since the iterator has its own variable *)
 let replace_var i j var = j :: List.remove Id.equal i var
@@ -150,6 +144,9 @@ let compare_notation_constr lt (vars1,vars2) t1 t2 =
     (* it had to be a single iteration of iter1 *)
     aux vars renaming tail1 t2
   | NApp (t1, a1), NApp (t2, a2) -> aux vars renaming t1 t2; List.iter2 (aux vars renaming) a1 a2
+  | NProj ((cst1,u1), l1, a1), NProj ((cst2,u2), l2, a2)
+    when GlobRef.equal (GlobRef.ConstRef cst1) (GlobRef.ConstRef cst2) && compare_glob_universe_instances lt strictly_lt u1 u2 ->
+    List.iter2 (aux vars renaming) l1 l2; aux vars renaming a1 a2
   | NLambda (na1, t1, u1), NLambda (na2, t2, u2)
   | NProd (na1, t1, u1), NProd (na2, t2, u2) ->
     (match t1, t2 with
@@ -202,16 +199,17 @@ let compare_notation_constr lt (vars1,vars2) t1 t2 =
     Array.iter3 (aux vars) renamings us1 us2;
     Array.iter3 (aux vars) (Array.map ((@) renaming) renamings) rs1 rs2
   | NSort s1, NSort s2 when glob_sort_eq s1 s2 -> ()
-  | NCast (t1, c1), NCast (t2, c2) ->
-    aux vars renaming t1 t2;
-    cast_type_iter2 (aux vars renaming) c1 c2
+  | NCast (c1, k1, t1), NCast (c2, k2, t2) ->
+    aux vars renaming c1 c2;
+    if not (cast_kind_eq k1 k2) then raise Exit;
+    aux vars renaming t1 t2
   | NInt i1, NInt i2 when Uint63.equal i1 i2 -> ()
   | NFloat f1, NFloat f2 when Float64.equal f1 f2 -> ()
   | NArray(t1,def1,ty1), NArray(t2,def2,ty2) ->
     Array.iter2 (aux vars renaming) t1 t2;
     aux vars renaming def1 def2;
     aux vars renaming ty1 ty2
-  | (NRef _ | NVar _ | NApp _ | NHole _ | NList _ | NLambda _ | NProd _
+  | (NRef _ | NVar _ | NApp _ | NProj _ | NHole _ | NList _ | NLambda _ | NProd _
     | NBinderList _ | NLetIn _ | NCases _ | NLetTuple _ | NIf _
     | NRec _ | NSort _ | NCast _ | NInt _ | NFloat _ | NArray _), _ -> raise Exit in
   try
@@ -327,6 +325,7 @@ let glob_constr_of_notation_constr_with_binders ?loc g f ?(h=default_binder_stat
   let lt x = DAst.make ?loc x in lt @@ match nc with
   | NVar id -> GVar id
   | NApp (a,args) -> let e = h.no e in DAst.get (mkGApp (f e a) (List.map (f e) args))
+  | NProj (p,args,c) -> let e = h.no e in GProj (p, List.map (f e) args, f e c)
   | NList (x,y,iter,tail,swap) ->
       let t = f e tail in let it = f e iter in
       let innerl = (ldots_var,t)::(if swap then [y, lt @@ GVar x] else []) in
@@ -395,7 +394,7 @@ let glob_constr_of_notation_constr_with_binders ?loc g f ?(h=default_binder_stat
           (e,(na,Explicit,Option.map (f e) oc,f e b)))) e dll in
       let e',idl = Array.fold_left_map (to_id (protect g)) e idl in
       GRec (fk,idl,dll,Array.map (f e) tl,Array.map (f e') bl)
-  | NCast (c,k) -> GCast (f e c,map_cast_type (f (h.slide e)) k)
+  | NCast (c,k,t) -> GCast (f e c, k, f (h.slide e) t)
   | NSort x -> GSort x
   | NHole (x, naming, arg)  -> GHole (x, naming, arg)
   | NRef (x,u) -> GRef (x,u)
@@ -600,6 +599,7 @@ let notation_constr_and_vars_of_glob_constr recvars a =
   | GApp (g,args) ->
      (* Treat applicative notes as binary nodes *)
      let a,args = List.sep_last args in mkNApp1 (aux (DAst.make (GApp (g, args))), aux a)
+  | GProj (p,args,c) -> NProj (p, List.map aux args, aux c)
   | GLambda (na,bk,ty,c) -> add_name found na; NLambda (na,aux_type ty,aux c)
   | GProd (na,bk,ty,c) -> add_name found na; NProd (na,aux_type ty,aux c)
   | GLetIn (na,b,t,c) -> add_name found na; NLetIn (na,aux b,Option.map aux t, aux c)
@@ -626,7 +626,7 @@ let notation_constr_and_vars_of_glob_constr recvars a =
            user_err Pp.(str "Binders marked as implicit not allowed in notations.");
          add_name found na; (na,Option.map aux oc,aux b))) dll in
       NRec (fk,idl,dll,Array.map aux tl,Array.map aux bl)
-  | GCast (c,k) -> NCast (aux c,map_cast_type aux k)
+  | GCast (c,k,t) -> NCast (aux c, k, aux t)
   | GSort s -> NSort s
   | GInt i -> NInt i
   | GFloat f -> NFloat f
@@ -742,6 +742,15 @@ let rec subst_notation_constr subst bound raw =
         if r' == r && rl' == rl then raw else
           NApp(r',rl')
 
+  | NProj ((cst,u),rl,r) ->
+      let ref = GlobRef.ConstRef cst in
+      let ref',t = subst_global subst ref in
+      assert (t = None);
+      let rl' = List.Smart.map (subst_notation_constr subst bound) rl
+      and r' = subst_notation_constr subst bound r in
+        if ref' == ref && rl' == rl && r' == r then raw else
+          NProj ((destConstRef ref',u),rl',r')
+
   | NList (id1,id2,r1,r2,b) ->
       let r1' = subst_notation_constr subst bound r1
       and r2' = subst_notation_constr subst bound r2 in
@@ -836,10 +845,10 @@ let rec subst_notation_constr subst bound raw =
     if nsolve == solve && nknd == knd then raw
     else NHole (nknd, naming, nsolve)
 
-  | NCast (r1,k) ->
+  | NCast (r1,k,t) ->
       let r1' = subst_notation_constr subst bound r1 in
-      let k' = smartmap_cast_type (subst_notation_constr subst bound) k in
-      if r1' == r1 && k' == k then raw else NCast(r1',k')
+      let t' = subst_notation_constr subst bound t in
+      if r1' == r1 && t' == t then raw else NCast(r1',k,t')
 
   | NArray (t,def,ty) ->
       let def' = subst_notation_constr subst bound def
@@ -1312,16 +1321,6 @@ let match_termlist match_fun alp metas sigma rest x y iter termin revert =
   else
     bind_termlist_env alp sigma x l
 
-let match_cast match_fun sigma c1 c2 =
-  match c1, c2 with
-  | CastConv t1, CastConv t2
-  | CastVM t1, CastVM t2
-  | CastNative t1, CastNative t2 ->
-    match_fun sigma t1 t2
-  | CastConv _, _
-  | CastVM _, _
-  | CastNative _, _ -> raise No_match
-
 let does_not_come_from_already_eta_expanded_var glob =
   (* This is hack to avoid looping on a rule with rhs of the form *)
   (* "?f (fun ?x => ?g)" since otherwise, matching "F H" expands in *)
@@ -1374,6 +1373,14 @@ let rec match_ inner u alp metas sigma a1 a2 =
       let may_use_eta = does_not_come_from_already_eta_expanded_var f1 in
       List.fold_left2 (match_ may_use_eta u alp metas)
         (match_hd u alp metas sigma f1 f2) l1 l2
+  | GProj ((cst1,u1),l1,a1), NProj ((cst2,u2),l2,a2) when GlobRef.equal (GlobRef.ConstRef cst1) (GlobRef.ConstRef cst2) && compare_glob_universe_instances_le u1 u2 ->
+     match_in u alp metas (List.fold_left2 (match_in u alp metas) sigma l1 l2) a1 a2
+  | GApp (f1,l1), NProj ((cst2,u2),l2,a2) ->
+     (match DAst.get f1 with
+     | GRef (r1,u1) when GlobRef.equal r1 (GlobRef.ConstRef cst2) && compare_glob_universe_instances_le u1 u2 &&
+         List.length l1 = List.length l2 + 1 ->
+        List.fold_left2 (match_in u alp metas) sigma l1 (l2@[a2])
+     | _ -> raise No_match)
   | GLambda (na1,bk1,t1,b1), NLambda (na2,t2,b2) ->
      match_extended_binders false u alp metas na1 na2 bk1 t1 (match_in_type u alp metas sigma t1 t2) b1 b2
   | GProd (na1,bk1,t1,b1), NProd (na2,t2,b2) ->
@@ -1427,8 +1434,10 @@ let rec match_ inner u alp metas sigma a1 a2 =
       let alp,sigma = Array.fold_right2 (fun id1 id2 alsig ->
         match_names metas alsig (Name id1) (Name id2)) idl1 idl2 (alp,sigma) in
       Array.fold_left2 (match_in u alp metas) sigma bl1 bl2
-  | GCast(t1, c1), NCast(t2, c2) ->
-    match_cast (match_in u alp metas) (match_in u alp metas sigma t1 t2) c1 c2
+  | GCast(c1, k1, t1), NCast(c2, k2, t2) ->
+    let sigma = match_in u alp metas sigma c1 c2 in
+    if not (cast_kind_eq k1 k2) then raise No_match;
+    match_in u alp metas sigma t1 t2
 
   (* Next pair of lines useful only if not coming from detyping *)
   | GSort (UNamed [(GProp|GSet),0]), NSort (UAnonymous _) -> raise No_match
@@ -1469,7 +1478,7 @@ let rec match_ inner u alp metas sigma a1 a2 =
       Array.fold_left2 (match_in u alp metas) sigma t nt
     else raise No_match
 
-  | (GRef _ | GVar _ | GEvar _ | GPatVar _ | GApp _ | GLambda _ | GProd _
+  | (GRef _ | GVar _ | GEvar _ | GPatVar _ | GApp _ | GProj _ | GLambda _ | GProd _
      | GLetIn _ | GCases _ | GLetTuple _ | GIf _ | GRec _ | GSort _ | GHole _
      | GCast _ | GInt _ | GFloat _ | GArray _), _ -> raise No_match
 
