@@ -49,7 +49,7 @@ type signature_mismatch_error =
   | NotEqualInductiveAliases
   | IncompatibleUniverses of Univ.univ_inconsistency
   | IncompatiblePolymorphism of env * types * types
-  | IncompatibleConstraints of { got : Univ.AUContext.t; expect : Univ.AUContext.t }
+  | IncompatibleConstraints of { got : Univ.AbstractContext.t; expect : Univ.AbstractContext.t }
   | IncompatibleVariance
 
 type module_typing_error =
@@ -137,10 +137,6 @@ let rec functor_smart_map fty f0 funct = match funct with
   |NoFunctor a ->
     let a' = f0 a in if a==a' then funct else NoFunctor a'
 
-let rec functor_iter fty f0 = function
-  |MoreFunctor (_mbid,ty,e) -> fty ty; functor_iter fty f0 e
-  |NoFunctor a -> f0 a
-
 (** {6 Misc operations } *)
 
 let module_type_of_module mb =
@@ -164,11 +160,6 @@ let implem_smartmap fs fa impl = match impl with
   |Algebraic a -> let a' = fa a in if a==a' then impl else Algebraic a'
   |Abstract|FullStruct -> impl
 
-let implem_iter fs fa impl = match impl with
-  |Struct e -> fs e
-  |Algebraic a -> fa a
-  |Abstract|FullStruct -> ()
-
 (** {6 Substitutions of modular structures } *)
 
 let id_delta x _y = x
@@ -269,6 +260,7 @@ let do_delta_dom reso sub = subst_dom_delta_resolver sub reso
 let do_delta_codom reso sub = subst_codom_delta_resolver sub reso
 let do_delta_dom_codom reso sub = subst_dom_codom_delta_resolver sub reso
 
+let subst_dom_codom_signature subst = subst_signature subst do_delta_dom_codom
 let subst_signature subst = subst_signature subst do_delta_codom
 let subst_structure subst = subst_structure subst do_delta_codom
 
@@ -541,11 +533,9 @@ let strengthen_and_subst_mb mb mp include_b = match mb.mod_type with
     let subst = map_mp mb.mod_mp mp empty_delta_resolver in
     subst_module subst do_delta_dom_codom mb
 
-let subst_modtype_and_resolver mtb mp =
-  let subst = map_mp mtb.mod_mp mp empty_delta_resolver in
-  let new_delta = subst_dom_codom_delta_resolver subst mtb.mod_delta in
-  let full_subst = map_mp mtb.mod_mp mp new_delta in
-  subst_modtype full_subst do_delta_dom_codom mtb
+let subst_modtype_signature_and_resolver mp_from mp_to sign reso =
+  let subst = map_mp mp_from mp_to empty_delta_resolver in
+  subst_dom_codom_signature subst sign, subst_dom_codom_delta_resolver subst reso
 
 (** {6 Cleaning a module expression from bounded parts }
 
@@ -601,27 +591,3 @@ let rec collect_mbid l sign =  match sign with
 
 let clean_bounded_mod_expr sign =
   if is_functor sign then collect_mbid MBIset.empty sign else sign
-
-(** {6 Stm machinery } *)
-let join_constant_body except otab cb =
-  match cb.const_body with
-  | OpaqueDef o -> Opaqueproof.join_opaque ~except otab o
-  | _ -> ()
-
-let join_structure except otab s =
-  let rec join_module : 'a. 'a generic_module_body -> unit = fun mb ->
-    Option.iter join_expression mb.mod_type_alg;
-    join_signature mb.mod_type
-  and join_field (_l,body) = match body with
-    |SFBconst sb -> join_constant_body except otab sb
-    |SFBmind _ -> ()
-    |SFBmodule m ->
-      implem_iter join_signature join_expression m.mod_expr;
-      join_module m
-    |SFBmodtype m -> join_module m
-  and join_structure struc = List.iter join_field struc
-  and join_signature sign =
-    functor_iter join_module join_structure sign
-  and join_expression me = functor_iter join_module (fun _ -> ()) me in
-  join_structure s
-
