@@ -65,16 +65,9 @@ struct
 
   let print id = str id
 
-  module Self =
-  struct
-    type t = string
-    let compare = compare
-  end
-
-  module Set = Set.Make(Self)
-  module Map = CMap.Make(Self)
-
-  module Pred = Predicate.Make(Self)
+  module Set = CString.Set
+  module Map = CString.Map
+  module Pred = CString.Pred
 
   module List = String.List
 
@@ -1090,6 +1083,11 @@ module GlobRef = struct
     let hash gr = GlobRefInternal.global_hash_gen Constant.SyntacticOrd.hash Ind.SyntacticOrd.hash Construct.SyntacticOrd.hash gr
   end
 
+  let is_bound = function
+  | VarRef _ -> false
+  | ConstRef cst -> ModPath.is_bound (Constant.modpath cst)
+  | IndRef (ind,_) | ConstructRef ((ind,_),_) -> ModPath.is_bound (MutInd.modpath ind)
+
   module Map = HMap.Make(CanOrd)
   module Set = Map.Set
 
