@@ -879,7 +879,7 @@ let compile ~fail_on_error ?universes:(universes=0) env sigma c =
   with TooLargeInductive msg as exn ->
     let _, info = Exninfo.capture exn in
     let fn = if fail_on_error then
-        CErrors.user_err ?loc:None ~info ~hdr:"compile"
+        CErrors.user_err ?loc:None ~info
       else
         (fun x -> Feedback.msg_warning x) in
     fn msg; None
@@ -887,7 +887,7 @@ let compile ~fail_on_error ?universes:(universes=0) env sigma c =
 let compile_constant_body ~fail_on_error env univs = function
   | Undef _ | OpaqueDef _ | Primitive _ -> Some BCconstant
   | Def body ->
-      let instance_size = Univ.AUContext.size (Declareops.universes_context univs) in
+      let instance_size = Univ.AbstractContext.size (Declareops.universes_context univs) in
       match kind body with
         | Const (kn',u) when is_univ_copy instance_size u ->
             (* we use the canonical name of the constant*)
