@@ -88,8 +88,8 @@ let classify_segment seg =
     end
     | (_,OpenedSection _) :: _ -> user_err Pp.(str "there are still opened sections")
     | (_,OpenedModule (ty,_,_,_)) :: _ ->
-      user_err ~hdr:"Lib.classify_segment"
-        (str "there are still opened " ++ str (module_kind ty) ++ str "s")
+      user_err
+        (str "there are still opened " ++ str (module_kind ty) ++ str "s.")
   in
     clean ([],[],[]) (List.rev seg)
 
@@ -277,10 +277,10 @@ let start_mod is_type export id mp fs =
   let prefix = Nametab.{ obj_dir = dir; obj_mp = mp; } in
   let exists =
     if is_type then Nametab.exists_cci (make_path id)
-    else Nametab.exists_dir dir
+    else Nametab.exists_module dir
   in
   if exists then
-    user_err ~hdr:"open_module" (Id.print id ++ str " already exists.");
+    user_err (Id.print id ++ str " already exists.");
   add_entry (make_foname id) (OpenedModule (is_type,export,prefix,fs));
   lib_state := { !lib_state with path_prefix = prefix} ;
   prefix
@@ -340,7 +340,7 @@ let open_blocks_message es =
 let end_compilation_checks dir =
   let _ = match find_entries_p is_opening_node with
     | [] -> ()
-    | es -> user_err ~hdr:"Lib.end_compilation_checks" (open_blocks_message es) in
+    | es -> user_err (open_blocks_message es) in
   let is_opening_lib = function _,CompilingLibrary _ -> true | _ -> false
   in
   let oname =
@@ -365,22 +365,23 @@ let end_compilation oname =
   !lib_state.path_prefix,after
 
 (* Returns true if we are inside an opened module or module type *)
+let is_module_or_modtype () =
+  match Safe_typing.module_is_modtype (Global.safe_env ()) with
+  | [] -> false
+  | _ -> true
 
-let is_module_gen which check =
-  let test = function
-    | _, OpenedModule (ty,_,_,_) -> which ty
-    | _ -> false
-  in
-  try
-    match find_entry_p test with
-    | _, OpenedModule (ty,_,_,_) -> check ty
-    | _ -> assert false
-  with Not_found -> false
+let is_modtype () =
+  let modules = Safe_typing.module_is_modtype (Global.safe_env ()) in
+  List.exists (fun x -> x) modules
 
-let is_module_or_modtype () = is_module_gen (fun _ -> true) (fun _ -> true)
-let is_modtype () = is_module_gen (fun b -> b) (fun _ -> true)
-let is_modtype_strict () = is_module_gen (fun _ -> true) (fun b -> b)
-let is_module () = is_module_gen (fun b -> not b) (fun _ -> true)
+let is_modtype_strict () =
+  match Safe_typing.module_is_modtype (Global.safe_env ()) with
+  | b :: _ -> b
+  | [] -> false
+
+let is_module () =
+  let modules = Safe_typing.module_is_modtype (Global.safe_env ()) in
+  List.exists (fun x -> not x) modules
 
 (* Returns the opening node of a given name *)
 let find_opening_node id =
@@ -388,7 +389,7 @@ let find_opening_node id =
     let oname,entry = find_entry_p is_opening_node in
     let id' = basename (fst oname) in
     if not (Names.Id.equal id id') then
-      user_err ~hdr:"Lib.find_opening_node"
+      user_err
         (str "Last block to end has name " ++ Id.print id' ++ str ".");
     entry
   with Not_found -> user_err Pp.(str "There is nothing to end.")
@@ -406,8 +407,8 @@ let instance_from_variable_context =
   List.rev %> List.filter is_local_assum %> List.map NamedDecl.get_id %> Array.of_list
 
 let extract_worklist info =
-  let args = instance_from_variable_context info.Section.abstr_ctx in
-  info.Section.abstr_subst, args
+  let args = instance_from_variable_context info.Declarations.abstr_ctx in
+  info.Declarations.abstr_subst, args
 
 let sections () = Safe_typing.sections_of_safe_env @@ Global.safe_env ()
 
@@ -433,7 +434,7 @@ let section_segment_of_reference = let open GlobRef in function
 | VarRef _ -> empty_segment
 
 let variable_section_segment_of_reference gr =
-  (section_segment_of_reference gr).Section.abstr_ctx
+  (section_segment_of_reference gr).Declarations.abstr_ctx
 
 let is_in_section ref = match sections () with
   | None -> false
@@ -459,7 +460,7 @@ let open_section id =
   let obj_dir = add_dirpath_suffix opp.Nametab.obj_dir id in
   let prefix = Nametab.{ obj_dir; obj_mp = opp.obj_mp; } in
   if Nametab.exists_dir obj_dir then
-    user_err ~hdr:"open_section" (Id.print id ++ str " already exists.");
+    user_err (Id.print id ++ str " already exists.");
   let fs = Summary.freeze_summaries ~marshallable:false in
   add_entry (make_foname id) (OpenedSection (prefix, fs));
   (*Pushed for the lifetime of the section: removed by unfrozing the summary*)
@@ -551,10 +552,10 @@ let discharge_proj_repr =
       let _, newpars = Mindmap.find mind (snd modlist) in
       mind, npars + Array.length newpars)
 
-let discharge_abstract_universe_context { Section.abstr_subst = subst; abstr_uctx = abs_ctx } auctx =
+let discharge_abstract_universe_context { Declarations.abstr_subst = subst; abstr_uctx = abs_ctx } auctx =
   let open Univ in
   let ainst = make_abstract_instance auctx in
   let subst = Instance.append subst ainst in
   let subst = make_instance_subst subst in
   let auctx = Univ.subst_univs_level_abstract_universe_context subst auctx in
-  subst, AUContext.union abs_ctx auctx
+  subst, AbstractContext.union abs_ctx auctx
