@@ -392,7 +392,6 @@ let comm_loc bp = match !comment_begin with
 let comments = ref []
 let current_comment = Buffer.create 8192
 let between_commands = ref true
-let loc_offset = ref 0
 
 let real_push_char c = Buffer.add_char current_comment c
 
@@ -737,9 +736,11 @@ let rec next_token ~diff_mode loc s =
         IDENT id, set_loc_pos loc bp ep end
   | Some ('0'..'9') ->
       let n = NumTok.Unsigned.parse s in
-      let ep = Stream.count s in
       comment_stop bp;
-      (NUMBER n, set_loc_pos loc bp ep)
+      begin try find_keyword loc (NumTok.Unsigned.sprint n) bp s
+      with Not_found ->
+        let ep = Stream.count s in
+        NUMBER n, set_loc_pos loc bp ep end
   | Some '\"' ->
       Stream.junk s;
       let (loc, len) =
@@ -789,50 +790,21 @@ let rec next_token ~diff_mode loc s =
         | EmptyStream ->
             comment_stop bp; (EOI, set_loc_pos loc bp (bp+1))
 
-let next_token ~diff_mode loc s =
-  let (t,loc as r) = next_token ~diff_mode loc s in
-  let open Loc in
- (* Debug: uncomment this for tracing tokens seen by coq...*)
-(*  Printf.eprintf "(line %i, %i-%i)[%s]\n%!" loc.line_nb loc.bp loc.ep (Tok.extract_string diff_mode t);*)
-  (t, {loc with bp = loc.bp + !loc_offset; ep = loc.ep + !loc_offset})
-
 (** {6 The lexer of Coq} *)
 
-(** Note: removing a token.
-   We do nothing because [remove_token] is called only when removing a grammar
-   rule with [Grammar.delete_rule]. The latter command is called only when
-   unfreezing the state of the grammar entries (see GRAMMAR summary, file
-   env/metasyntax.ml). Therefore, instead of removing tokens one by one,
-   we unfreeze the state of the lexer. This restores the behaviour of the
-   lexer. B.B. *)
-
-(** Names of tokens, for this lexer, used in Grammar error messages *)
-
-let token_text : type c. c Tok.p -> string = function
-  | PKEYWORD t -> "'" ^ t ^ "'"
-  | PIDENT None -> "identifier"
-  | PIDENT (Some t) -> "'" ^ t ^ "'"
-  | PNUMBER None -> "number"
-  | PNUMBER (Some n) -> "'" ^ NumTok.Unsigned.sprint n ^ "'"
-  | PSTRING None -> "string"
-  | PSTRING (Some s) -> "STRING \"" ^ s ^ "\""
-  | PLEFTQMARK -> "LEFTQMARK"
-  | PEOI -> "end of input"
-  | PPATTERNIDENT None -> "PATTERNIDENT"
-  | PPATTERNIDENT (Some s) -> "PATTERNIDENT \"" ^ s ^ "\""
-  | PFIELD None -> "FIELD"
-  | PFIELD (Some s) -> "FIELD \"" ^ s ^ "\""
-  | PBULLET None -> "BULLET"
-  | PBULLET (Some s) -> "BULLET \"" ^ s ^ "\""
-  | PQUOTATION lbl -> "QUOTATION \"" ^ lbl ^ "\""
-
 let func next_token ?(loc=Loc.(initial ToplevelInput)) cs =
+  let bp_ = Loc.(loc.bp) in
   let cur_loc = ref loc in
   LStream.from ~loc
     (fun i ->
       let (tok, loc) = next_token !cur_loc cs in
       cur_loc := after loc;
-      Some (tok,loc))
+      let aloc = Loc.{loc with bol_pos = loc.bol_pos + bp_;
+                                    bp = loc.bp + bp_;
+                                    ep = loc.ep + bp_} in
+     (* Debug: uncomment this for tracing tokens seen by coq...*)
+     (*  Printf.eprintf "(line %i, %i-%i)[%s]\n%!" aloc.line_nb aloc.bp aloc.ep (Tok.extract_string diff_mode t);*)
+      Some (tok,aloc))
 
 module MakeLexer (Diff : sig val mode : bool end) = struct
   type te = Tok.t
@@ -844,9 +816,14 @@ module MakeLexer (Diff : sig val mode : bool end) = struct
     | PKEYWORD s -> add_keyword ~quotation:NoQuotation s
     | PQUOTATION s -> add_keyword ~quotation:Quotation s
     | _ -> ()
-  let tok_removing = (fun _ -> ())
+  let tok_removing : type c. c pattern -> unit = function
+    (* Normally useless because backtracking relies on state freezing *)
+    (* Nevertheless consistent with tok_using *)
+    | PKEYWORD s -> remove_keyword s
+    | PQUOTATION s -> remove_keyword s
+    | _ -> ()
   let tok_match = Tok.match_pattern
-  let tok_text = token_text
+  let tok_text = Tok.token_text
 
   (* The state of the lexer visible from outside *)
   module State = struct
@@ -863,7 +840,6 @@ module MakeLexer (Diff : sig val mode : bool end) = struct
       (!comment_begin, Buffer.contents current_comment, !between_commands, !comments)
     let drop () = set (init ())
     let get_comments (_,_,_,c) = c
-    let set_loc_offset off = loc_offset := off
 
   end
 end
