@@ -68,8 +68,8 @@ let build_newrecursive lnameargsardef =
         let def = abstract_glob_constr body_def binders in
         interp_casted_constr_with_implicits rec_sign sigma rec_impls def
       | None ->
-        CErrors.user_err ~hdr:"Function"
-          (Pp.str "Body of Function must be given")
+        CErrors.user_err
+          (Pp.str "Body of Function must be given.")
     in
     Vernacstate.System.protect (List.map f) lnameargsardef
   in
@@ -85,7 +85,7 @@ let is_rec names =
     | GVar id -> check_id id names
     | GRef _ | GEvar _ | GPatVar _ | GSort _ | GHole _ | GInt _ | GFloat _ ->
       false
-    | GCast (b, _) -> lookup names b
+    | GCast (b, _, _) -> lookup names b
     | GRec _ -> CErrors.user_err (Pp.str "GRec not handled")
     | GIf (b, _, lhs, rhs) ->
       lookup names b || lookup names lhs || lookup names rhs
@@ -103,7 +103,7 @@ let is_rec names =
               (fun acc na -> Nameops.Name.fold_right Id.Set.remove na acc)
               names nal)
            b
-    | GApp (f, args) -> List.exists (lookup names) (f :: args)
+    | GApp (c, args) | GProj (_, args, c) -> List.exists (lookup names) (c :: args)
     | GArray (_u, t, def, ty) ->
       Array.exists (lookup names) t || lookup names def || lookup names ty
     | GCases (_, _, el, brl) ->
@@ -236,7 +236,7 @@ let change_property_sort evd toSort princ princName =
         Global.add_constraints
           (Univ.enforce_leq
              (Sorts.univ_of_sort toSort)
-             (Sorts.univ_of_sort s) Univ.Constraint.empty);
+             (Sorts.univ_of_sort s) Univ.Constraints.empty);
         Term.compose_prod args (Constr.mkSort toSort) )
   in
   let evd, princName_as_constr =
@@ -397,8 +397,8 @@ let register_struct is_rec fixpoint_exprl =
       match body_def with
       | Some body -> body
       | None ->
-        CErrors.user_err ~hdr:"Function"
-          Pp.(str "Body of Function must be given")
+        CErrors.user_err
+          Pp.(str "Body of Function must be given.")
     in
     ComDefinition.do_definition ~name:fname.CAst.v ~poly:false
       ~scope:(Locality.Global Locality.ImportDefaultBehavior)
@@ -489,10 +489,10 @@ let generate_type evd g_to_f f graph =
   in
   let named_ctxt = Id.Set.of_list (List.map_filter filter fun_ctxt) in
   let res_id =
-    Namegen.next_ident_away_in_goal (Id.of_string "_res") named_ctxt
+    Namegen.next_ident_away_in_goal (Global.env ()) (Id.of_string "_res") named_ctxt
   in
   let fv_id =
-    Namegen.next_ident_away_in_goal (Id.of_string "fv")
+    Namegen.next_ident_away_in_goal (Global.env ()) (Id.of_string "fv")
       (Id.Set.add res_id named_ctxt)
   in
   (*i we can then type the argument to be applied to the function [f] i*)
@@ -588,7 +588,7 @@ let find_induction_principle evd f =
 let rec generate_fresh_id x avoid i =
   if i == 0 then []
   else
-    let id = Namegen.next_ident_away_in_goal x (Id.Set.of_list avoid) in
+    let id = Namegen.next_ident_away_in_goal (Global.env ()) x (Id.Set.of_list avoid) in
     id :: generate_fresh_id x (id :: avoid) (pred i)
 
 let prove_fun_correct evd graphs_constr schemes lemmas_types_infos i :
@@ -596,9 +596,9 @@ let prove_fun_correct evd graphs_constr schemes lemmas_types_infos i :
   let open Constr in
   let open EConstr in
   let open Context.Rel.Declaration in
-  let open Tacmach.New in
+  let open Tacmach in
   let open Tactics in
-  let open Tacticals.New in
+  let open Tacticals in
   Proofview.Goal.enter (fun g ->
       (* first of all we recreate the lemmas types to be used as predicates of the induction principle
          that is~:
@@ -623,7 +623,7 @@ let prove_fun_correct evd graphs_constr schemes lemmas_types_infos i :
          using a name
       *)
       let principle_id =
-        Namegen.next_ident_away_in_goal (Id.of_string "princ")
+        Namegen.next_ident_away_in_goal (Global.env ()) (Id.of_string "princ")
           (Id.Set.of_list ids)
       in
       let ids = principle_id :: ids in
@@ -874,7 +874,7 @@ let tauto =
 *)
 let generalize_dependent_of x hyp =
   let open Context.Named.Declaration in
-  let open Tacticals.New in
+  let open Tacticals in
   Proofview.Goal.enter (fun g ->
       tclMAP
         (function
@@ -892,9 +892,9 @@ let rec intros_with_rewrite () =
 and intros_with_rewrite_aux () : unit Proofview.tactic =
   let open Constr in
   let open EConstr in
-  let open Tacmach.New in
+  let open Tacmach in
   let open Tactics in
-  let open Tacticals.New in
+  let open Tacticals in
   Proofview.Goal.enter (fun g ->
       let eq_ind = make_eq () in
       let sigma = Proofview.Goal.sigma g in
@@ -969,7 +969,7 @@ and intros_with_rewrite_aux () : unit Proofview.tactic =
         | Ind _
           when EConstr.eq_constr sigma t
                  (EConstr.of_constr
-                    ( UnivGen.constr_of_monomorphic_global
+                    ( UnivGen.constr_of_monomorphic_global (Global.env ())
                     @@ Coqlib.lib_ref "core.False.type" )) ->
           tauto
         | Case (_, _, _, _, _, v, _) ->
@@ -994,9 +994,9 @@ and intros_with_rewrite_aux () : unit Proofview.tactic =
 let rec reflexivity_with_destruct_cases () =
   let open Constr in
   let open EConstr in
-  let open Tacmach.New in
+  let open Tacmach in
   let open Tactics in
-  let open Tacticals.New in
+  let open Tacticals in
   Proofview.Goal.enter (fun g ->
       let destruct_case () =
         try
@@ -1018,11 +1018,8 @@ let rec reflexivity_with_destruct_cases () =
       let my_inj_flags =
         Some
           { Equality.keep_proof_equalities = false
-          ; injection_in_context = false
-          ; (* for compatibility, necessary *)
-            injection_pattern_l2r_order =
-              false
-              (* probably does not matter; except maybe with dependent hyps *)
+          ; injection_pattern_l2r_order = false
+            (* probably does not matter; except maybe with dependent hyps *)
           }
       in
       let discr_inject =
@@ -1045,7 +1042,7 @@ let rec reflexivity_with_destruct_cases () =
                         (Proofview.Goal.sigma g) ~keep_proofs:None t1 t2
                     then
                       tclTHENLIST
-                        [ Equality.injHyp my_inj_flags None id
+                        [ Equality.injHyp my_inj_flags ~injection_in_context:false None id
                         ; thin [id]
                         ; intros_with_rewrite () ]
                     else Proofview.tclUNIT ()
@@ -1068,9 +1065,9 @@ let rec reflexivity_with_destruct_cases () =
 let prove_fun_complete funcs graphs schemes lemmas_types_infos i :
     unit Proofview.tactic =
   let open EConstr in
-  let open Tacmach.New in
+  let open Tacmach in
   let open Tactics in
-  let open Tacticals.New in
+  let open Tacticals in
   Proofview.Goal.enter (fun g ->
       (* We compute the types of the different mutually recursive lemmas
          in $\zeta$ normal form
@@ -1660,7 +1657,7 @@ let register_wf interactive_proof ?(is_mes = false) fname rec_impls wf_rel_expr
     let f_app_args =
       CAst.make
       @@ Constrexpr.CAppExpl
-           ( (None, Libnames.qualid_of_ident fname, None)
+           ( (Libnames.qualid_of_ident fname, None)
            , List.map
                (function
                  | {CAst.v = Anonymous} -> assert false
@@ -1669,7 +1666,7 @@ let register_wf interactive_proof ?(is_mes = false) fname rec_impls wf_rel_expr
     in
     CAst.make
     @@ Constrexpr.CApp
-         ( (None, Constrexpr_ops.mkRefC (Libnames.qualid_of_string "Logic.eq"))
+         ( Constrexpr_ops.mkRefC (Libnames.qualid_of_string "Logic.eq")
          , [(f_app_args, None); (body, None)] )
   in
   let eq = Constrexpr_ops.mkCProdN args unbounded_eq in
@@ -1779,8 +1776,8 @@ let do_generate_principle_aux pconstants on_error register_built
         match body_def with
         | Some body -> body
         | None ->
-          CErrors.user_err ~hdr:"Function"
-            (Pp.str "Body of Function must be given")
+          CErrors.user_err
+            (Pp.str "Body of Function must be given.")
       in
       let recdefs, rec_impls = build_newrecursive fixpoint_exprl in
       let using_lemmas = [] in
@@ -1810,8 +1807,8 @@ let do_generate_principle_aux pconstants on_error register_built
         match body_def with
         | Some body -> body
         | None ->
-          CErrors.user_err ~hdr:"Function"
-            Pp.(str "Body of Function must be given")
+          CErrors.user_err
+            Pp.(str "Body of Function must be given.")
       in
       let pre_hook pconstants =
         generate_principle
@@ -1951,7 +1948,7 @@ let rec add_args id new_args =
   CAst.map (function
     | CRef (qid, _) as b ->
       if qualid_is_ident qid && Id.equal (qualid_basename qid) id then
-        CAppExpl ((None, qid, None), new_args)
+        CAppExpl ((qid, None), new_args)
       else b
     | CFix _ | CCoFix _ -> CErrors.anomaly ~label:"add_args " (Pp.str "todo.")
     | CProdN (nal, b1) ->
@@ -1990,15 +1987,20 @@ let rec add_args id new_args =
         , add_args id new_args b1
         , Option.map (add_args id new_args) t
         , add_args id new_args b2 )
-    | CAppExpl ((pf, qid, us), exprl) ->
+    | CAppExpl ((qid, us), exprl) ->
       if qualid_is_ident qid && Id.equal (qualid_basename qid) id then
         CAppExpl
-          ((pf, qid, us), new_args @ List.map (add_args id new_args) exprl)
-      else CAppExpl ((pf, qid, us), List.map (add_args id new_args) exprl)
-    | CApp ((pf, b), bl) ->
+          ((qid, us), new_args @ List.map (add_args id new_args) exprl)
+      else CAppExpl ((qid, us), List.map (add_args id new_args) exprl)
+    | CApp (b, bl) ->
       CApp
-        ( (pf, add_args id new_args b)
+        ( add_args id new_args b
         , List.map (fun (e, o) -> (add_args id new_args e, o)) bl )
+    | CProj (expl, f, bl, b) ->
+      CProj
+        (expl, f
+        , List.map (fun (e, o) -> (add_args id new_args e, o)) bl
+        , add_args id new_args b)
     | CCases (sty, b_option, cel, cal) ->
       CCases
         ( sty
@@ -2022,10 +2024,8 @@ let rec add_args id new_args =
         , add_args id new_args b2
         , add_args id new_args b3 )
     | (CHole _ | CPatVar _ | CEvar _ | CPrim _ | CSort _) as b -> b
-    | CCast (b1, b2) ->
-      CCast
-        ( add_args id new_args b1
-        , Glob_ops.map_cast_type (add_args id new_args) b2 )
+    | CCast (b1, k, b2) ->
+      CCast (add_args id new_args b1, k, add_args id new_args b2)
     | CRecord pars ->
       CRecord (List.map (fun (e, o) -> (e, add_args id new_args o)) pars)
     | CNotation _ -> CErrors.anomaly ~label:"add_args " (Pp.str "CNotation.")
@@ -2164,8 +2164,8 @@ let build_scheme fas =
         let f_as_constant =
           try Smartlocate.global_with_alias f
           with Not_found ->
-            CErrors.user_err ~hdr:"FunInd.build_scheme"
-              Pp.(str "Cannot find " ++ Libnames.pr_qualid f)
+            CErrors.user_err
+              Pp.(str "Cannot find " ++ Libnames.pr_qualid f ++ str ".")
         in
         let evd', f = Evd.fresh_global (Global.env ()) !evd f_as_constant in
         let _ = evd := evd' in
@@ -2209,8 +2209,8 @@ let build_case_scheme fa =
       | ConstRef c -> c
       | IndRef _ | ConstructRef _ | VarRef _ -> assert false
     with Not_found ->
-      CErrors.user_err ~hdr:"FunInd.build_case_scheme"
-        Pp.(str "Cannot find " ++ Libnames.pr_qualid f)
+      CErrors.user_err
+        Pp.(str "Cannot find " ++ Libnames.pr_qualid f ++ str ".")
   in
   let sigma, (_, u) = Evd.fresh_constant_instance env sigma funs in
   let first_fun = funs in
