@@ -23,7 +23,7 @@ open Nameops
 open CErrors
 open Util
 open UnivGen
-open Tacticals.New
+open Tacticals
 open Tactics
 open Nametab
 open Tacred
@@ -42,10 +42,10 @@ open Context.Rel.Declaration
 (* Ugly things which should not be here *)
 
 let coq_constant s =
-  EConstr.of_constr @@ UnivGen.constr_of_monomorphic_global @@ Coqlib.lib_ref s
+  EConstr.of_constr @@ UnivGen.constr_of_monomorphic_global (Global.env ()) @@ Coqlib.lib_ref s
 
 let coq_init_constant s =
-  EConstr.of_constr (UnivGen.constr_of_monomorphic_global @@ Coqlib.lib_ref s)
+  EConstr.of_constr (UnivGen.constr_of_monomorphic_global (Global.env ()) @@ Coqlib.lib_ref s)
 
 let find_reference sl s =
   let dp = Names.DirPath.make (List.rev_map Id.of_string sl) in
@@ -85,7 +85,7 @@ let type_of_const sigma t =
     Typeops.type_of_constant_in (Global.env ()) (sp, u)
   | _ -> assert false
 
-let constant sl s = UnivGen.constr_of_monomorphic_global (find_reference sl s)
+let constant sl s = UnivGen.constr_of_monomorphic_global (Global.env ()) (find_reference sl s)
 
 let const_of_ref = function
   | GlobRef.ConstRef kn -> kn
@@ -93,7 +93,7 @@ let const_of_ref = function
 
 (* Generic values *)
 let pf_get_new_ids idl g =
-  let ids = Tacmach.New.pf_ids_of_hyps g in
+  let ids = Tacmach.pf_ids_of_hyps g in
   let ids = Id.Set.of_list ids in
   List.fold_right
     (fun id acc ->
@@ -101,12 +101,16 @@ let pf_get_new_ids idl g =
     idl []
 
 let next_ident_away_in_goal ids avoid =
-  next_ident_away_in_goal ids (Id.Set.of_list avoid)
+  next_ident_away_in_goal (Global.env ()) ids (Id.Set.of_list avoid)
 
 let compute_renamed_type gls id =
-  rename_bound_vars_as_displayed (Proofview.Goal.sigma gls)
-    (*no avoid*) Id.Set.empty (*no rels*) []
-    (Tacmach.New.pf_get_hyp_typ id gls)
+  let t = Tacmach.pf_get_hyp_typ id gls in
+  if Clenv.rename_with () then
+    rename_bound_vars_as_displayed (Global.env ()) (Proofview.Goal.sigma gls)
+      (*no avoid*) Id.Set.empty (*no rels*) []
+      t
+  else
+    t
 
 let h'_id = Id.of_string "h'"
 let teq_id = Id.of_string "teq"
@@ -124,10 +128,10 @@ let nat = function () -> coq_init_constant "num.nat.type"
 
 let iter_ref () =
   try find_reference ["Recdef"] "iter"
-  with Not_found -> user_err Pp.(str "module Recdef not loaded")
+  with Not_found -> user_err Pp.(str "Module Recdef not loaded.")
 
 let iter_rd = function
-  | () -> constr_of_monomorphic_global (delayed_force iter_ref)
+  | () -> constr_of_monomorphic_global (Global.env ()) (delayed_force iter_ref)
 
 let eq = function () -> coq_init_constant "core.eq.type"
 let le_lt_SS = function () -> constant ["Recdef"] "le_lt_SS"
@@ -147,7 +151,7 @@ let max_ref = function () -> find_reference ["Recdef"] "max"
 
 let max_constr = function
   | () ->
-    EConstr.of_constr (constr_of_monomorphic_global (delayed_force max_ref))
+    EConstr.of_constr (constr_of_monomorphic_global (Global.env ()) (delayed_force max_ref))
 
 let f_S t = mkApp (delayed_force coq_S, [|t|])
 
@@ -219,7 +223,7 @@ let (declare_f :
   declare_fun f_id kind (value_f input_type fterm_ref)
 
 module New = struct
-  open Tacticals.New
+  open Tacticals
 
   let observe_tac = New.observe_tac ~header:(Pp.mt ())
 
@@ -243,7 +247,7 @@ end
 (* The boolean value is_mes expresses that the termination is expressed
   using a measure function instead of a well-founded relation. *)
 let tclUSER tac is_mes l =
-  let open Tacticals.New in
+  let open Tacticals in
   let clear_tac =
     match l with
     | None -> tclIDTAC
@@ -264,7 +268,7 @@ let tclUSER tac is_mes l =
 
 let tclUSER_if_not_mes concl_tac is_mes names_to_suppress =
   if is_mes then
-    Tacticals.New.tclCOMPLETE (Simple.apply (delayed_force well_founded_ltof))
+    Tacticals.tclCOMPLETE (Simple.apply (delayed_force well_founded_ltof))
   else
     (* tclTHEN (Simple.apply (delayed_force acc_intro_generator_function) ) *)
     tclUSER concl_tac is_mes names_to_suppress
@@ -284,8 +288,8 @@ let check_not_nested env sigma forbidden e =
     | Int _ | Float _ -> ()
     | Var x ->
       if Id.List.mem x forbidden then
-        user_err ~hdr:"Recdef.check_not_nested"
-          (str "check_not_nested: failure " ++ Id.print x)
+        user_err
+          (str "check_not_nested: failure " ++ Id.print x ++ str ".")
     | Meta _ | Evar _ | Sort _ -> ()
     | Cast (e, _, t) -> check_not_nested e; check_not_nested t
     | Prod (_, t, b) -> check_not_nested t; check_not_nested b
@@ -310,9 +314,9 @@ let check_not_nested env sigma forbidden e =
     | CoFix _ -> user_err Pp.(str "check_not_nested : Fix")
   in
   try check_not_nested e
-  with UserError (_, p) ->
-    user_err ~hdr:"_"
-      (str "on expr : " ++ Printer.pr_leconstr_env env sigma e ++ str " " ++ p)
+  with UserError p ->
+    user_err
+      (str "on expr : " ++ Printer.pr_leconstr_env env sigma e ++ str " " ++ p ++ str ".")
 
 (* ['a info] contains the local information for traveling *)
 type 'a infos =
@@ -405,14 +409,14 @@ let treat_case forbid_new_ids to_intros finalize_tac nb_lam e infos :
         (fun _ _ -> str "treat_case1")
         [ h_intros (List.rev rev_ids)
         ; intro_using_then teq_id (fun _ -> Proofview.tclUNIT ())
-        ; Tacticals.New.onLastHypId (fun heq ->
+        ; Tacticals.onLastHypId (fun heq ->
               New.observe_tclTHENLIST
                 (fun _ _ -> str "treat_case2")
                 [ clear to_intros
                 ; h_intros to_intros
                 ; Proofview.Goal.enter (fun g' ->
                       let sigma = Proofview.Goal.sigma g' in
-                      let ty_teq = Tacmach.New.pf_get_hyp_typ heq g' in
+                      let ty_teq = Tacmach.pf_get_hyp_typ heq g' in
                       let teq_lhs, teq_rhs =
                         let _, args =
                           try destApp sigma ty_teq with DestKO -> assert false
@@ -456,7 +460,7 @@ let rec travel_aux jinfo continuation_tac (expr_info : constr infos) =
             expr_info.info;
           jinfo.otherS () expr_info continuation_tac expr_info
         with e when CErrors.noncritical e ->
-          user_err ~hdr:"Recdef.travel"
+          user_err
             ( str "the term "
             ++ Printer.pr_leconstr_env env sigma expr_info.info
             ++ str " can not contain a recursive call to "
@@ -468,11 +472,11 @@ let rec travel_aux jinfo continuation_tac (expr_info : constr infos) =
             expr_info.info;
           jinfo.otherS () expr_info continuation_tac expr_info
         with e when CErrors.noncritical e ->
-          user_err ~hdr:"Recdef.travel"
+          user_err
             ( str "the term "
             ++ Printer.pr_leconstr_env env sigma expr_info.info
             ++ str " can not contain a recursive call to "
-            ++ Id.print expr_info.f_id ) )
+            ++ Id.print expr_info.f_id ++ str ".") )
       | Case (ci, u, pms, t, iv, a, l) ->
         let (ci, t, iv, a, l) = EConstr.expand_case env sigma (ci, u, pms, t, iv, a, l) in
         let continuation_tac_a =
@@ -496,12 +500,12 @@ let rec travel_aux jinfo continuation_tac (expr_info : constr infos) =
             travel_args jinfo expr_info.is_main_branch new_continuation_tac
               new_infos
           | Case _ ->
-            user_err ~hdr:"Recdef.travel"
+            user_err
               ( str "the term "
               ++ Printer.pr_leconstr_env env sigma expr_info.info
               ++ str
                    " can not contain an applied match (See Limitation in \
-                    Section 2.3 of refman)" )
+                    Section 2.3 of refman)." )
           | _ ->
             anomaly
               ( Pp.str "travel_aux : unexpected "
@@ -545,7 +549,7 @@ let rec prove_lt hyple =
         let h =
           List.find
             (fun id ->
-              match decompose_app sigma (Tacmach.New.pf_get_hyp_typ id g) with
+              match decompose_app sigma (Tacmach.pf_get_hyp_typ id g) with
               | _, t :: _ -> EConstr.eq_constr sigma t varx
               | _ -> false)
             hyple
@@ -553,7 +557,7 @@ let rec prove_lt hyple =
         let y =
           List.hd
             (List.tl
-               (snd (decompose_app sigma (Tacmach.New.pf_get_hyp_typ h g))))
+               (snd (decompose_app sigma (Tacmach.pf_get_hyp_typ h g))))
         in
         New.observe_tclTHENLIST
           (fun _ _ -> str "prove_lt1")
@@ -570,11 +574,11 @@ let rec prove_lt hyple =
               assumption ])
 
 let rec destruct_bounds_aux infos (bound, hyple, rechyps) lbounds =
-  let open Tacticals.New in
+  let open Tacticals in
   Proofview.Goal.enter (fun g ->
       match lbounds with
       | [] ->
-        let ids = Tacmach.New.pf_ids_of_hyps g in
+        let ids = Tacmach.pf_ids_of_hyps g in
         let s_max = mkApp (delayed_force coq_S, [|bound|]) in
         let k = next_ident_away_in_goal k_id ids in
         let ids = k :: ids in
@@ -706,7 +710,7 @@ let terminate_letin (na, b, t, e) expr_info continuation_tac info =
       continuation_tac {info with info = new_e; forbidden_ids = new_forbidden})
 
 let pf_type c tac =
-  let open Tacticals.New in
+  let open Tacticals in
   Proofview.Goal.enter (fun gl ->
       let env = Proofview.Goal.env gl in
       let sigma = Proofview.Goal.sigma gl in
@@ -763,7 +767,7 @@ let mkDestructEq not_on_hyp env sigma expr =
 
 let terminate_case next_step (ci, a, iv, t, l) expr_info continuation_tac infos
     =
-  let open Tacticals.New in
+  let open Tacticals in
   Proofview.Goal.enter (fun g ->
       let sigma = Proofview.Goal.sigma g in
       let env = Proofview.Goal.env g in
@@ -792,31 +796,20 @@ let terminate_case next_step (ci, a, iv, t, l) expr_info continuation_tac infos
           ++ int (Array.length l)
           ++ str ")" ++ spc ()
           ++ Printer.pr_leconstr_env env sigma a')
-        ( try
-            tclTHENS destruct_tac
-              (List.map_i
-                 (fun i e ->
-                   New.observe_tac
-                     (fun _ _ -> str "do treat case")
-                     (treat_case f_is_present to_thin_intro
-                        (next_step continuation_tac)
-                        ci.ci_cstr_ndecls.(i) e new_info))
-                 0 (Array.to_list l))
-          with
-          | UserError (Some "Refiner.thensn_tac3", _)
-           |UserError (Some "Refiner.tclFAIL_s", _)
-          ->
-            New.observe_tac
-              (fun _ _ ->
-                str "is computable "
-                ++ Printer.pr_leconstr_env env sigma new_info.info)
-              (next_step continuation_tac
-                 { new_info with
-                   info = Reductionops.nf_betaiotazeta env sigma new_info.info
-                 }) ))
+        ((* We used to try-catch the error from tclTHENS before the
+            port to the new engine, not sure if it's worth putting it back *)
+          tclTHENS destruct_tac
+            (List.map_i
+               (fun i e ->
+                  New.observe_tac
+                    (fun _ _ -> str "do treat case")
+                    (treat_case f_is_present to_thin_intro
+                       (next_step continuation_tac)
+                       ci.ci_cstr_ndecls.(i) e new_info))
+               0 (Array.to_list l))))
 
 let terminate_app_rec (f, args) expr_info continuation_tac _ =
-  let open Tacticals.New in
+  let open Tacticals in
   Proofview.Goal.enter (fun g ->
       let sigma = Proofview.Goal.sigma g in
       let env = Proofview.Goal.env g in
@@ -918,7 +911,7 @@ let equation_case next_step case expr_info continuation_tac infos =
     (terminate_case next_step case expr_info continuation_tac infos)
 
 let rec prove_le () =
-  let open Tacticals.New in
+  let open Tacticals in
   Proofview.Goal.enter (fun g ->
       let sigma = Proofview.Goal.sigma g in
       let x, z =
@@ -941,7 +934,7 @@ let rec prove_le () =
               let h, t =
                 List.find
                   (fun (_, t) -> matching_fun t)
-                  (Tacmach.New.pf_hyps_types g)
+                  (Tacmach.pf_hyps_types g)
               in
               let y =
                 let _, args = decompose_app sigma t in
@@ -953,13 +946,13 @@ let rec prove_le () =
                 ; New.observe_tac
                     (fun _ _ -> str "prove_le (rec)")
                     (prove_le ()) ]
-            with Not_found -> Tacticals.New.tclFAIL 0 (mt ())
+            with Not_found -> Tacticals.tclFAIL 0 (mt ())
           end ])
 
 let rec make_rewrite_list expr_info max = function
   | [] -> Proofview.tclUNIT ()
   | (_, p, hp) :: l ->
-    let open Tacticals.New in
+    let open Tacticals in
     New.observe_tac
       (fun _ _ -> str "make_rewrite_list")
       (tclTHENS
@@ -979,8 +972,8 @@ let rec make_rewrite_list expr_info max = function
                    (* dep proofs also: *) ~dep:true ~with_evars:false
                    ( mkVar hp
                    , ExplicitBindings
-                       [ CAst.make @@ (NamedHyp def, expr_info.f_constr)
-                       ; CAst.make @@ (NamedHyp k, f_S max) ] )
+                       [ CAst.make @@ (NamedHyp (CAst.make def), expr_info.f_constr)
+                       ; CAst.make @@ (NamedHyp (CAst.make k), f_S max) ] )
                    )))
          [ make_rewrite_list expr_info max l
          ; New.observe_tclTHENLIST
@@ -990,7 +983,7 @@ let rec make_rewrite_list expr_info max = function
              ; New.observe_tac (fun _ _ -> str "prove_le(2)") (prove_le ()) ] ])
 
 let make_rewrite expr_info l hp max =
-  let open Tacticals.New in
+  let open Tacticals in
   tclTHENFIRST
     (New.observe_tac
        (fun _ _ -> str "make_rewrite")
@@ -1014,8 +1007,8 @@ let make_rewrite expr_info l hp max =
                     (* dep proofs also: *) ~dep:true ~with_evars:false
                     ( mkVar hp
                     , ExplicitBindings
-                        [ CAst.make @@ (NamedHyp def, expr_info.f_constr)
-                        ; CAst.make @@ (NamedHyp k, f_S (f_S max)) ] )
+                        [ CAst.make @@ (NamedHyp (CAst.make def), expr_info.f_constr)
+                        ; CAst.make @@ (NamedHyp (CAst.make k), f_S (f_S max)) ] )
                     )))
           [ New.observe_tac
               (fun _ _ -> str "make_rewrite finalize")
@@ -1044,7 +1037,7 @@ let rec compute_max rew_tac max l =
   match l with
   | [] -> rew_tac max
   | (_, p, _) :: l ->
-    let open Tacticals.New in
+    let open Tacticals in
     New.observe_tclTHENLIST
       (fun _ _ -> str "compute_max")
       [ simplest_elim (mkApp (delayed_force max_constr, [|max; mkVar p|]))
@@ -1055,7 +1048,7 @@ let rec compute_max rew_tac max l =
             | _ -> assert false) ]
 
 let rec destruct_hex expr_info acc l =
-  let open Tacticals.New in
+  let open Tacticals in
   match l with
   | [] -> (
     match List.rev acc with
@@ -1079,7 +1072,7 @@ let rec destruct_hex expr_info acc l =
                   (destruct_hex expr_info ((v, p, hp) :: acc) l))) ]
 
 let rec intros_values_eq expr_info acc =
-  let open Tacticals.New in
+  let open Tacticals in
   tclORELSE
     (New.observe_tclTHENLIST
        (fun _ _ -> str "intros_values_eq")
@@ -1090,7 +1083,7 @@ let rec intros_values_eq expr_info acc =
     (tclCOMPLETE (destruct_hex expr_info [] acc))
 
 let equation_others _ expr_info continuation_tac infos =
-  let open Tacticals.New in
+  let open Tacticals in
   if expr_info.is_final && expr_info.is_main_branch then
     New.observe_tac
       (fun env sigma ->
@@ -1171,7 +1164,7 @@ let compute_terminate_type nb_args func =
   let open Constr in
   let open CVars in
   let _, a_arrow_b, _ =
-    destLambda (def_of_const (constr_of_monomorphic_global func))
+    destLambda (def_of_const (constr_of_monomorphic_global (Global.env ()) func))
   in
   let rev_args, b = decompose_prod_n nb_args a_arrow_b in
   let left =
@@ -1179,7 +1172,7 @@ let compute_terminate_type nb_args func =
       ( delayed_force iter_rd
       , Array.of_list
           ( lift 5 a_arrow_b :: mkRel 3
-          :: constr_of_monomorphic_global func
+          :: constr_of_monomorphic_global (Global.env ()) func
           :: mkRel 1
           :: List.rev (List.map_i (fun i _ -> mkRel (6 + i)) 0 rev_args) ) )
   in
@@ -1204,14 +1197,14 @@ let compute_terminate_type nb_args func =
   in
   let value =
     mkApp
-      ( constr_of_monomorphic_global (Util.delayed_force coq_sig_ref)
+      ( constr_of_monomorphic_global (Global.env ()) (Util.delayed_force coq_sig_ref)
       , [|b; mkLambda (make_annot (Name v_id) Sorts.Relevant, b, nb_iter)|] )
   in
   compose_prod rev_args value
 
 let termination_proof_header is_mes input_type ids args_id relation rec_arg_num
     rec_arg_id tac wf_tac : unit Proofview.tactic =
-  let open Tacticals.New in
+  let open Tacticals in
   Proofview.Goal.enter (fun g ->
       let nargs = List.length args_id in
       let pre_rec_args =
@@ -1286,7 +1279,7 @@ let whole_start concl_tac nb_args is_mes func input_type relation rec_arg_num :
       let sigma = Proofview.Goal.sigma g in
       let hyps = Proofview.Goal.hyps g in
       let ids = Termops.ids_of_named_context hyps in
-      let func_body = def_of_const (constr_of_monomorphic_global func) in
+      let func_body = def_of_const (constr_of_monomorphic_global (Global.env ()) func) in
       let func_body = EConstr.of_constr func_body in
       let f_name, _, body1 = destLambda sigma func_body in
       let f_id =
@@ -1337,16 +1330,33 @@ let whole_start concl_tac nb_args is_mes func input_type relation rec_arg_num :
             ; args_assoc = [] })
         (fun b ids -> tclUSER_if_not_mes concl_tac b ids))
 
+(* Goal represented as a type, doesn't take into account section variables *)
+let abstract_type sigma gl =
+  let open EConstr in
+  let genv = Global.env () in
+  let evi = Evd.find sigma gl in
+  let env = Evd.evar_filtered_env genv evi in
+  let is_proof_var decl =
+    try ignore (Environ.lookup_named (Context.Named.Declaration.get_id decl) genv); false
+    with Not_found -> true in
+  Environ.fold_named_context_reverse (fun t decl ->
+                                        if is_proof_var decl then
+                                          let decl = Termops.map_named_decl EConstr.of_constr decl in
+                                          mkNamedProd_or_LetIn decl t
+                                        else
+                                          t
+                                      ) ~init:(Evd.evar_concl evi) env
+
 let get_current_subgoals_types pstate =
   let p = Declare.Proof.get pstate in
   let Proof.{goals = sgs; sigma; _} = Proof.data p in
-  (sigma, List.map (Goal.V82.abstract_type sigma) sgs)
+  (sigma, List.map (abstract_type sigma) sgs)
 
 exception EmptySubgoals
 
 let build_and_l sigma l =
   let and_constr =
-    UnivGen.constr_of_monomorphic_global @@ Coqlib.lib_ref "core.and.type"
+    UnivGen.constr_of_monomorphic_global (Global.env ()) @@ Coqlib.lib_ref "core.and.type"
   in
   let conj_constr = Coqlib.lib_ref "core.and.conj" in
   let mk_and p1 p2 = mkApp (EConstr.of_constr and_constr, [|p1; p2|]) in
@@ -1370,7 +1380,7 @@ let build_and_l sigma l =
       let c, tac, nb = f pl in
       ( mk_and p1 c
       , tclTHENS
-          (apply (EConstr.of_constr (constr_of_monomorphic_global conj_constr)))
+          (apply (EConstr.of_constr (constr_of_monomorphic_global (Global.env ()) conj_constr)))
           [tclIDTAC; tac]
       , nb + 1 )
   in
@@ -1443,8 +1453,8 @@ let open_new_goal ~lemma build_proof sigma using_lemmas ref_ goal_name
     let h_num = ref (-1) in
     let env = Global.env () in
     let start_tac =
-      let open Tacmach.New in
-      let open Tacticals.New in
+      let open Tacmach in
+      let open Tacticals in
       Proofview.Goal.enter (fun gl ->
           let hid = next_ident_away_in_goal h_id (pf_ids_of_hyps gl) in
           New.observe_tclTHENLIST
@@ -1462,8 +1472,8 @@ let open_new_goal ~lemma build_proof sigma using_lemmas ref_ goal_name
                          tclIDTAC))) ])
     in
     let end_tac =
-      let open Tacmach.New in
-      let open Tacticals.New in
+      let open Tacmach in
+      let open Tacticals in
       Proofview.Goal.enter (fun gl ->
           let sigma = project gl in
           match EConstr.kind sigma (pf_concl gl) with
@@ -1501,12 +1511,12 @@ let open_new_goal ~lemma build_proof sigma using_lemmas ref_ goal_name
                  (tclFIRST
                     (List.map
                        (fun c ->
-                         Tacticals.New.tclTHENLIST
+                         Tacticals.tclTHENLIST
                            [ intros
                            ; Simple.apply
                                (fst (interp_constr (Global.env ()) Evd.empty c))
                              (*FIXME*)
-                           ; Tacticals.New.tclCOMPLETE Auto.default_auto ])
+                           ; Tacticals.tclCOMPLETE Auto.default_auto ])
                        using_lemmas))
                  tclIDTAC))
            lemma
@@ -1542,7 +1552,7 @@ let com_terminate interactive_proof tcc_lemma_name tcc_lemma_ref is_mes
   let lemma =
     start_proof
       Global.(env ())
-      ctx Tacticals.New.tclIDTAC Tacticals.New.tclIDTAC
+      ctx Tacticals.tclIDTAC Tacticals.tclIDTAC
   in
   try
     let sigma, new_goal_type = build_new_goal_type lemma in
@@ -1558,8 +1568,8 @@ let start_equation (f : GlobRef.t) (term_f : GlobRef.t)
     (cont_tactic : Id.t list -> unit Proofview.tactic) =
   Proofview.Goal.enter (fun g ->
       let sigma = Proofview.Goal.sigma g in
-      let ids = Tacmach.New.pf_ids_of_hyps g in
-      let terminate_constr = constr_of_monomorphic_global term_f in
+      let ids = Tacmach.pf_ids_of_hyps g in
+      let terminate_constr = constr_of_monomorphic_global (Global.env ()) term_f in
       let terminate_constr = EConstr.of_constr terminate_constr in
       let nargs =
         nb_prod sigma (EConstr.of_constr (type_of_const sigma terminate_constr))
@@ -1587,7 +1597,7 @@ let com_eqn uctx nb_arg eq_name functional_ref f_ref terminate_ref
     | _ -> anomaly ~label:"terminate_lemma" (Pp.str "not a constant.")
   in
   let evd = Evd.from_ctx uctx in
-  let f_constr = constr_of_monomorphic_global f_ref in
+  let f_constr = constr_of_monomorphic_global (Global.env ()) f_ref in
   let equation_lemma_type = subst1 f_constr equation_lemma_type in
   let info = Declare.Info.make () in
   let cinfo =
@@ -1605,15 +1615,15 @@ let com_eqn uctx nb_arg eq_name functional_ref f_ref terminate_ref
                 { nb_arg
                 ; f_terminate =
                     EConstr.of_constr
-                      (constr_of_monomorphic_global terminate_ref)
+                      (constr_of_monomorphic_global (Global.env ()) terminate_ref)
                 ; f_constr = EConstr.of_constr f_constr
-                ; concl_tac = Tacticals.New.tclIDTAC
+                ; concl_tac = Tacticals.tclIDTAC
                 ; func = functional_ref
                 ; info =
                     instantiate_lambda Evd.empty
                       (EConstr.of_constr
                          (def_of_const
-                            (constr_of_monomorphic_global functional_ref)))
+                            (constr_of_monomorphic_global (Global.env ()) functional_ref)))
                       (EConstr.of_constr f_constr :: List.map mkVar x)
                 ; is_main_branch = true
                 ; is_final = true
@@ -1740,9 +1750,9 @@ let recursive_definition ~interactive_proof ~is_mes function_name rec_impls
       with e when CErrors.noncritical e ->
         if do_observe () then
           Feedback.msg_debug
-            (str "Cannot create equation Lemma " ++ CErrors.print e)
+            (str "Cannot create equation Lemma " ++ CErrors.print e ++ str ".")
         else
-          CErrors.user_err ~hdr:"Cannot create equation Lemma"
+          CErrors.user_err
             ( str "Cannot create equation lemma."
             ++ spc ()
             ++ str "This may be because the function is nested-recursive." );
@@ -1750,10 +1760,10 @@ let recursive_definition ~interactive_proof ~is_mes function_name rec_impls
     in
     if not stop then
       let eq_ref = Nametab.locate (qualid_of_ident equation_id) in
-      let f_ref = destConst (constr_of_monomorphic_global f_ref)
+      let f_ref = destConst (constr_of_monomorphic_global (Global.env ()) f_ref)
       and functional_ref =
-        destConst (constr_of_monomorphic_global functional_ref)
-      and eq_ref = destConst (constr_of_monomorphic_global eq_ref) in
+        destConst (constr_of_monomorphic_global (Global.env ()) functional_ref)
+      and eq_ref = destConst (constr_of_monomorphic_global (Global.env ()) eq_ref) in
       generate_induction_principle f_ref tcc_lemma_constr functional_ref eq_ref
         rec_arg_num
         (EConstr.of_constr rec_arg_type)
