@@ -406,7 +406,7 @@ let intern_typed_pattern_or_ref_with_occurrences ist (l,p) =
           Inr (bound_names,(c,None),dummy_pat) in
   (l, match p with
   | Inl r -> interp_ref r
-  | Inr { v = CAppExpl((None,r,None),[]) } ->
+  | Inr { v = CAppExpl((r,None),[]) } ->
       (* We interpret similarly @ref and ref *)
       interp_ref (make @@ AN r)
   | Inr c ->
@@ -505,9 +505,11 @@ let rec intern_match_goal_hyps ist ?(as_type=false) lfun = function
 let extract_let_names lrc =
   let fold accu ({loc;v=name}, _) =
     Nameops.Name.fold_right (fun id accu ->
-    if Id.Set.mem id accu then user_err ?loc
-      ~hdr:"glob_tactic" (str "This variable is bound several times.")
-    else Id.Set.add id accu) name accu
+        if Id.Set.mem id accu then
+          user_err ?loc (str "This variable is bound several times.")
+        else Id.Set.add id accu)
+      name
+      accu
   in
   List.fold_left fold Id.Set.empty lrc
 
@@ -621,7 +623,6 @@ and intern_tactic_seq onlytac ist tac =
   | TacFail (g,n,l) ->
       ist.ltacvars, CAst.make ?loc (TacFail (g,intern_int_or_var ist n,intern_message ist l))
   | TacProgress tac -> ist.ltacvars, CAst.make ?loc (TacProgress (intern_pure_tactic ist tac))
-  | TacShowHyps tac -> ist.ltacvars, CAst.make ?loc (TacShowHyps (intern_pure_tactic ist tac))
   | TacAbstract (tac,s) ->
       ist.ltacvars, CAst.make ?loc (TacAbstract (intern_pure_tactic ist tac,s))
   | TacThen (t1,t2) ->
