@@ -27,7 +27,6 @@ open Tacinterp
 open Libobject
 open Printer
 open Declare
-open Entries
 open Ring_ast
 open Proofview.Notations
 
@@ -71,13 +70,12 @@ let add_map s m = protect_maps := String.Map.add s m !protect_maps
 let lookup_map map =
   try String.Map.find map !protect_maps
   with Not_found ->
-    CErrors.user_err ~hdr:"lookup_map" (str"Map "++qs map++str"not found")
+    CErrors.user_err (str "Map " ++ qs map ++ str "not found.")
 
 let protect_red map env sigma c0 =
-  let evars ev = Evarutil.safe_evar_value sigma ev in
   let c = EConstr.Unsafe.to_constr c0 in
   let tab = create_tab () in
-  let infos = create_clos_infos ~univs:(Evd.universes sigma) ~evars all env in
+  let infos = Evarutil.create_clos_infos env sigma all in
   let map = lookup_map map sigma c0 in
   let rec eval n c = match Constr.kind c with
   | Prod (na, t, u) -> Constr.mkProd (na, eval n t, eval (n + 1) u)
@@ -110,7 +108,7 @@ let closed_term args _ = match args with
   let l = List.map (fun c -> Value.cast (Genarg.topwit Stdarg.wit_ref) c) (Option.get (Value.to_list l)) in
   Proofview.tclEVARMAP >>= fun sigma ->
   let cs = List.fold_right GlobRef.Set_env.add l GlobRef.Set_env.empty in
-  if closed_under sigma cs t then Proofview.tclUNIT () else Tacticals.New.tclFAIL 0 (mt())
+  if closed_under sigma cs t then Proofview.tclUNIT () else Tacticals.tclFAIL 0 (mt())
 | _ -> assert false
 
 let closed_term_ast =
@@ -149,7 +147,7 @@ let decl_constant name univs c =
   let univs = UState.restrict_universe_context ~lbound:(Global.universes_lbound ()) univs vars in
   let () = DeclareUctx.declare_universe_context ~poly:false univs in
   let types = (Typeops.infer (Global.env ()) c).uj_type in
-  let univs = Monomorphic_entry Univ.ContextSet.empty in
+  let univs = UState.Monomorphic_entry Univ.ContextSet.empty, UnivNames.empty_binders in
   mkConst(declare_constant ~name
             ~kind:Decls.(IsProof Lemma)
             (DefinitionEntry (definition_entry ~opaque:true ~types ~univs c)))
@@ -162,11 +160,6 @@ let decl_constant na suff univs c =
 let ltac_call tac (args:glob_tactic_arg list) =
   CAst.make @@ TacArg (TacCall (CAst.make (ArgArg(Loc.tag @@ Lazy.force tac),args)))
 
-let dummy_goal env sigma =
-  let (gl,_,sigma) =
-    Goal.V82.mk_goal sigma (named_context_val env) EConstr.mkProp in
-  {Evd.it = gl; Evd.sigma = sigma}
-
 let constr_of sigma v = match Value.to_constr v with
   | Some c -> EConstr.to_constr sigma c
   | None -> failwith "Ring.exec_tactic: anomaly"
@@ -200,13 +193,14 @@ let exec_tactic env sigma n f args =
   let get_res = CAst.make (TacML (get_res, [TacGeneric (None, n)])) in
   let getter = Tacexp (CAst.make (TacFun (List.map (fun n -> Name n) lid, get_res))) in
   (* Evaluate the whole result *)
-  let gl = dummy_goal env sigma in
-  let gls = Proofview.V82.of_tactic (Tacinterp.eval_tactic_ist ist (ltac_call f (args@[getter]))) gl in
-  let sigma = Evd.minimize_universes gls.Evd.sigma in
+  let _, pv = Proofview.init sigma [env, EConstr.mkProp] in
+  let tac = Tacinterp.eval_tactic_ist ist (ltac_call f (args@[getter])) in
+  let ((), pv, _, _) = Proofview.apply ~name:(Id.of_string "ring") ~poly:false (Global.env ()) tac pv in
+  let sigma = Evd.minimize_universes (Proofview.return pv) in
   let nf c = constr_of sigma c in
   Array.map nf !tactic_res, Evd.universe_context_set sigma
 
-let gen_constant n = lazy (EConstr.of_constr (UnivGen.constr_of_monomorphic_global (Coqlib.lib_ref n)))
+let gen_constant n = lazy (EConstr.of_constr (UnivGen.constr_of_monomorphic_global (Global.env ()) (Coqlib.lib_ref n)))
 let gen_reference n = lazy (Coqlib.lib_ref n)
 
 let coq_mk_Setoid = gen_constant "plugins.ring.Build_Setoid_Theory"
@@ -220,7 +214,7 @@ let coq_nil = gen_reference "core.list.nil"
 let lapp f args = mkApp(Lazy.force f,args)
 
 let plapp sigma f args =
-  let sigma, fc = Evarutil.new_global sigma (Lazy.force f) in
+  let sigma, fc = Evd.fresh_global (Global.env ()) sigma (Lazy.force f) in
   sigma, mkApp(fc,args)
 
 let dest_rel0 sigma t =
@@ -250,7 +244,7 @@ let plugin_modules =
     ]
 
 let my_constant c =
-  lazy (EConstr.of_constr (UnivGen.constr_of_monomorphic_global @@ Coqlib.gen_reference_in_modules "Ring" plugin_modules c))
+  lazy (EConstr.of_constr (UnivGen.constr_of_monomorphic_global (Global.env ()) @@ Coqlib.gen_reference_in_modules "Ring" plugin_modules c))
     [@@ocaml.warning "-3"]
 let my_reference c =
   lazy (Coqlib.gen_reference_in_modules "Ring" plugin_modules c)
@@ -346,13 +340,13 @@ let find_ring_structure env sigma l =
         let check c =
           let ty' = Retyping.get_type_of env sigma c in
           if not (Reductionops.is_conv env sigma ty ty') then
-            CErrors.user_err ~hdr:"ring"
+            CErrors.user_err
               (str"Arguments of ring_simplify do not have all the same type.")
         in
         List.iter check cl';
         (try ring_for_carrier (EConstr.to_constr sigma ty)
         with Not_found ->
-          CErrors.user_err ~hdr:"ring"
+          CErrors.user_err
             (str"Cannot find a declared ring structure over"++
              spc() ++ str"\"" ++ pr_econstr_env env sigma ty ++ str"\"."))
     | [] -> assert false
@@ -512,7 +506,7 @@ let make_hyp env sigma c =
   plapp sigma coq_mkhypo [|t;c|]
 
 let make_hyp_list env sigma lH =
-  let sigma, carrier = Evarutil.new_global sigma (Lazy.force coq_hypo) in
+  let sigma, carrier = Evd.fresh_global env sigma (Lazy.force coq_hypo) in
   let sigma, l =
     List.fold_right
       (fun c (sigma,l) ->
@@ -525,7 +519,7 @@ let make_hyp_list env sigma lH =
   sigma, Evarutil.nf_evars_universes sigma l'
 
 let interp_power env sigma pow =
-  let sigma, carrier = Evarutil.new_global sigma (Lazy.force coq_hypo) in
+  let sigma, carrier = Evd.fresh_global env sigma (Lazy.force coq_hypo) in
   match pow with
   | None ->
       let t = ArgArg(Loc.tag (Lazy.force ltac_inv_morph_nothing)) in
@@ -543,7 +537,7 @@ let interp_power env sigma pow =
       sigma, (tac, pow)
 
 let interp_sign env sigma sign =
-  let sigma, carrier = Evarutil.new_global sigma (Lazy.force coq_hypo) in
+  let sigma, carrier = Evd.fresh_global env sigma (Lazy.force coq_hypo) in
   match sign with
   | None -> plapp sigma coq_None [|carrier|]
   | Some spec ->
@@ -552,7 +546,7 @@ let interp_sign env sigma sign =
        (* Same remark on ill-typed terms ... *)
 
 let interp_div env sigma div =
-  let sigma, carrier = Evarutil.new_global sigma (Lazy.force coq_hypo) in
+  let sigma, carrier = Evd.fresh_global env sigma (Lazy.force coq_hypo) in
   match div with
   | None -> plapp sigma coq_None [|carrier|]
   | Some spec ->
@@ -683,7 +677,7 @@ let ltac_ring_structure e =
 
 let ring_lookup (f : Value.t) lH rl t =
   Proofview.Goal.enter begin fun gl ->
-    let sigma = Tacmach.New.project gl in
+    let sigma = Tacmach.project gl in
     let env = Proofview.Goal.env gl in
     let rl = make_args_list sigma rl t in
     let e = find_ring_structure env sigma rl in
@@ -795,13 +789,13 @@ let find_field_structure env sigma l =
         let check c =
           let ty' = Retyping.get_type_of env sigma c in
           if not (Reductionops.is_conv env sigma ty ty') then
-            CErrors.user_err ~hdr:"field"
+            CErrors.user_err
               (str"Arguments of field_simplify do not have all the same type.")
         in
         List.iter check cl';
         (try field_for_carrier (EConstr.to_constr sigma ty)
         with Not_found ->
-          CErrors.user_err ~hdr:"field"
+          CErrors.user_err
             (str"Cannot find a declared field structure over"++
              spc()++str"\""++pr_econstr_env env sigma ty++str"\"."))
     | [] -> assert false
@@ -855,7 +849,7 @@ let ftheory_to_obj : field_info -> obj =
 let field_equality env sigma r inv req =
   match EConstr.kind sigma req with
     | App (f, [| _ |]) when eq_constr_nounivs sigma f (Lazy.force coq_eq) ->
-        let c = UnivGen.constr_of_monomorphic_global Coqlib.(lib_ref "core.eq.congr") in
+        let c = UnivGen.constr_of_monomorphic_global (Global.env ()) Coqlib.(lib_ref "core.eq.congr") in
         let c = EConstr.of_constr c in
         mkApp(c,[|r;r;inv|])
     | _ ->
@@ -976,7 +970,7 @@ let ltac_field_structure e =
 
 let field_lookup (f : Value.t) lH rl t =
   Proofview.Goal.enter begin fun gl ->
-    let sigma = Tacmach.New.project gl in
+    let sigma = Tacmach.project gl in
     let env = Proofview.Goal.env gl in
     let rl = make_args_list sigma rl t in
     let e = find_field_structure env sigma rl in
