@@ -570,17 +570,16 @@ let force_eqs c =
   Set.fold
     (fun c acc ->
        let c' = match c with
-         (* Should we be forcing weak constraints? *)
-         | ULub (l, r) | UWeak (l, r) -> UEq (Univ.Universe.make l,Univ.Universe.make r)
-         | ULe _ | UEq _ -> c
+         | ULub (l, r) -> UEq (Univ.Universe.make l,Univ.Universe.make r)
+         | ULe _ | UEq _ | UWeak _ -> c
        in
         Set.add c' acc)
     c Set.empty
 
-let constr_cmp pb env sigma flags t u =
+let constr_cmp pb env sigma flags ?nargs t u =
   let cstrs =
-    if pb == Reduction.CONV then EConstr.eq_constr_universes env sigma t u
-    else EConstr.leq_constr_universes env sigma t u
+    if pb == Reduction.CONV then EConstr.eq_constr_universes env sigma ?nargs t u
+    else EConstr.leq_constr_universes env sigma ?nargs t u
   in
   match cstrs with
   | Some cstrs ->
@@ -619,8 +618,10 @@ let subst_defined_metas_evars sigma (bl,el) c =
     | Evar (evk,args) ->
       let eq c1 c2 = Constr.equal c1 (EConstr.Unsafe.to_constr c2) in
       let select (_,(evk',args'),_) = Evar.equal evk evk' && List.for_all2 eq args args' in
-      (try substrec (EConstr.Unsafe.to_constr (pi3 (List.find select el)))
-       with Not_found -> Constr.map substrec c)
+      begin match List.find select el with
+      | (_, _, c) -> substrec (EConstr.Unsafe.to_constr c)
+      | exception Not_found -> Constr.map substrec c
+      end
     | _ -> Constr.map substrec c
   in try Some (EConstr.of_constr (substrec c)) with Not_found -> None
 
@@ -708,7 +709,7 @@ let fast_occur_meta_or_undefined_evar sigma (c, gnd) = match gnd with
 | NotGround -> true
 
 let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top env cv_pb flags m n =
-  let rec unirec_rec (curenv,nb as curenvnb) pb opt ((sigma,metasubst,evarsubst) as substn : subst0) curm curn =
+  let rec unirec_rec (curenv,nb as curenvnb) pb opt ((sigma,metasubst,evarsubst) as substn : subst0) ?(nargs=0) curm curn =
     let cM = Evarutil.whd_head_evar sigma curm
     and cN = Evarutil.whd_head_evar sigma curn in
     let () =
@@ -820,7 +821,7 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
           (try unify_same_proj curenvnb cv_pb {opt with at_top = true}
                substn c1 c2
            with ex when precatchable_exception ex ->
-             unify_not_same_head curenvnb pb opt substn cM cN)
+             unify_not_same_head curenvnb pb opt substn ~nargs cM cN)
 
         (* eta-expansion *)
         | Lambda (na,t1,c1), _ when flags.modulo_eta ->
@@ -838,7 +839,7 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
           (try
              let l1', l2' = eta_constructor_app curenv sigma f1 l1 cN in
              let opt' = {opt with at_top = true; with_cs = false} in
-               Array.fold_left2 (unirec_rec curenvnb CONV opt') substn l1' l2'
+               Array.fold_left2 (unirec_rec curenvnb CONV opt' ~nargs:0) substn l1' l2'
            with ex when precatchable_exception ex ->
              match EConstr.kind sigma cN with
              | App(f2,l2) when
@@ -852,7 +853,7 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
           (try
              let l2', l1' = eta_constructor_app curenv sigma f2 l2 cM in
              let opt' = {opt with at_top = true; with_cs = false} in
-               Array.fold_left2 (unirec_rec curenvnb CONV opt') substn l1' l2'
+               Array.fold_left2 (unirec_rec curenvnb CONV opt' ~nargs:0) substn l1' l2'
            with ex when precatchable_exception ex ->
              match EConstr.kind sigma cM with
              | App(f1,l1) when
@@ -867,7 +868,7 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
             (try
              if not (Ind.CanOrd.equal ci1.ci_ind ci2.ci_ind) then error_cannot_unify curenv sigma (cM,cN);
              let opt' = {opt with at_top = true; with_types = false} in
-               Array.fold_left2 (unirec_rec curenvnb CONV {opt with at_top = true})
+               Array.fold_left2 (unirec_rec curenvnb CONV {opt with at_top = true} ~nargs:0)
                (unirec_rec curenvnb CONV opt'
                 (unirec_rec curenvnb CONV opt' substn p1 p2) c1 c2)
                  cl1 cl2
@@ -879,8 +880,8 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
             (try
              let opt' = {opt with at_top = true; with_types = false} in
              let curenvnb' = Array.fold_right2 (fun na t -> push (na,t)) lna1 tl1 curenvnb in
-               Array.fold_left2 (unirec_rec curenvnb' CONV opt')
-               (Array.fold_left2 (unirec_rec curenvnb CONV opt') substn tl1 tl2) bl1 bl2
+               Array.fold_left2 (unirec_rec curenvnb' CONV opt' ~nargs:0)
+               (Array.fold_left2 (unirec_rec curenvnb CONV opt' ~nargs:0) substn tl1 tl2) bl1 bl2
              with ex when precatchable_exception ex ->
                reduce curenvnb pb opt substn cM cN)
 
@@ -889,8 +890,8 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
             (try
              let opt' = {opt with at_top = true; with_types = false} in
              let curenvnb' = Array.fold_right2 (fun na t -> push (na,t)) lna1 tl1 curenvnb in
-               Array.fold_left2 (unirec_rec curenvnb' CONV opt')
-               (Array.fold_left2 (unirec_rec curenvnb CONV opt') substn tl1 tl2) bl1 bl2
+               Array.fold_left2 (unirec_rec curenvnb' CONV opt' ~nargs:0)
+               (Array.fold_left2 (unirec_rec curenvnb CONV opt' ~nargs:0) substn tl1 tl2) bl1 bl2
              with ex when precatchable_exception ex ->
                reduce curenvnb pb opt substn cM cN)
 
@@ -914,7 +915,7 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
           unify_app curenvnb pb opt substn cM cM [||] cN f2 l2
 
         | _ ->
-          unify_not_same_head curenvnb pb opt substn cM cN
+          unify_not_same_head curenvnb pb opt substn ~nargs cM cN
 
   and unify_app_pattern dir curenvnb pb opt (sigma, _, _ as substn) cM f1 l1 cN f2 l2 =
     let f, l, t = if dir then f1, l1, cN else f2, l2, cM in
@@ -925,7 +926,9 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
           if dir then unify_app curenvnb pb opt substn cM f1 l1 t f' l'
           else unify_app curenvnb pb opt substn t f' l' cN f2 l2
         | Proj _ -> unify_app curenvnb pb opt substn cM f1 l1 cN f2 l2
-        | _ -> unify_not_same_head curenvnb pb opt substn cM cN)
+        | _ ->
+          (* XXX nargs could be better? *)
+          unify_not_same_head curenvnb pb opt substn ~nargs:0 cM cN)
       | Some l ->
         solve_pattern_eqn_array curenvnb f l t substn
 
@@ -954,8 +957,8 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
       let (f1,l1,f2,l2) = adjust_app_array_size f1 l1 f2 l2 in
         if Array.length l1 == 0 then error_cannot_unify (fst curenvnb) sigma (cM,cN)
         else
-          Array.fold_left2 (unirec_rec curenvnb CONV opta)
-            (unirec_rec curenvnb CONV optf substn f1 f2) l1 l2
+          Array.fold_left2 (unirec_rec curenvnb CONV opta ~nargs:0)
+            (unirec_rec curenvnb CONV optf substn f1 f2 ~nargs:(Array.length l1)) l1 l2
     with ex when precatchable_exception ex ->
     try reduce curenvnb pb {opt with with_types = false} substn cM cN
     with ex when precatchable_exception ex ->
@@ -976,10 +979,10 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
             (ty1, Unknown) (ty2, Unknown)
       with RetypeError _ -> substn
 
-  and unify_not_same_head curenvnb pb opt (sigma, metas, evars as substn : subst0) cM cN =
+  and unify_not_same_head curenvnb pb opt (sigma, metas, evars as substn : subst0) ~nargs cM cN =
     try canonical_projections curenvnb pb opt cM cN substn
     with ex when precatchable_exception ex ->
-    match constr_cmp cv_pb env sigma flags cM cN with
+    match constr_cmp cv_pb env sigma flags ~nargs cM cN with
     | Some sigma -> (sigma, metas, evars)
     | None ->
         try reduce curenvnb pb opt substn cM cN
@@ -1628,6 +1631,8 @@ let make_pattern_test from_prefix_of_ind is_correct_type env sigma (pending,c) =
   let n = Array.length (snd (decompose_app_vect sigma c)) in
   let cgnd = if occur_meta_or_undefined_evar sigma c then NotGround else Ground in
   let matching_fun _ t =
+    (* make_pattern_test is only ever called with an empty rel context *)
+    if not (EConstr.Vars.closed0 sigma t) then raise (NotUnifiable None);
     try
       let t',l2 =
         if from_prefix_of_ind then
@@ -1684,9 +1689,9 @@ let make_abstraction_core name (test,out) env sigma c ty occs check_occs concl =
     let t = match ty with Some t -> t | None -> get_type_of env sigma c in
     let x = id_of_name_using_hdchar env sigma t name in
     let ids = Environ.ids_of_named_context_val (named_context_val env) in
-    if name == Anonymous then next_ident_away_in_goal x ids else
+    if name == Anonymous then next_ident_away_in_goal env x ids else
     if mem_named_context_val x (named_context_val env) then
-      user_err ~hdr:"Unification.make_abstraction_core"
+      user_err
         (str "The variable " ++ Id.print x ++ str " is already declared.")
     else
       x
@@ -1780,6 +1785,124 @@ let keyed_unify env evd kop =
           | None -> false
           | Some kc -> Keys.equiv_keys kop kc
 
+module AConstr :
+sig
+  type t
+  val proj : t -> EConstr.t
+  val make : evar_map -> EConstr.t -> t
+  val kind : t -> (t, t, ESorts.t, EInstance.t) kind_of_term
+  val mkApp : t * t array -> t
+  val closed0 : t -> bool
+end =
+struct
+
+type t = {
+  proj : EConstr.t;
+  self : (t, t, ESorts.t, EInstance.t) kind_of_term;
+  data : int;
+}
+
+let proj c = c.proj
+
+let closed0 c = Int.equal c.data 0
+
+let max (i : int) (j : int) = if i < j then j else i
+let max_array f a = Array.fold_left (fun n v -> max (f v) n) 0 a
+let lift (i : int) = if Int.equal i 0 then 0 else i - 1
+let liftn k (i : int) = if i < k then 0 else i - k
+
+let data v = v.data
+
+let kind v = v.self
+
+let mkApp (c, al) =
+  if Array.is_empty al then c
+  else match kind c with
+  | App (c0, al0) ->
+    { proj = mkApp (c.proj, Array.map proj al); self = App (c0, Array.append al0 al); data = max c.data (max_array data al) }
+  | _ ->
+    { proj = mkApp (c.proj, Array.map proj al); self = App (c, al); data = max c.data (max_array data al) }
+
+let rec make sigma c0 = match EConstr.kind sigma c0 with
+| (Meta _ | Var _ | Sort _ | Const _ | Ind _ | Construct _ | Int _ | Float _) as r ->
+  { proj = c0; self = r; data = 0 }
+| Rel n ->
+  { proj = c0; self = Rel n; data = n }
+| Cast (c, k, t) ->
+  let c = make sigma c in
+  let t = make sigma t in
+  { proj = c0; self = Cast (c, k, t); data = max c.data t.data }
+| Prod (na, t, c) ->
+  let t = make sigma t in
+  let c = make sigma c in
+  { proj = c0; self = Prod (na, t, c); data = max t.data (lift c.data) }
+| Lambda (na, t, c) ->
+  let t = make sigma t in
+  let c = make sigma c in
+  { proj = c0; self = Lambda (na, t, c); data = max t.data (lift c.data) }
+| LetIn (na, b, t, c) ->
+  let b = make sigma b in
+  let t = make sigma t in
+  let c = make sigma c in
+  { proj = c0; self = LetIn (na, b, t, c); data = max b.data (max t.data (lift c.data)) }
+| App (c, al) ->
+  let c = make sigma c in
+  let ald, al = make_array sigma al in
+  { proj = c0; self = App (c, al); data = max c.data ald }
+| Proj (p, t) ->
+  let t = make sigma t in
+  { proj = c0; self = Proj (p, t); data = t.data }
+| Evar (e, al) ->
+  let al = List.map (fun c -> make sigma c) al in
+  let data = List.fold_left (fun accu v -> max accu v.data) 0 al in
+  { proj = c0; self = Evar (e, al); data }
+| Case (ci, u, pms, p, iv, c, bl) ->
+  let pmsd, pms = make_array sigma pms in
+  let pd, p =
+    let (nas, p) = p in
+    let p = make sigma p in
+    liftn (Array.length nas) p.data, (nas, p)
+  in
+  let ivd, iv = match iv with
+  | NoInvert -> 0, NoInvert
+  | CaseInvert { indices } ->
+    let n, indices = make_array sigma indices in
+    n, CaseInvert { indices }
+  in
+  let c = make sigma c in
+  let fold accu (nas, p) =
+    let p = make sigma p in
+    max accu (liftn (Array.length nas) p.data), (nas, p)
+  in
+  let bld, bl = Array.fold_left_map fold 0 bl in
+  let data = max pmsd @@ max pd @@ max ivd @@ max c.data bld in
+  { proj = c0; self = Case (ci, u, pms, p, iv, c, bl); data }
+| Fix (ln, (lna, tl, bl)) ->
+  let tld, tl = make_array sigma tl in
+  let bld, bl = make_array sigma bl in
+  let data = max tld (liftn (Array.length tl) bld) in
+  { proj = c0; self = Fix (ln, (lna, tl, bl)); data }
+| CoFix(ln,(lna,tl,bl)) ->
+  let tld, tl = make_array sigma tl in
+  let bld, bl = make_array sigma bl in
+  let data = max tld (liftn (Array.length tl) bld) in
+  { proj = c0; self = CoFix (ln, (lna, tl, bl)); data }
+| Array(u,t,def,ty) ->
+  let td, t = make_array sigma t in
+  let def = make sigma def in
+  let ty = make sigma ty in
+  let data = max td (max def.data ty.data) in
+  { proj = c0; self = Array(u,t,def,ty); data }
+
+and make_array sigma v =
+  let fold accu c =
+    let c = make sigma c in
+    max accu c.data, c
+  in
+  Array.fold_left_map fold 0 v
+
+end
+
 (* Tries to find an instance of term [cl] in term [op].
    Unifies [cl] to every subterm of [op] until it finds a match.
    Fails if no match is found *)
@@ -1788,9 +1911,15 @@ let w_unify_to_subterm env evd ?(flags=default_unify_flags ()) (op,cl) =
   let kop = Keys.constr_key (fun c -> EConstr.kind evd c) op in
   let opgnd = if occur_meta_or_undefined_evar evd op then NotGround else Ground in
   let rec matchrec cl =
-    let cl = strip_outer_cast evd cl in
+    let rec strip_outer_cast c = match AConstr.kind c with
+    | Cast (c, _, _) -> strip_outer_cast c
+    | _ -> c
+    in
+    let cl = strip_outer_cast cl in
     (try
-       if closed0 evd cl && not (isEvar evd cl) && keyed_unify env evd kop cl then
+      let is_closed = AConstr.closed0 cl in
+      let cl = AConstr.proj cl in
+       if is_closed && not (isEvar evd cl) && keyed_unify env evd kop cl then
        (try
          if is_keyed_unification () then
            let f1, l1 = decompose_app_vect evd op in
@@ -1801,11 +1930,11 @@ let w_unify_to_subterm env evd ?(flags=default_unify_flags ()) (op,cl) =
             bestexn := Some ex; user_err Pp.(str "Unsat"))
        else user_err Pp.(str "Bound 1")
      with ex when precatchable_exception ex ->
-       (match EConstr.kind evd cl with
+       (match AConstr.kind cl with
           | App (f,args) ->
               let n = Array.length args in
               assert (n>0);
-              let c1 = mkApp (f,Array.sub args 0 (n-1)) in
+              let c1 = AConstr.mkApp (f,Array.sub args 0 (n-1)) in
               let c2 = args.(n-1) in
               (try
                  matchrec c1
@@ -1960,7 +2089,7 @@ let w_unify_to_subterm_list env evd flags hdmeta oplist t =
                unify pre-existing non frozen evars of the goal or of the
                pattern *)
           set_no_delta_flags flags in
-        let t' = (strip_outer_cast evd op,t) in
+        let t' = (strip_outer_cast evd op, AConstr.make evd t) in
         let (evd',cl) =
           try
             if is_keyed_unification () then
@@ -1986,6 +2115,9 @@ let w_unify_to_subterm_list env evd flags hdmeta oplist t =
     oplist
     (evd,[])
 
+let w_unify_to_subterm env sigma ?flags (c, t) =
+  w_unify_to_subterm env sigma ?flags (c, AConstr.make sigma t)
+
 let secondOrderAbstraction env evd flags typ (p, oplist) =
   (* Remove delta when looking for a subterm *)
   let flags = { flags with core_unify_flags = flags.subterm_unify_flags } in
