@@ -72,14 +72,12 @@ let find_rectype_a env sigma c =
 
 let type_constructor mind mib u (ctx, typ) params =
   let typ = it_mkProd_or_LetIn typ ctx in
-  let s = ind_subst mind mib u in
-  let ctyp = substl s typ in
-  let ctyp = subst_instance_constr u ctyp in
+  let ctyp = subst_instance_constr u typ in
   let ndecls = Context.Rel.length mib.mind_params_ctxt in
   if Int.equal ndecls 0 then ctyp
   else
     let _,ctyp = decompose_prod_n_assum ndecls ctyp in
-    substl (subst_of_rel_context_instance mib.mind_params_ctxt (Array.to_list params))
+    substl (subst_of_rel_context_instance mib.mind_params_ctxt params)
       ctyp
 
 
@@ -181,7 +179,7 @@ and nf_whd env sigma whd typ =
   | Vatom_stk(Aind ((mi,i) as ind), stk) ->
      let mib = Environ.lookup_mind mi env in
      let nb_univs =
-       Univ.AUContext.size (Declareops.inductive_polymorphic_context mib)
+       Univ.AbstractContext.size (Declareops.inductive_polymorphic_context mib)
      in
      let mk u =
        let pind = (ind, u) in (mkIndU pind, type_of_ind env pind)
@@ -235,7 +233,7 @@ and constr_type_of_idkey env sigma (idkey : Vmvalues.id_key) stk =
   | ConstKey cst ->
      let cbody = Environ.lookup_constant cst env in
      let nb_univs =
-       Univ.AUContext.size (Declareops.constant_polymorphic_context cbody)
+       Univ.AbstractContext.size (Declareops.constant_polymorphic_context cbody)
      in
      let mk u =
        let pcst = (cst, u) in (mkConstU pcst, Typeops.type_of_constant_in env pcst)
@@ -427,7 +425,8 @@ let cbv_vm env sigma c t  =
   EConstr.of_constr (nf_val env sigma v t)
 
 let vm_infer_conv ?(pb=Reduction.CUMUL) env sigma t1 t2 =
-  Reductionops.infer_conv_gen (fun pb ~l2r sigma ts -> Vconv.vm_conv_gen pb)
+  Reductionops.infer_conv_gen (fun pb ~l2r sigma ts ->
+      Vconv.vm_conv_gen pb (Evd.existential_opt_value0 sigma))
     ~catch_incon:true ~pb env sigma t1 t2
 
 let _ = if Coq_config.bytecode_compiler then Reductionops.set_vm_infer_conv vm_infer_conv
