@@ -8,7 +8,6 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
-open Pp
 open CErrors
 open Util
 open Names
@@ -55,7 +54,7 @@ let clenv_term clenv c = meta_instance clenv.env clenv.cache c
 let clenv_meta_type clenv mv =
   let ty =
     try Evd.meta_ftype clenv.evd mv
-    with Not_found -> anomaly (str "unknown meta ?" ++ str (Nameops.string_of_meta mv) ++ str ".") in
+    with Not_found -> anomaly Pp.(str "unknown meta ?" ++ str (Nameops.string_of_meta mv) ++ str ".") in
   meta_instance clenv.env clenv.cache ty
 let clenv_value clenv = meta_instance clenv.env clenv.cache clenv.templval
 let clenv_type clenv = meta_instance clenv.env clenv.cache clenv.templtyp
@@ -64,8 +63,6 @@ let clenv_hnf_constr ce t = hnf_constr (cl_env ce) (cl_sigma ce) t
 
 let clenv_get_type_of ce c = Retyping.get_type_of (cl_env ce) (cl_sigma ce) c
 
-exception NotExtensibleClause
-
 let clenv_push_prod cl =
   let typ = whd_all (cl_env cl) (cl_sigma cl) (clenv_type cl) in
   let rec clrec typ = match EConstr.kind cl.evd typ with
@@ -77,12 +74,12 @@ let clenv_push_prod cl =
         let e' = meta_declare mv t ~name:na' cl.evd in
         let concl = if dep then subst1 (mkMeta mv) u else u in
         let def = applist (cl.templval.rebus,[mkMeta mv]) in
-        { templval = mk_freelisted def;
+        Some (mv, dep, { templval = mk_freelisted def;
           templtyp = mk_freelisted concl;
           evd = e';
           env = cl.env;
-          cache = create_meta_instance_subst e' }
-    | _ -> raise NotExtensibleClause
+          cache = create_meta_instance_subst e' })
+    | _ -> None
   in clrec typ
 
 (* Instantiate the first [bound] products of [t] with metas (all products if
@@ -126,12 +123,8 @@ let mk_clenv_from_env env sigma n (c,cty) =
     env = env;
     cache = create_meta_instance_subst evd }
 
-let mk_clenv_from_n gls n (c,cty) =
-  let env = Proofview.Goal.env gls in
-  let sigma = Tacmach.New.project gls in
-  mk_clenv_from_env env sigma n (c, cty)
-
-let mk_clenv_from gls = mk_clenv_from_n gls None
+let mk_clenv_from env sigma c = mk_clenv_from_env env sigma None c
+let mk_clenv_from_n env sigma n c = mk_clenv_from_env env sigma (Some n) c
 
 (******************************************************************)
 
@@ -153,12 +146,12 @@ let mentions clenv mv0 =
 let error_incompatible_inst clenv mv  =
   let na = meta_name clenv.evd mv in
   match na with
-      Name id ->
-        user_err ~hdr:"clenv_assign"
-          (str "An incompatible instantiation has already been found for " ++
-           Id.print id)
-    | _ ->
-        anomaly ~label:"clenv_assign" (Pp.str "non dependent metavar already assigned.")
+  | Name id ->
+    user_err
+      Pp.(str "An incompatible instantiation has already been found for " ++
+          Id.print id)
+  | _ ->
+    anomaly ~label:"clenv_assign" (Pp.str "non dependent metavar already assigned.")
 
 (* TODO: replace by clenv_unify (mkMeta mv) rhs ? *)
 let clenv_assign mv rhs clenv =
@@ -319,7 +312,7 @@ let clenv_unify ?(flags=default_unify_flags ()) cv_pb t1 t2 clenv =
 let clenv_unify_meta_types ?(flags=default_unify_flags ()) clenv =
   update_clenv_evd clenv (w_unify_meta_types ~flags:flags clenv.env clenv.evd)
 
-let clenv_unique_resolver_gen ?(flags=default_unify_flags ()) clenv concl =
+let clenv_unique_resolver ?(flags=default_unify_flags ()) clenv concl =
   if isMeta clenv.evd (fst (decompose_app_vect clenv.evd (whd_nored clenv.env clenv.evd clenv.templtyp.rebus))) then
     clenv_unify CUMUL ~flags (clenv_type clenv) concl
       (clenv_unify_meta_types ~flags clenv)
@@ -327,10 +320,6 @@ let clenv_unique_resolver_gen ?(flags=default_unify_flags ()) clenv concl =
     clenv_unify CUMUL ~flags
       (meta_reducible_instance clenv.env clenv.evd clenv.templtyp) concl clenv
 
-let clenv_unique_resolver ?flags clenv gl =
-  let concl = Proofview.Goal.concl gl in
-  clenv_unique_resolver_gen ?flags clenv concl
-
 let adjust_meta_source evd mv = function
   | loc,Evar_kinds.VarInstance id ->
     let rec match_name c l =
@@ -428,6 +417,11 @@ let fchain_flags () =
   { (default_unify_flags ()) with
     allow_K_in_toplevel_higher_order_unification = true }
 
+let clenv_instantiate ?(flags=fchain_flags ()) mv clenv (c, ty) =
+  (* unify the type of the template of [nextclenv] with the type of [mv] *)
+  let clenv = clenv_unify ~flags CUMUL ty (clenv_meta_type clenv mv) clenv in
+  clenv_assign mv c clenv
+
 let clenv_fchain ?with_univs ?(flags=fchain_flags ()) mv clenv nextclenv =
   (* Add the metavars of [nextclenv] to [clenv], with their name-environment *)
   let evd = meta_merge ?with_univs nextclenv.evd clenv.evd in
@@ -465,40 +459,32 @@ let clenv_independent clenv =
   List.filter (fun mv -> not (Metaset.mem mv deps)) mvs
 
 let qhyp_eq h1 h2 = match h1, h2 with
-| NamedHyp n1, NamedHyp n2 -> Id.equal n1 n2
+| NamedHyp n1, NamedHyp n2 -> lident_eq n1 n2
 | AnonHyp i1, AnonHyp i2 -> Int.equal i1 i2
 | _ -> false
 
 let check_bindings bl =
   match List.duplicates qhyp_eq (List.map (fun {CAst.v=x} -> fst x) bl) with
     | NamedHyp s :: _ ->
-        user_err
-          (str "The variable " ++ Id.print s ++
-           str " occurs more than once in binding list.");
+        user_err ?loc:s.CAst.loc
+          Pp.(str "The variable " ++ Id.print s.CAst.v ++
+              str " occurs more than once in binding list.");
     | AnonHyp n :: _ ->
         user_err
-          (str "The position " ++ int n ++
-           str " occurs more than once in binding list.")
+          Pp.(str "The position " ++ int n ++
+              str " occurs more than once in binding list.")
     | [] -> ()
 
-let explain_no_such_bound_variable evd id =
-  let fold l (n, clb) =
-    let na = match clb with
-    | Cltyp (na, _) -> na
-    | Clval (na, _, _) -> na
-    in
-    if na != Anonymous then Name.get_id na :: l else l
+let explain_no_such_bound_variable mvl {CAst.v=id;loc} =
+  let open Pp in
+  let expl = match mvl with
+  | [] -> str "(no bound variables at all in the expression)."
+  | [id] -> str "(possible name is: " ++ Id.print id ++ str ")."
+  | _ -> str "(possible names are: " ++ pr_enum Id.print mvl ++ str ")."
   in
-  let mvl = List.fold_left fold [] (Evd.meta_list evd) in
-  user_err ~hdr:"Evd.meta_with_name"
-    (str"No such bound variable " ++ Id.print id ++
-     (if mvl == [] then str " (no bound variables at all in the expression)."
-      else
-        (str" (possible name" ++
-         str (if List.length mvl == 1 then " is: " else "s are: ") ++
-         pr_enum Id.print mvl ++ str").")))
+  user_err ?loc (str "No such bound variable " ++ Id.print id ++ spc () ++ expl)
 
-let meta_with_name evd id =
+let meta_with_name evd ({CAst.v=id} as lid) =
   let na = Name id in
   let fold (l1, l2 as l) (n, clb) =
     let (na',def) = match clb with
@@ -511,30 +497,34 @@ let meta_with_name evd id =
   let (mvl, mvnodef) = List.fold_left fold ([], []) (Evd.meta_list evd) in
   match mvnodef, mvl with
     | _,[]  ->
-      explain_no_such_bound_variable evd id
-    | ([n],_|_,[n]) ->
+      let fold l (n, clb) =
+        let na = match clb with
+          | Cltyp (na, _) -> na
+          | Clval (na, _, _) -> na
+        in
+        if na != Anonymous then Name.get_id na :: l else l
+      in
+      let mvl = List.fold_left fold [] (Evd.meta_list evd) in
+      explain_no_such_bound_variable mvl lid
+    | (n::_,_|_,n::_) ->
         n
-    | _  ->
-        user_err ~hdr:"Evd.meta_with_name"
-          (str "Binder name \"" ++ Id.print id ++
-           strbrk "\" occurs more than once in clause.")
 
 let meta_of_binder clause loc mvs = function
   | NamedHyp s -> meta_with_name clause.evd s
   | AnonHyp n ->
       try List.nth mvs (n-1)
       with (Failure _|Invalid_argument _) ->
-        user_err  (str "No such binder.")
+        user_err Pp.(str "No such binder.")
 
 let error_already_defined b =
   match b with
     | NamedHyp id ->
-        user_err
-          (str "Binder name \"" ++ Id.print id ++
-           str"\" already defined with incompatible value.")
+        user_err ?loc:id.CAst.loc
+          Pp.(str "Binder name \"" ++ Id.print id.CAst.v ++
+              str"\" already defined with incompatible value.")
     | AnonHyp n ->
         anomaly
-          (str "Position " ++ int n ++ str" already defined.")
+          Pp.(str "Position " ++ int n ++ str" already defined.")
 
 let clenv_unify_binding_type clenv c t u =
   if isMeta clenv.evd (fst (decompose_app_vect clenv.evd (whd_nored clenv.env clenv.evd u))) then
@@ -583,8 +573,8 @@ let clenv_constrain_last_binding c clenv =
 
 let error_not_right_number_missing_arguments n =
   user_err
-    (strbrk "Not the right number of missing arguments (expected " ++
-      int n ++ str ").")
+    Pp.(strbrk "Not the right number of missing arguments (expected " ++
+        int n ++ str ").")
 
 let clenv_constrain_dep_args hyps_only bl clenv =
   if List.is_empty bl then
@@ -646,34 +636,30 @@ let clenv_pose_dependent_evars ?(with_evars=false) clenv =
       (RefinerError (env, sigma, UnresolvedBindings (List.map (meta_name clenv.evd) dep_mvs)));
   clenv_pose_metas_as_evars clenv dep_mvs
 
-let clenv_refine ?(with_evars=false) ?(with_classes=true) clenv =
-  Proofview.Goal.enter begin fun gl ->
-  let clenv = clenv_pose_dependent_evars ~with_evars clenv in
-  let evd' =
-    if with_classes then
-      let evd' =
-        Typeclasses.resolve_typeclasses ~filter:Typeclasses.all_evars
-          ~fail:(not with_evars) clenv.env clenv.evd
-      in
-      (* After an apply, all the subgoals including those dependent shelved ones are in
-         the hands of the user and resolution won't be called implicitely on them. *)
-      Typeclasses.make_unresolvables (fun x -> true) evd'
-    else clenv.evd
-  in
-  let clenv = update_clenv_evd clenv evd' in
-  Proofview.tclTHEN
-    (Proofview.Unsafe.tclEVARS (Evd.clear_metas evd'))
-    (refiner ~check:false EConstr.Unsafe.(to_constr (clenv_cast_meta clenv (clenv_value clenv))))
-  end
-
 open Unification
 
 let dft = default_unify_flags
 
-let res_pf ?with_evars ?(with_classes=true) ?(flags=dft ()) clenv =
+let res_pf ?(with_evars=false) ?(with_classes=true) ?(flags=dft ()) clenv =
   Proofview.Goal.enter begin fun gl ->
-    let clenv = clenv_unique_resolver ~flags clenv gl in
-    clenv_refine ?with_evars ~with_classes clenv
+    let concl = Proofview.Goal.concl gl in
+    let clenv = clenv_unique_resolver ~flags clenv concl in
+    let clenv = clenv_pose_dependent_evars ~with_evars clenv in
+    let evd' =
+      if with_classes then
+        let evd' =
+          Typeclasses.resolve_typeclasses ~filter:Typeclasses.all_evars
+            ~fail:(not with_evars) clenv.env clenv.evd
+        in
+        (* After an apply, all the subgoals including those dependent shelved ones are in
+          the hands of the user and resolution won't be called implicitely on them. *)
+        Typeclasses.make_unresolvables (fun x -> true) evd'
+      else clenv.evd
+    in
+    let clenv = update_clenv_evd clenv evd' in
+    Proofview.tclTHEN
+      (Proofview.Unsafe.tclEVARS (Evd.clear_metas evd'))
+      (refiner ~check:false EConstr.Unsafe.(to_constr (clenv_cast_meta clenv (clenv_value clenv))))
   end
 
 (* [unifyTerms] et [unify] ne semble pas gérer les Meta, en
@@ -707,9 +693,9 @@ let fail_quick_unif_flags = {
 (* let unifyTerms m n = walking (fun wc -> fst (w_Unify CONV m n [] wc)) *)
 let unify ?(flags=fail_quick_unif_flags) m =
   Proofview.Goal.enter begin fun gl ->
-    let env = Tacmach.New.pf_env gl in
-    let n = Tacmach.New.pf_concl gl in
-    let evd = clear_metas (Tacmach.New.project gl) in
+    let env = Tacmach.pf_env gl in
+    let n = Tacmach.pf_concl gl in
+    let evd = clear_metas (Tacmach.project gl) in
     try
       let evd' = w_unify env evd CONV ~flags m n in
         Proofview.Unsafe.tclEVARSADVANCE evd'
@@ -721,15 +707,16 @@ let unify ?(flags=fail_quick_unif_flags) m =
 (****************************************************************)
 (* Clausal environment for an application *)
 
+let rename_with = Goptions.declare_bool_option_and_ref ~depr:true ~key:["Apply";"With";"Renaming"]
+    ~value:false
+
 let make_clenv_binding_gen hyps_only n env sigma (c,t) = function
   | ImplicitBindings largs ->
       let clause = mk_clenv_from_env env sigma n (c,t) in
       clenv_constrain_dep_args hyps_only largs clause
   | ExplicitBindings lbind ->
-      let t = rename_bound_vars_as_displayed sigma Id.Set.empty [] t in
-      let clause = mk_clenv_from_env env sigma n
-        (c, t)
-      in clenv_match_args lbind clause
+      let t = if rename_with () then rename_bound_vars_as_displayed env sigma Id.Set.empty [] t else t in
+      let clause = mk_clenv_from_env env sigma n (c, t) in clenv_match_args lbind clause
   | NoBindings ->
       mk_clenv_from_env env sigma n (c,t)
 
@@ -740,10 +727,10 @@ let make_clenv_binding env sigma = make_clenv_binding_gen false None env sigma
 (* Pretty-print *)
 
 let pr_clenv clenv =
-  h
-    (str"TEMPL: " ++ Termops.Internal.print_constr_env clenv.env clenv.evd clenv.templval.rebus ++
-     str" : " ++ Termops.Internal.print_constr_env clenv.env clenv.evd clenv.templtyp.rebus ++ fnl () ++
-     pr_evar_map (Some 2) clenv.env clenv.evd)
+  let prc = Termops.Internal.print_constr_env clenv.env clenv.evd in
+  Pp.(h (str"TEMPL: " ++ prc clenv.templval.rebus ++
+         str" : " ++ prc clenv.templtyp.rebus ++ fnl () ++
+         pr_evar_map (Some 2) clenv.env clenv.evd))
 
 (****************************************************************)
 (** Evar version of mk_clenv *)
@@ -767,22 +754,22 @@ let make_evar_clause env sigma ?len t =
   | None -> -1
   | Some n -> assert (0 <= n); n
   in
-  (* FIXME: do the renaming online *)
-  let t = rename_bound_vars_as_displayed sigma Id.Set.empty [] t in
+  let t = if rename_with () then rename_bound_vars_as_displayed env sigma Id.Set.empty [] t else t in
   let rec clrec (sigma, holes) inst n t =
     if n = 0 then (sigma, holes, t)
     else match EConstr.kind sigma t with
     | Cast (t, _, _) -> clrec (sigma, holes) inst n t
     | Prod (na, t1, t2) ->
       (* Share the evar instances as we are living in the same context *)
-      let inst, ctx, args, subst = match inst with
+      let inst, ctx, args, identity, subst = match inst with
       | None ->
         (* Dummy type *)
         let ctx, _, args, subst = push_rel_context_to_named_context env sigma mkProp in
-        Some (ctx, args, subst), ctx, args, subst
-      | Some (ctx, args, subst) -> inst, ctx, args, subst
+        let id = if Int.equal (Environ.nb_rel env) 0 then Identity.make args else Identity.none () in
+        Some (ctx, args, id, subst), ctx, args, id, subst
+      | Some (ctx, args, id, subst) -> inst, ctx, args, id, subst
       in
-      let (sigma, ev) = new_pure_evar ~typeclass_candidate:false ctx sigma (csubst_subst subst t1) in
+      let (sigma, ev) = new_pure_evar ~identity ~typeclass_candidate:false ctx sigma (csubst_subst subst t1) in
       let ev = mkEvar (ev, args) in
       let dep = not (noccurn sigma 1 t2) in
       let hole = {
@@ -792,7 +779,7 @@ let make_evar_clause env sigma ?len t =
         (* We fix it later *)
         hole_name = na.binder_name;
       } in
-      let t2 = if dep then subst1 ev t2 else t2 in
+      let t2 = subst1 ev t2 in
       clrec (sigma, hole :: holes) inst (pred n) t2
     | LetIn (na, b, _, t) -> clrec (sigma, holes) inst n (subst1 b t)
     | _ -> (sigma, holes, t)
@@ -802,32 +789,21 @@ let make_evar_clause env sigma ?len t =
   let clause = { cl_concl = t; cl_holes = holes } in
   (sigma, clause)
 
-let explain_no_such_bound_variable holes id =
-  let fold h accu = match h.hole_name with
-  | Anonymous -> accu
-  | Name id -> id :: accu
-  in
-  let mvl = List.fold_right fold holes [] in
-  let expl = match mvl with
-  | [] -> str " (no bound variables at all in the expression)."
-  | [id] -> str "(possible name is: " ++ Id.print id ++ str ")."
-  | _ -> str "(possible names are: " ++ pr_enum Id.print mvl ++ str ")."
-  in
-  user_err  (str "No such bound variable " ++ Id.print id ++ expl)
-
-let evar_with_name holes id =
+let evar_with_name holes ({CAst.v=id} as lid) =
   let map h = match h.hole_name with
   | Anonymous -> None
   | Name id' -> if Id.equal id id' then Some h else None
   in
   let hole = List.map_filter map holes in
   match hole with
-  | [] -> explain_no_such_bound_variable holes id
-  | [h] -> h.hole_evar
-  | _ ->
-    user_err
-      (str "Binder name \"" ++ Id.print id ++
-        str "\" occurs more than once in clause.")
+  | [] ->
+    let fold h accu = match h.hole_name with
+      | Anonymous -> accu
+      | Name id -> id :: accu
+    in
+    let mvl = List.fold_right fold holes [] in
+    explain_no_such_bound_variable mvl lid
+  | h::_ -> h.hole_evar
 
 let evar_of_binder holes = function
 | NamedHyp s -> evar_with_name holes s
@@ -837,7 +813,7 @@ let evar_of_binder holes = function
     let h = List.nth nondeps (pred n) in
     h.hole_evar
   with e when CErrors.noncritical e ->
-    user_err  (str "No such binder.")
+    user_err Pp.(str "No such binder.")
 
 let define_with_type sigma env ev c =
   let open EConstr in
