@@ -27,7 +27,7 @@ open Hints
 (*          tactics with a trace mechanism for automatic search           *)
 (**************************************************************************)
 
-let priority l = List.filter (fun (_, hint) -> Int.equal (FullHint.priority hint) 0) l
+let priority l = List.filter (fun hint -> Int.equal (FullHint.priority hint) 0) l
 
 let compute_secvars gl =
   let hyps = Proofview.Goal.hyps gl in
@@ -68,12 +68,7 @@ let auto_unif_flags =
 (* Try unification with the precompiled clause, then use registered Apply *)
 
 let unify_resolve flags h = Hints.hint_res_pf ~flags h
-
-let unify_resolve_nodelta h = unify_resolve auto_unif_flags h
-
-let unify_resolve_gen = function
-  | None -> unify_resolve_nodelta
-  | Some flags -> unify_resolve flags
+let unify_resolve_nodelta h = Hints.hint_res_pf ~flags:auto_unif_flags h
 
 let exact h =
   Proofview.Goal.enter begin fun gl ->
@@ -106,11 +101,11 @@ let conclPattern concl pat tac =
           Proofview.tclUNIT (Constr_matching.matches env sigma pat concl)
         with Constr_matching.PatternMatchingFailure as exn ->
           let _, info = Exninfo.capture exn in
-          Tacticals.New.tclZEROMSG ~info (str "pattern-matching failed")
+          Tacticals.tclZEROMSG ~info (str "pattern-matching failed")
   in
   Proofview.Goal.enter begin fun gl ->
      let env = Proofview.Goal.env gl in
-     let sigma = Tacmach.New.project gl in
+     let sigma = Tacmach.project gl in
      constr_bindings env sigma >>= fun constr_bindings ->
      Proofview.tclProofInfo [@ocaml.warning "-3"] >>= fun (_name, poly) ->
      let open Genarg in
@@ -250,7 +245,7 @@ let tclTRY_dbg d tac =
       pr_info_trace env sigma d;
       Proofview.tclUNIT () in
   let after = delay (fun () -> pr_info_nop d; Proofview.tclUNIT ()) in
-  Tacticals.New.tclORELSE0 tac after
+  Tacticals.tclORELSE0 tac after
 
 (**************************************************************************)
 (*                           The Trivial tactic                           *)
@@ -260,9 +255,6 @@ let tclTRY_dbg d tac =
 (* Papageno : cette fonction a été pas mal simplifiée depuis que la base
   de Hint impérative a été remplacée par plusieurs bases fonctionnelles *)
 
-let flags_of_state st =
-  auto_unif_flags_of st st
-
 let auto_flags_of_state st =
   auto_unif_flags_of TransparentState.full st
 
@@ -271,8 +263,8 @@ let hintmap_of env sigma secvars hdc concl =
   | None -> Hint_db.map_none ~secvars
   | Some hdc ->
      if occur_existential sigma concl then
-       (fun db -> match Hint_db.map_existential sigma ~secvars hdc concl db with
-                  | ModeMatch l -> l
+       (fun db -> match Hint_db.map_eauto env sigma ~secvars hdc concl db with
+                  | ModeMatch (_, l) -> l
                   | ModeMismatch -> [])
      else Hint_db.map_auto env sigma ~secvars hdc concl
 
@@ -283,90 +275,52 @@ let exists_evaluable_reference env = function
 let dbg_intro dbg = tclLOG dbg (fun _ _ -> str "intro") intro
 let dbg_assumption dbg = tclLOG dbg (fun _ _ -> str "assumption") assumption
 
-let rec trivial_fail_db dbg mod_delta db_list local_db =
+let rec trivial_fail_db dbg db_list local_db =
   let intro_tac =
-    Tacticals.New.tclTHEN (dbg_intro dbg)
+    Tacticals.tclTHEN (dbg_intro dbg)
       ( Proofview.Goal.enter begin fun gl ->
-          let sigma = Tacmach.New.project gl in
+          let sigma = Tacmach.project gl in
           let env = Proofview.Goal.env gl in
-          let decl = Tacmach.New.pf_last_hyp gl in
+          let decl = Tacmach.pf_last_hyp gl in
           let hyp = Context.Named.Declaration.get_id decl in
           let local_db = push_resolve_hyp env sigma hyp local_db in
-          trivial_fail_db dbg mod_delta db_list local_db
+          trivial_fail_db dbg db_list local_db
       end)
   in
   Proofview.Goal.enter begin fun gl ->
-    let concl = Tacmach.New.pf_concl gl in
-    let sigma = Tacmach.New.project gl in
+    let concl = Tacmach.pf_concl gl in
+    let sigma = Tacmach.project gl in
     let env = Proofview.Goal.env gl in
     let secvars = compute_secvars gl in
-    Tacticals.New.tclFIRST
+    Tacticals.tclFIRST
       ((dbg_assumption dbg)::intro_tac::
-          (List.map Tacticals.New.tclCOMPLETE
-             (trivial_resolve env sigma dbg mod_delta db_list local_db secvars concl)))
+          (List.map Tacticals.tclCOMPLETE
+             (trivial_resolve env sigma dbg db_list local_db secvars concl)))
   end
 
-and my_find_search_nodelta env sigma db_list local_db secvars hdc concl =
-  List.map (fun hint -> (None,hint))
-    (List.map_append (hintmap_of env sigma secvars hdc concl) (local_db::db_list))
+and my_find_search env sigma db_list local_db secvars hdc concl =
+  List.map_append (hintmap_of env sigma secvars hdc concl) (local_db::db_list)
 
-and my_find_search mod_delta =
-  if mod_delta then my_find_search_delta
-  else my_find_search_nodelta
-
-and my_find_search_delta env sigma db_list local_db secvars hdc concl =
-  let f = hintmap_of env sigma secvars hdc concl in
-    if occur_existential sigma concl then
-      List.map_append
-        (fun db ->
-          if Hint_db.use_dn db then
-            let flags = flags_of_state (Hint_db.transparent_state db) in
-              List.map (fun x -> (Some flags,x)) (f db)
-          else
-            let flags = auto_flags_of_state (Hint_db.transparent_state db) in
-              List.map (fun x -> (Some flags,x)) (f db))
-        (local_db::db_list)
-    else
-      List.map_append (fun db ->
-        if Hint_db.use_dn db then
-          let flags = flags_of_state (Hint_db.transparent_state db) in
-            List.map (fun x -> (Some flags, x)) (f db)
-        else
-          let st = Hint_db.transparent_state db in
-          let flags, l =
-            let l =
-              match hdc with
-              | None -> Hint_db.map_none ~secvars db
-              | Some hdc ->
-                  if TransparentState.is_empty st
-                  then Hint_db.map_auto env sigma ~secvars hdc concl db
-                  else match Hint_db.map_existential sigma ~secvars hdc concl db with
-                      | ModeMatch l -> l
-                      | ModeMismatch -> []
-            in auto_flags_of_state st, l
-          in List.map (fun x -> (Some flags,x)) l)
-        (local_db::db_list)
-
-and tac_of_hint dbg db_list local_db concl (flags, h) =
+and tac_of_hint dbg db_list local_db concl h =
   let tactic = function
-    | Res_pf h -> unify_resolve_gen flags h
+    | Res_pf h -> unify_resolve_nodelta h
     | ERes_pf _ -> Proofview.Goal.enter (fun gl ->
         let info = Exninfo.reify () in
-        Tacticals.New.tclZEROMSG ~info (str "eres_pf"))
+        Tacticals.tclZEROMSG ~info (str "eres_pf"))
     | Give_exact h  -> exact h
     | Res_pf_THEN_trivial_fail h ->
-      Tacticals.New.tclTHEN
-        (unify_resolve_gen flags h)
+      Tacticals.tclTHEN
+        (unify_resolve_nodelta h)
         (* With "(debug) trivial", we shouldn't end here, and
            with "debug auto" we don't display the details of inner trivial *)
-        (trivial_fail_db (no_dbg dbg) (not (Option.is_empty flags)) db_list local_db)
+        (trivial_fail_db (no_dbg dbg) db_list local_db)
     | Unfold_nth c ->
       Proofview.Goal.enter begin fun gl ->
-       if exists_evaluable_reference (Tacmach.New.pf_env gl) c then
-         Tacticals.New.tclPROGRESS (reduce (Unfold [AllOccurrences,c]) Locusops.onConcl)
+       if exists_evaluable_reference (Tacmach.pf_env gl) c then
+         Tacticals.tclPROGRESS (reduce (Unfold [AllOccurrences,c]) Locusops.onConcl)
        else
          let info = Exninfo.reify () in
-         Tacticals.New.tclFAIL ~info 0 (str"Unbound reference")
+         Tacticals.tclFAIL ~info 0 (str"Unbound reference")
        end
     | Extern (p, tacast) ->
       conclPattern concl p tacast
@@ -380,7 +334,7 @@ and tac_of_hint dbg db_list local_db concl (flags, h) =
   in
   tclLOG dbg pr_hint (FullHint.run h tactic)
 
-and trivial_resolve env sigma dbg mod_delta db_list local_db secvars cl =
+and trivial_resolve env sigma dbg db_list local_db secvars cl =
   try
     let head =
       try let hdconstr = decompose_app_bound sigma cl in
@@ -389,7 +343,7 @@ and trivial_resolve env sigma dbg mod_delta db_list local_db secvars cl =
     in
       List.map (tac_of_hint dbg db_list local_db cl)
         (priority
-            (my_find_search mod_delta env sigma db_list local_db secvars head cl))
+            (my_find_search env sigma db_list local_db secvars head cl))
   with Not_found -> []
 
 (** The use of the "core" database can be de-activated by passing
@@ -399,24 +353,24 @@ let trivial ?(debug=Off) lems dbnames =
   Hints.wrap_hint_warning @@
   Proofview.Goal.enter begin fun gl ->
   let env = Proofview.Goal.env gl in
-  let sigma = Tacmach.New.project gl in
+  let sigma = Tacmach.project gl in
   let db_list = make_db_list dbnames in
   let d = mk_trivial_dbg debug in
   let hints = make_local_hint_db env sigma false lems in
   tclTRY_dbg d
-    (trivial_fail_db d false db_list hints)
+    (trivial_fail_db d db_list hints)
   end
 
 let full_trivial ?(debug=Off) lems =
   Hints.wrap_hint_warning @@
   Proofview.Goal.enter begin fun gl ->
   let env = Proofview.Goal.env gl in
-  let sigma = Tacmach.New.project gl in
+  let sigma = Tacmach.project gl in
   let db_list = current_pure_db () in
   let d = mk_trivial_dbg debug in
   let hints = make_local_hint_db env sigma false lems in
   tclTRY_dbg d
-    (trivial_fail_db d false db_list hints)
+    (trivial_fail_db d db_list hints)
   end
 
 let gen_trivial ?(debug=Off) lems = function
@@ -429,7 +383,7 @@ let h_trivial ?(debug=Off) lems l = gen_trivial ~debug lems l
 (*                       The classical Auto tactic                        *)
 (**************************************************************************)
 
-let possible_resolve env sigma dbg mod_delta db_list local_db secvars cl =
+let possible_resolve env sigma dbg db_list local_db secvars cl =
   try
     let head =
       try let hdconstr = decompose_app_bound sigma cl in
@@ -437,47 +391,47 @@ let possible_resolve env sigma dbg mod_delta db_list local_db secvars cl =
       with Bound -> None
     in
       List.map (tac_of_hint dbg db_list local_db cl)
-        (my_find_search mod_delta env sigma db_list local_db secvars head cl)
+        (my_find_search env sigma db_list local_db secvars head cl)
   with Not_found -> []
 
 (* Introduce an hypothesis, then call the continuation tactic [kont]
    with the hint db extended with the so-obtained hypothesis *)
 
 let intro_register dbg kont db =
-  Tacticals.New.tclTHEN (dbg_intro dbg)
+  Tacticals.tclTHEN (dbg_intro dbg)
     (Proofview.Goal.enter begin fun gl ->
       let extend_local_db decl db =
-        let env = Tacmach.New.pf_env gl in
-        let sigma = Tacmach.New.project gl in
+        let env = Tacmach.pf_env gl in
+        let sigma = Tacmach.project gl in
         push_resolve_hyp env sigma (Context.Named.Declaration.get_id decl) db
       in
-      Tacticals.New.onLastDecl (fun decl -> kont (extend_local_db decl db))
+      Tacticals.onLastDecl (fun decl -> kont (extend_local_db decl db))
     end)
 
 (* n is the max depth of search *)
 (* local_db contains the local Hypotheses *)
 
-let search d n mod_delta db_list local_db =
+let search d n db_list local_db =
   let rec search d n local_db =
     (* spiwack: the test of [n] to 0 must be done independently in
        each goal. Hence the [tclEXTEND] *)
     Proofview.tclEXTEND [] begin
       if Int.equal n 0 then
         let info = Exninfo.reify () in
-        Tacticals.New.tclZEROMSG ~info (str"BOUND 2")
+        Tacticals.tclZEROMSG ~info (str"BOUND 2")
       else
-        Tacticals.New.tclORELSE0 (dbg_assumption d)
-          (Tacticals.New.tclORELSE0 (intro_register d (search d n) local_db)
+        Tacticals.tclORELSE0 (dbg_assumption d)
+          (Tacticals.tclORELSE0 (intro_register d (search d n) local_db)
              ( Proofview.Goal.enter begin fun gl ->
-               let concl = Tacmach.New.pf_concl gl in
-               let sigma = Tacmach.New.project gl in
+               let concl = Tacmach.pf_concl gl in
+               let sigma = Tacmach.project gl in
                let env = Proofview.Goal.env gl in
                let secvars = compute_secvars gl in
                let d' = incr_dbg d in
-               Tacticals.New.tclFIRST
+               Tacticals.tclFIRST
                  (List.map
-                    (fun ntac -> Tacticals.New.tclTHEN ntac (search d' (n-1) local_db))
-                    (possible_resolve env sigma d mod_delta db_list local_db secvars concl))
+                    (fun ntac -> Tacticals.tclTHEN ntac (search d' (n-1) local_db))
+                    (possible_resolve env sigma d db_list local_db secvars concl))
              end))
     end []
   in
@@ -485,38 +439,35 @@ let search d n mod_delta db_list local_db =
 
 let default_search_depth = ref 5
 
-let delta_auto debug mod_delta n lems dbnames =
+let delta_auto debug n lems dbnames =
   Hints.wrap_hint_warning @@
   Proofview.Goal.enter begin fun gl ->
   let env = Proofview.Goal.env gl in
-  let sigma = Tacmach.New.project gl in
+  let sigma = Tacmach.project gl in
   let db_list = make_db_list dbnames in
   let d = mk_auto_dbg debug in
   let hints = make_local_hint_db env sigma false lems in
   tclTRY_dbg d
-    (search d n mod_delta db_list hints)
+    (search d n db_list hints)
   end
 
-let auto ?(debug=Off) n = delta_auto debug false n
-
-let new_auto ?(debug=Off) n = delta_auto debug true n
+let auto ?(debug=Off) n = delta_auto debug n
 
 let default_auto = auto !default_search_depth [] []
 
-let delta_full_auto ?(debug=Off) mod_delta n lems =
+let delta_full_auto ?(debug=Off) n lems =
   Hints.wrap_hint_warning @@
   Proofview.Goal.enter begin fun gl ->
   let env = Proofview.Goal.env gl in
-  let sigma = Tacmach.New.project gl in
+  let sigma = Tacmach.project gl in
   let db_list = current_pure_db () in
   let d = mk_auto_dbg debug in
   let hints = make_local_hint_db env sigma false lems in
   tclTRY_dbg d
-    (search d n mod_delta db_list hints)
+    (search d n db_list hints)
   end
 
-let full_auto ?(debug=Off) n = delta_full_auto ~debug false n
-let new_full_auto ?(debug=Off) n = delta_full_auto ~debug true n
+let full_auto ?(debug=Off) n = delta_full_auto ~debug n
 
 let default_full_auto = full_auto !default_search_depth []
 
