@@ -28,10 +28,10 @@ open Globnames
 open Reductionops
 open Typing
 open Retyping
-open Tacmach.New
+open Tacmach
 open Logic
 open Hipattern
-open Tacticals.New
+open Tacticals
 open Tactics
 open Tacred
 open Coqlib
@@ -53,7 +53,6 @@ module NamedDecl = Context.Named.Declaration
 
 type inj_flags = {
     keep_proof_equalities : bool;
-    injection_in_context : bool;
     injection_pattern_l2r_order : bool;
   }
 
@@ -63,18 +62,9 @@ let use_injection_pattern_l2r_order = function
   | None -> true
   | Some flags -> flags.injection_pattern_l2r_order
 
-let injection_in_context = ref false
-
-let use_injection_in_context = function
-  | None -> !injection_in_context
-  | Some flags -> flags.injection_in_context
-
-let () =
-  declare_bool_option
-    { optdepr  = false;
-      optkey   = ["Structural";"Injection"];
-      optread  = (fun () -> !injection_in_context) ;
-      optwrite = (fun b -> injection_in_context := b) }
+let injection_in_context_flag =
+  declare_bool_option_and_ref
+    ~depr:false ~key:["Structural";"Injection"] ~value:false
 
 (* Rewriting tactics *)
 
@@ -157,7 +147,7 @@ let instantiate_lemma_all frzevars gl c ty l l2r concl =
     let clenv = Clenv.update_clenv_evd eqclause evd' in
     Clenv.clenv_pose_dependent_evars ~with_evars:true clenv
   in
-  let flags = make_flags frzevars (Tacmach.New.project gl) rewrite_unif_flags eqclause in
+  let flags = make_flags frzevars (Tacmach.project gl) rewrite_unif_flags eqclause in
   let occs =
     w_unify_to_subterm_all ~flags env eqclause.evd
       ((if l2r then c1 else c2),concl)
@@ -242,7 +232,7 @@ let rewrite_elim with_evars frzevars cls c e =
   Proofview.Goal.enter begin fun gl ->
   let flags = if Unification.is_keyed_unification ()
               then rewrite_keyed_unif_flags else rewrite_conv_closed_unif_flags in
-  let flags = make_flags frzevars (Tacmach.New.project gl) flags c in
+  let flags = make_flags frzevars (Tacmach.project gl) flags c in
   general_elim_clause with_evars flags cls c e
   end
 
@@ -383,7 +373,7 @@ let find_elim hdcncl lft2rgt dep cls ot =
             let l' = Label.of_id (add_suffix (Label.to_id l) "_r")  in
             let c1' = Global.constant_of_delta_kn (KerName.make mp l') in
             if not (Environ.mem_constant c1' (Global.env ())) then
-              user_err ~hdr:"Equality.find_elim"
+              user_err
                 (str "Cannot find rewrite principle " ++ Label.print l' ++ str ".");
             c1'
           end
@@ -459,7 +449,7 @@ let general_rewrite ~where:cls ~l2r:lft2rgt occs ~freeze:frzevars ~dep:dep_proof
     rewrite_side_tac (Hook.get forward_general_setoid_rewrite_clause cls lft2rgt occs (c,l) ~new_goals:[]) tac)
   else
     Proofview.Goal.enter begin fun gl ->
-      let sigma = Tacmach.New.project gl in
+      let sigma = Tacmach.project gl in
       let env = Proofview.Goal.env gl in
     let ctype = get_type_of env sigma c in
     let rels, t = decompose_prod_assum sigma (whd_betaiotazeta env sigma ctype) in
@@ -538,7 +528,7 @@ let general_rewrite_clause l2r with_evars ?tac c cl =
 
 let apply_special_clear_request clear_flag f =
   Proofview.Goal.enter begin fun gl ->
-    let sigma = Tacmach.New.project gl in
+    let sigma = Tacmach.project gl in
     let env = Proofview.Goal.env gl in
     try
       let (sigma, (c, bl)) = f env sigma in
@@ -556,7 +546,7 @@ type multi =
 let general_multi_rewrite with_evars l cl tac =
   let do1 l2r f =
     Proofview.Goal.enter begin fun gl ->
-      let sigma = Tacmach.New.project gl in
+      let sigma = Tacmach.project gl in
       let env = Proofview.Goal.env gl in
       let (sigma, c) = f env sigma in
       tclWITHHOLES with_evars
@@ -632,9 +622,9 @@ let replace_using_leibniz clause c1 c2 l2r unsafe try_prove_eq_opt =
   let t1 = get_type_of c1
   and t2 = get_type_of c2 in
   let evd =
-    if unsafe then Some (Tacmach.New.project gl)
+    if unsafe then Some (Tacmach.project gl)
     else
-      try Some (Evarconv.unify_delay (Proofview.Goal.env gl) (Tacmach.New.project gl) t1 t2)
+      try Some (Evarconv.unify_delay (Proofview.Goal.env gl) (Tacmach.project gl) t1 t2)
       with Evarconv.UnableToUnify _ -> None
   in
   match evd with
@@ -647,8 +637,8 @@ let replace_using_leibniz clause c1 c2 l2r unsafe try_prove_eq_opt =
       try lib_ref "core.identity.type", lib_ref "core.identity.sym"
       with UserError _ ->
         user_err (strbrk "Need a registration for either core.eq.type and core.eq.sym or core.identity.type and core.identity.sym.") in
-    Tacticals.New.pf_constr_of_global sym >>= fun sym ->
-    Tacticals.New.pf_constr_of_global e >>= fun e ->
+    Tacticals.pf_constr_of_global sym >>= fun sym ->
+    Tacticals.pf_constr_of_global e >>= fun e ->
     let eq = applist (e, [t1;c1;c2]) in
     tclTHENLAST
       (replace_core clause l2r eq)
@@ -728,15 +718,49 @@ let keep_proof_equalities = function
   | None -> !keep_proof_equalities_for_injection
   | Some flags -> flags.keep_proof_equalities
 
+module KeepEqualities =
+struct
+  type t = inductive
+  module Set = Indset_env
+  let encode _env r = Nametab.global_inductive r
+  let subst subst obj = Mod_subst.subst_ind subst obj
+  let printer ind = Nametab.pr_global_env Id.Set.empty (GlobRef.IndRef ind)
+  let key = ["Keep"; "Equalities"]
+  let title = "Prop-valued inductive types for which injection keeps equality proofs"
+  let member_message ind b =
+    let b = if b then mt () else str "not " in
+    str "Equality proofs over " ++ (printer ind) ++
+      str " are " ++ b ++ str "kept by injection"
+end
+
+module KeepEqualitiesTable = Goptions.MakeRefTable(KeepEqualities)
+
+let set_keep_equality = KeepEqualitiesTable.set
+
 (* [keep_proofs] is relevant for types in Prop with elimination in Type *)
 (* In particular, it is relevant for injection but not for discriminate *)
 
+let keep_head_inductive sigma c =
+  (* Note that we do not weak-head normalize c before checking it is an
+     applied inductive, because [get_sort_family_of] did not use to either.
+     As a matter of fact, if it reduces to an applied template inductive
+     type but is not syntactically equal to it, it will fail to project. *)
+  let _, hd = EConstr.decompose_prod sigma c in
+  let hd, _ = Termops.decompose_app_vect sigma hd in
+  match EConstr.kind sigma hd with
+  | Ind (ind, _) -> KeepEqualitiesTable.active ind
+  | _ -> false
+
 let find_positions env sigma ~keep_proofs ~no_discr t1 t2 =
   let project env sorts posn t1 t2 =
     let ty1 = get_type_of env sigma t1 in
-    let s = get_sort_family_of ~truncation_style:true env sigma ty1 in
-    if List.mem_f Sorts.family_equal s sorts
-    then [(List.rev posn,t1,t2)] else []
+    let keep =
+      if keep_head_inductive sigma ty1 then true
+      else
+        let s = get_sort_family_of env sigma ty1 in
+        List.mem_f Sorts.family_equal s sorts
+    in
+    if keep then [(List.rev posn,t1,t2)] else []
   in
   let rec findrec sorts posn t1 t2 =
     let hd1,args1 = whd_all_stack env sigma t1 in
@@ -918,7 +942,7 @@ let build_selector env sigma dirn c ind special default =
           on (c bool true) = (c bool false)
           CP : changed assert false in a more informative error
        *)
-      user_err ~hdr:"Equality.construct_discriminator"
+      user_err
         (str "Cannot discriminate on inductive constructors with \
                  dependent types.") in
   let (indp,_) = dest_ind_family indf in
@@ -984,7 +1008,8 @@ let gen_absurdity id =
 *)
 
 let ind_scheme_of_eq lbeq to_kind =
-  let (mib,mip) = Global.lookup_inductive (destIndRef lbeq.eq) in
+  Proofview.tclENV >>= fun env ->
+  let (mib,mip) = Inductive.lookup_mind_specif env (destIndRef lbeq.eq) in
   let from_kind = inductive_sort_family mip in
   (* use ind rather than case by compatibility *)
   let kind = Elimschemes.nondep_elim_scheme from_kind to_kind in
@@ -1062,7 +1087,7 @@ let onEquality with_evars tac (c,lbindc) =
 
 let onNegatedEquality with_evars tac =
   Proofview.Goal.enter begin fun gl ->
-    let sigma = Tacmach.New.project gl in
+    let sigma = Tacmach.project gl in
     let ccl = Proofview.Goal.concl gl in
     let env = Proofview.Goal.env gl in
     match EConstr.kind sigma (hnf_constr env sigma ccl) with
@@ -1319,7 +1344,7 @@ let set_eq_dec_scheme_kind k = eq_dec_scheme_kind_name := (fun _ -> k)
 let inject_if_homogenous_dependent_pair ty =
   Proofview.Goal.enter begin fun gl ->
   try
-    let sigma = Tacmach.New.project gl in
+    let sigma = Tacmach.project gl in
     let eq,u,(t,t1,t2) = pf_apply find_this_eq_data_decompose gl ty in
     (* fetch the informations of the  pair *)
     let sigTconstr   = Coqlib.(lib_ref "core.sigT.type") in
@@ -1347,10 +1372,10 @@ let inject_if_homogenous_dependent_pair ty =
       [
        intro;
        onLastHyp (fun hyp ->
-        Tacticals.New.pf_constr_of_global Coqlib.(lib_ref "core.eq.type") >>= fun ceq ->
+        Tacticals.pf_constr_of_global Coqlib.(lib_ref "core.eq.type") >>= fun ceq ->
         tclTHENS (cut (mkApp (ceq,new_eq_args)))
           [clear [destVar sigma hyp];
-           Tacticals.New.pf_constr_of_global inj2 >>= fun inj2 ->
+           Tacticals.pf_constr_of_global inj2 >>= fun inj2 ->
            Logic.refiner ~check:true EConstr.Unsafe.(to_constr
              (mkApp(inj2,[|ar1.(0);mkConst c;ar1.(1);ar1.(2);ar1.(3);ar2.(3);hyp|])))
           ])]
@@ -1395,7 +1420,7 @@ let inject_at_positions env sigma l2r eq posns tac =
     tclZEROMSG (str "Failed to decompose the equality.")
   else
     Proofview.tclTHEN (Proofview.Unsafe.tclEVARS !evdref)
-    (Tacticals.New.tclTHENFIRST
+    (Tacticals.tclTHENFIRST
       (Proofview.tclIGNORE (Proofview.Monad.List.map
          (fun (pf,ty) -> tclTHENS (cut ty)
            [inject_if_homogenous_dependent_pair ty;
@@ -1437,11 +1462,10 @@ let get_previous_hyp_position id gl =
   in
   aux MoveLast (Proofview.Goal.hyps gl)
 
-let injEq flags ?(old=false) with_evars clear_flag ipats =
+let injEq flags ?(injection_in_context = injection_in_context_flag ()) with_evars clear_flag ipats =
   (* Decide which compatibility mode to use *)
   let ipats_style, l2r, dft_clear_flag, bounded_intro = match ipats with
-    | None when not old && use_injection_in_context flags ->
-      Some [], true, true, true
+    | None when injection_in_context -> Some [], true, true, true
     | None -> None, false, false, false
     | _ -> let b = use_injection_pattern_l2r_order flags in ipats, b, b, b in
   (* Built the post tactic depending on compatibility mode *)
@@ -1465,18 +1489,19 @@ let injEq flags ?(old=false) with_evars clear_flag ipats =
     | None -> tclIDTAC in
   injEqThen (keep_proof_equalities flags) post_tac l2r
 
-let inj flags ipats with_evars clear_flag = onEquality with_evars (injEq flags with_evars clear_flag ipats)
+let inj flags ?injection_in_context ipats with_evars clear_flag =
+  onEquality with_evars (injEq flags ?injection_in_context with_evars clear_flag ipats)
 
-let injClause flags ipats with_evars = function
-  | None -> onNegatedEquality with_evars (injEq flags with_evars None ipats)
-  | Some c -> onInductionArg (inj flags ipats with_evars) c
+let injClause flags ?injection_in_context ipats with_evars = function
+  | None -> onNegatedEquality with_evars (injEq flags ?injection_in_context with_evars None ipats)
+  | Some c -> onInductionArg (inj flags ?injection_in_context ipats with_evars) c
 
 let simpleInjClause flags with_evars = function
-  | None -> onNegatedEquality with_evars (injEq flags ~old:true with_evars None None)
-  | Some c -> onInductionArg (fun clear_flag -> onEquality with_evars (injEq flags ~old:true with_evars clear_flag None)) c
+  | None -> onNegatedEquality with_evars (injEq flags ~injection_in_context:false with_evars None None)
+  | Some c -> onInductionArg (fun clear_flag -> onEquality with_evars (injEq flags ~injection_in_context:false with_evars clear_flag None)) c
 
-let injConcl flags = injClause flags None false None
-let injHyp flags clear_flag id = injClause flags None false (Some (clear_flag,ElimOnIdent CAst.(make id)))
+let injConcl flags ?injection_in_context () = injClause flags ?injection_in_context None false None
+let injHyp flags ?injection_in_context clear_flag id = injClause flags ?injection_in_context None false (Some (clear_flag,ElimOnIdent CAst.(make id)))
 
 let decompEqThen keep_proofs ntac eq =
   let { eq_data = (_, (_,t1,t2) as u); eq_clenv = clause } = eq in
@@ -1678,15 +1703,6 @@ user = raise user error specific to rewrite
 (**********************************************************************)
 (* Substitutions tactics (JCF) *)
 
-let regular_subst_tactic = ref true
-
-let () =
-  declare_bool_option
-    { optdepr  = true;
-      optkey   = ["Regular";"Subst";"Tactic"];
-      optread  = (fun () -> !regular_subst_tactic);
-      optwrite = (:=) regular_subst_tactic }
-
 let restrict_to_eq_and_identity eq = (* compatibility *)
   if not (Constr.isRefX (lib_ref "core.eq.type") eq) &&
     not (Constr.isRefX (lib_ref "core.identity.type") eq)
@@ -1713,7 +1729,7 @@ exception FoundDepInGlobal of Id.t option * GlobRef.t
 
 let test_non_indirectly_dependent_section_variable gl x =
   let env = Proofview.Goal.env gl in
-  let sigma = Tacmach.New.project gl in
+  let sigma = Tacmach.project gl in
   let hyps = Proofview.Goal.hyps gl in
   let concl = Proofview.Goal.concl gl in
   List.iter (fun decl ->
@@ -1731,7 +1747,7 @@ let check_non_indirectly_dependent_section_variable gl x =
     let where = match pos with
       | Some id -> str "hypothesis " ++ Id.print id
       | None -> str "the conclusion of the goal" in
-    user_err ~hdr:"Subst"
+    user_err
       (strbrk "Section variable " ++ Id.print x ++
        strbrk " occurs implicitly in global declaration " ++ Printer.pr_global gr ++
        strbrk " present in " ++ where ++ strbrk ".")
@@ -1745,7 +1761,7 @@ let is_non_indirectly_dependent_section_variable gl z =
 
 let subst_one dep_proof_ok x (hyp,rhs,dir) =
   Proofview.Goal.enter begin fun gl ->
-  let sigma = Tacmach.New.project gl in
+  let sigma = Tacmach.project gl in
   let hyps = Proofview.Goal.hyps gl in
   let concl = Proofview.Goal.concl gl in
   (* The set of hypotheses using x *)
@@ -1755,8 +1771,7 @@ let subst_one dep_proof_ok x (hyp,rhs,dir) =
       if not (Id.equal id hyp)
          && List.exists (fun y -> local_occur_var_in_decl sigma y dcl) deps
       then
-        let id_dest = if !regular_subst_tactic then dest else MoveLast in
-        (dest,id::deps,(id_dest,id)::allhyps)
+        (dest,id::deps,(dest,id)::allhyps)
       else
         (MoveBefore id,deps,allhyps))
       hyps
@@ -1789,11 +1804,11 @@ let subst_one_var dep_proof_ok x =
           let hyps = Proofview.Goal.hyps gl in
           let test hyp _ = is_eq_x gl x hyp in
           Context.Named.fold_outside test ~init:() hyps;
-          user_err ~hdr:"Subst"
+          user_err
             (str "Cannot find any non-recursive equality over " ++ Id.print x ++
                str".")
         with FoundHyp res -> res in
-      if is_section_variable x then
+      if is_section_variable (Global.env ()) x then
         check_non_indirectly_dependent_section_variable gl x;
       subst_one dep_proof_ok x res
   end
@@ -1826,64 +1841,36 @@ let subst_all ?(flags=default_subst_tactic_flags) () =
       warn_deprecated_simple_subst ()
   in
 
-  if !regular_subst_tactic then
-
-    (* Find hypotheses to treat in linear time *)
-    let process hyp =
-      Proofview.Goal.enter begin fun gl ->
-      let env = Proofview.Goal.env gl in
-      let sigma = project gl in
-      let c = pf_get_hyp hyp gl |> NamedDecl.get_type in
-      try
-        let lbeq,u,(_,x,y) = pf_apply find_eq_data_decompose gl c in
-        let u = EInstance.kind sigma u in
-        let eq = Constr.mkRef (lbeq.eq,u) in
-        if flags.only_leibniz then restrict_to_eq_and_identity eq;
-        match EConstr.kind sigma x, EConstr.kind sigma y with
-        | Var x, Var y when Id.equal x y ->
-            Proofview.tclUNIT ()
-        | Var x', _ when not (Termops.local_occur_var sigma x' y) &&
-                        not (is_evaluable env (EvalVarRef x')) &&
-                        is_non_indirectly_dependent_section_variable gl x' ->
-            subst_one flags.rewrite_dependent_proof x' (hyp,y,true)
-        | _, Var y' when not (Termops.local_occur_var sigma y' x) &&
-                        not (is_evaluable env (EvalVarRef y')) &&
-                        is_non_indirectly_dependent_section_variable gl y' ->
-            subst_one flags.rewrite_dependent_proof y' (hyp,x,false)
-        | _ ->
-            Proofview.tclUNIT ()
-      with Constr_matching.PatternMatchingFailure ->
-         Proofview.tclUNIT ()
-        end
-    in
+  (* Find hypotheses to treat in linear time *)
+  let process hyp =
     Proofview.Goal.enter begin fun gl ->
-      tclMAP process (List.rev (List.map NamedDecl.get_id (Proofview.Goal.hyps gl)))
-    end
-
-  else
-
-(* Old implementation, not able to manage configurations like a=b, a=t,
-   or situations like "a = S b, b = S a", or also accidentally unfolding
-   let-ins *)
-  Proofview.Goal.enter begin fun gl ->
-  let sigma = project gl in
-  let find_eq_data_decompose = pf_apply find_eq_data_decompose gl in
-  let test (_,c) =
+    let env = Proofview.Goal.env gl in
+    let sigma = project gl in
+    let c = pf_get_hyp hyp gl |> NamedDecl.get_type in
     try
-      let lbeq,u,(_,x,y) = find_eq_data_decompose c in
+      let lbeq,u,(_,x,y) = pf_apply find_eq_data_decompose gl c in
       let u = EInstance.kind sigma u in
       let eq = Constr.mkRef (lbeq.eq,u) in
       if flags.only_leibniz then restrict_to_eq_and_identity eq;
-      (* J.F.: added to prevent failure on goal containing x=x as an hyp *)
-      if EConstr.eq_constr sigma x y then failwith "caught";
-      match EConstr.kind sigma x with Var x -> x | _ ->
-      match EConstr.kind sigma y with Var y -> y | _ -> failwith "caught"
-    with Constr_matching.PatternMatchingFailure -> failwith "caught" in
-  let test p = try Some (test p) with Failure _ -> None in
-  let hyps = pf_hyps_types gl in
-  let ids = List.map_filter test hyps in
-  let ids = List.uniquize ids in
-  subst_gen flags.rewrite_dependent_proof ids
+      match EConstr.kind sigma x, EConstr.kind sigma y with
+      | Var x, Var y when Id.equal x y ->
+          Proofview.tclUNIT ()
+      | Var x', _ when not (Termops.local_occur_var sigma x' y) &&
+                      not (is_evaluable env (EvalVarRef x')) &&
+                      is_non_indirectly_dependent_section_variable gl x' ->
+          subst_one flags.rewrite_dependent_proof x' (hyp,y,true)
+      | _, Var y' when not (Termops.local_occur_var sigma y' x) &&
+                      not (is_evaluable env (EvalVarRef y')) &&
+                      is_non_indirectly_dependent_section_variable gl y' ->
+          subst_one flags.rewrite_dependent_proof y' (hyp,x,false)
+      | _ ->
+          Proofview.tclUNIT ()
+    with Constr_matching.PatternMatchingFailure ->
+        Proofview.tclUNIT ()
+      end
+  in
+  Proofview.Goal.enter begin fun gl ->
+    tclMAP process (List.rev (List.map NamedDecl.get_id (Proofview.Goal.hyps gl)))
   end
 
 (* Rewrite the first assumption for which a condition holds
