@@ -97,39 +97,70 @@ end
 
 (** Declaration of constants and parameters *)
 
-type 'a proof_entry = {
-  proof_entry_body   : 'a Entries.const_entry_body;
+type 'a pproof_entry = {
+  proof_entry_body   : 'a;
   (* List of section variables *)
   proof_entry_secctx : Id.Set.t option;
   (* State id on which the completion of type checking is reported *)
   proof_entry_feedback : Stateid.t option;
   proof_entry_type        : Constr.types option;
-  proof_entry_universes   : Entries.universes_entry;
+  proof_entry_universes   : UState.named_universes_entry;
   proof_entry_opaque      : bool;
   proof_entry_inline_code : bool;
 }
 
-let default_univ_entry = Entries.Monomorphic_entry Univ.ContextSet.empty
+type proof_entry = Evd.side_effects Opaques.const_entry_body pproof_entry
+
+type parameter_entry = {
+  parameter_entry_secctx : Id.Set.t option;
+  parameter_entry_type : Constr.types;
+  parameter_entry_universes : UState.named_universes_entry;
+  parameter_entry_inline_code : Entries.inline;
+}
+
+type primitive_entry = {
+  prim_entry_type : (Constr.types * UState.named_universes_entry) option;
+  prim_entry_content : CPrimitives.op_or_type;
+}
+
+let default_univ_entry = UState.Monomorphic_entry Univ.ContextSet.empty
+let default_named_univ_entry = default_univ_entry, UnivNames.empty_binders
 
 (** [univsbody] are universe-constraints attached to the body-only,
    used in vio-delayed opaque constants and private poly universes *)
-let definition_entry_core ?(opaque=false) ?using ?(inline=false) ?feedback_id ?types
-    ?(univs=default_univ_entry) ?(eff=Evd.empty_side_effects) ?(univsbody=Univ.ContextSet.empty) body =
+let definition_entry_core ?(opaque=false) ?using ?(inline=false) ?types
+    ?(univs=default_named_univ_entry) ?(eff=Evd.empty_side_effects) ?(univsbody=Univ.ContextSet.empty) body =
   { proof_entry_body = Future.from_val ((body,univsbody), eff);
     proof_entry_secctx = using;
     proof_entry_type = types;
     proof_entry_universes = univs;
     proof_entry_opaque = opaque;
-    proof_entry_feedback = feedback_id;
+    proof_entry_feedback = None;
     proof_entry_inline_code = inline}
 
 let definition_entry =
-  definition_entry_core ?eff:None ?univsbody:None ?feedback_id:None
+  definition_entry_core ?eff:None ?univsbody:None
 
-type 'a constant_entry =
-  | DefinitionEntry of 'a proof_entry
-  | ParameterEntry of Entries.parameter_entry
-  | PrimitiveEntry of Entries.primitive_entry
+let parameter_entry ?inline ?(univs=default_named_univ_entry) typ = {
+  parameter_entry_secctx = None;
+  parameter_entry_type = typ;
+  parameter_entry_universes = univs;
+  parameter_entry_inline_code = inline;
+}
+
+let primitive_entry ?types c = {
+  prim_entry_type = types;
+  prim_entry_content = c;
+}
+
+type constant_entry =
+  | DefinitionEntry of proof_entry
+  | ParameterEntry of parameter_entry
+  | PrimitiveEntry of primitive_entry
+
+let local_csts = Summary.ref ~name:"local-csts" Cset_env.empty
+
+let is_local_constant c = Cset_env.mem c !local_csts
 
 type constant_obj = {
   cst_kind : Decls.logical_kind;
@@ -141,17 +172,20 @@ let load_constant i ((sp,kn), obj) =
     raise (DeclareUniv.AlreadyDeclared (None, Libnames.basename sp));
   let con = Global.constant_of_delta_kn kn in
   Nametab.push (Nametab.Until i) sp (GlobRef.ConstRef con);
-  Dumpglob.add_constant_kind con obj.cst_kind
+  Dumpglob.add_constant_kind con obj.cst_kind;
+  begin match obj.cst_locl with
+    | Locality.ImportNeedQualified -> local_csts := Cset_env.add con !local_csts
+    | Locality.ImportDefaultBehavior -> ()
+  end
 
 (* Opening means making the name without its module qualification available *)
-let open_constant f i ((sp,kn), obj) =
+let open_constant i ((sp,kn), obj) =
   (* Never open a local definition *)
   match obj.cst_locl with
   | Locality.ImportNeedQualified -> ()
   | Locality.ImportDefaultBehavior ->
     let con = Global.constant_of_delta_kn kn in
-    if Libobject.in_filter_ref (GlobRef.ConstRef con) f then
-      Nametab.push (Nametab.Exactly i) sp (GlobRef.ConstRef con)
+    Nametab.push (Nametab.Exactly i) sp (GlobRef.ConstRef con)
 
 let exists_name id =
   Decls.variable_exists id || Global.exists_objlabel (Label.of_id id)
@@ -181,7 +215,7 @@ let (objConstant : constant_obj Libobject.Dyn.tag) =
   declare_object_full { (default_object "CONSTANT") with
     cache_function = cache_constant;
     load_function = load_constant;
-    open_function = open_constant;
+    open_function = simple_open open_constant;
     classify_function = classify_constant;
     subst_function = ident_subst_function;
     discharge_function = discharge_constant }
@@ -201,16 +235,21 @@ let register_constant kn kind local =
   let _ = Lib.add_leaf id o in
   update_tables kn
 
-let register_side_effect (c, role) =
+let register_side_effect (c, body, role) =
+  (* Register the body in the opaque table *)
+  let () = match body with
+  | None -> ()
+  | Some opaque -> Opaques.declare_private_opaque opaque
+  in
   let () = register_constant c Decls.(IsProof Theorem) Locality.ImportDefaultBehavior in
   match role with
   | None -> ()
   | Some (Evd.Schema (ind, kind)) -> DeclareScheme.declare_scheme kind [|ind,c|]
 
 let get_roles export eff =
-  let map c =
+  let map (c, body) =
     let role = try Some (Cmap.find c eff.Evd.seff_roles) with Not_found -> None in
-    (c, role)
+    (c, body, role)
   in
   List.map map export
 
@@ -232,8 +271,8 @@ let record_aux env s_ty s_bo =
   Aux_file.record_in_aux "context_used" v
 
 let pure_definition_entry ?(opaque=false) ?(inline=false) ?types
-    ?(univs=default_univ_entry) body =
-  { proof_entry_body = Future.from_val ((body,Univ.ContextSet.empty), ());
+    ?(univs=default_named_univ_entry) body =
+  { proof_entry_body = ((body,Univ.ContextSet.empty), ());
     proof_entry_secctx = None;
     proof_entry_type = types;
     proof_entry_universes = univs;
@@ -251,30 +290,35 @@ let delayed_definition_entry ~opaque ?feedback_id ~using ~univs ?types body =
   ; proof_entry_inline_code = false
   }
 
+let extract_monomorphic = function
+| UState.Monomorphic_entry ctx -> Entries.Monomorphic_entry, ctx
+| UState.Polymorphic_entry uctx -> Entries.Polymorphic_entry uctx, Univ.ContextSet.empty
+
 let cast_proof_entry e =
-  let (body, ctx), () = Future.force e.proof_entry_body in
-  let univs =
-    if Univ.ContextSet.is_empty ctx then e.proof_entry_universes
-    else match e.proof_entry_universes with
-      | Entries.Monomorphic_entry ctx' ->
+  let (body, ctx), () = e.proof_entry_body in
+  let univ_entry =
+    if Univ.ContextSet.is_empty ctx then fst (e.proof_entry_universes)
+    else match fst (e.proof_entry_universes) with
+      | UState.Monomorphic_entry ctx' ->
         (* This can actually happen, try compiling EqdepFacts for instance *)
-        Entries.Monomorphic_entry (Univ.ContextSet.union ctx' ctx)
-      | Entries.Polymorphic_entry _ ->
+        UState.Monomorphic_entry (Univ.ContextSet.union ctx' ctx)
+      | UState.Polymorphic_entry _ ->
         CErrors.anomaly Pp.(str "Local universes in non-opaque polymorphic definition.");
   in
+  let univ_entry, ctx = extract_monomorphic univ_entry in
   { Entries.const_entry_body = body;
     const_entry_secctx = e.proof_entry_secctx;
-    const_entry_feedback = e.proof_entry_feedback;
     const_entry_type = e.proof_entry_type;
-    const_entry_universes = univs;
+    const_entry_universes = univ_entry;
     const_entry_inline_code = e.proof_entry_inline_code;
-  }
+  },
+  ctx
 
 type ('a, 'b) effect_entry =
-| EffectEntry : (private_constants, private_constants Entries.const_entry_body) effect_entry
-| PureEntry : (unit, Constr.constr) effect_entry
+| EffectEntry : (private_constants Opaques.const_entry_body, unit) effect_entry
+| PureEntry : (unit Entries.proof_output, Constr.constr) effect_entry
 
-let cast_opaque_proof_entry (type a b) (entry : (a, b) effect_entry) (e : a proof_entry) : b Entries.opaque_entry =
+let cast_opaque_proof_entry (type a b) (entry : (a, b) effect_entry) (e : a pproof_entry) : b Entries.opaque_entry * _ =
   let typ = match e.proof_entry_type with
   | None -> assert false
   | Some typ -> typ
@@ -290,7 +334,7 @@ let cast_opaque_proof_entry (type a b) (entry : (a, b) effect_entry) (e : a proo
         let ids_typ = global_vars_set env typ in
         let pf, env = match entry with
         | PureEntry ->
-          let (pf, _), () = Future.force e.proof_entry_body in
+          let (pf, _), () = e.proof_entry_body in
           pf, env
         | EffectEntry ->
           let (pf, _), eff = Future.force e.proof_entry_body in
@@ -304,25 +348,25 @@ let cast_opaque_proof_entry (type a b) (entry : (a, b) effect_entry) (e : a proo
     Environ.really_needed env (Id.Set.union hyp_typ hyp_def)
   | Some hyps -> hyps
   in
-  let (body, univs : b * _) = match entry with
+  let (body, (univ_entry, ctx) : b * _) = match entry with
   | PureEntry ->
-    let (body, uctx), () = Future.force e.proof_entry_body in
-    let univs = match e.proof_entry_universes with
-    | Entries.Monomorphic_entry uctx' ->
-      Entries.Monomorphic_entry (Univ.ContextSet.union uctx uctx')
-    | Entries.Polymorphic_entry _ ->
+    let (body, uctx), () = e.proof_entry_body in
+    let univ_entry = match fst (e.proof_entry_universes) with
+    | UState.Monomorphic_entry uctx' ->
+      Entries.Monomorphic_entry, (Univ.ContextSet.union uctx uctx')
+    | UState.Polymorphic_entry uctx' ->
       assert (Univ.ContextSet.is_empty uctx);
-      e.proof_entry_universes
+      Entries.Polymorphic_entry uctx', Univ.ContextSet.empty
     in
-    body, univs
-  | EffectEntry -> e.proof_entry_body, e.proof_entry_universes
+    body, univ_entry
+  | EffectEntry -> (), extract_monomorphic (fst (e.proof_entry_universes))
   in
   { Entries.opaque_entry_body = body;
     opaque_entry_secctx = secctx;
-    opaque_entry_feedback = e.proof_entry_feedback;
     opaque_entry_type = typ;
-    opaque_entry_universes = univs;
-  }
+    opaque_entry_universes = univ_entry;
+  },
+  ctx
 
 let feedback_axiom () = Feedback.(feedback AddedAxiom)
 
@@ -331,9 +375,13 @@ let is_unsafe_typing_flags flags =
   let open Declarations in
   not (flags.check_universes && flags.check_guarded && flags.check_positive)
 
-let define_constant ~name ~typing_flags cd =
+let make_ubinders uctx (univs, ubinders as u) = match univs with
+| UState.Polymorphic_entry _ -> u
+| UState.Monomorphic_entry _ -> (UState.Monomorphic_entry uctx, ubinders)
+
+let declare_constant_core ~name ~typing_flags cd =
   (* Logically define the constant and its subproofs, no libobject tampering *)
-  let decl, unsafe = match cd with
+  let decl, unsafe, ubinders, delayed = match cd with
     | DefinitionEntry de ->
       (* We deal with side effects *)
       if not de.proof_entry_opaque then
@@ -341,42 +389,81 @@ let define_constant ~name ~typing_flags cd =
         (* This globally defines the side-effects in the environment
            and registers their libobjects. *)
         let () = export_side_effects eff in
-        let de = { de with proof_entry_body = Future.from_val (body, ()) } in
-        let cd = Entries.DefinitionEntry (cast_proof_entry de) in
-        ConstantEntry cd, false
+        let de = { de with proof_entry_body = body, () } in
+        let e, ctx = cast_proof_entry de in
+        let ubinders = make_ubinders ctx de.proof_entry_universes in
+        (* We register the global universes after exporting side-effects, since
+           the latter depend on the former. *)
+        let () = DeclareUctx.declare_universe_context ~poly:false ctx in
+        let cd = Entries.DefinitionEntry e in
+        ConstantEntry cd, false, ubinders, None
       else
         let map (body, eff) = body, eff.Evd.seff_private in
         let body = Future.chain de.proof_entry_body map in
+        let feedback_id = de.proof_entry_feedback in
         let de = { de with proof_entry_body = body } in
-        let de = cast_opaque_proof_entry EffectEntry de in
-        OpaqueEntry de, false
+        let cd, ctx = cast_opaque_proof_entry EffectEntry de in
+        let ubinders = make_ubinders ctx de.proof_entry_universes in
+        let () = DeclareUctx.declare_universe_context ~poly:false ctx in
+        OpaqueEntry cd, false, ubinders, Some (body, feedback_id)
     | ParameterEntry e ->
-      ConstantEntry (Entries.ParameterEntry e), not (Lib.is_modtype_strict())
+      let univ_entry, ctx = extract_monomorphic (fst e.parameter_entry_universes) in
+      let ubinders = make_ubinders ctx e.parameter_entry_universes in
+      let () = DeclareUctx.declare_universe_context ~poly:false ctx in
+      let e = {
+        Entries.parameter_entry_secctx = e.parameter_entry_secctx;
+        Entries.parameter_entry_type = e.parameter_entry_type;
+        Entries.parameter_entry_universes = univ_entry;
+        Entries.parameter_entry_inline_code = e.parameter_entry_inline_code;
+      } in
+      ConstantEntry (Entries.ParameterEntry e), not (Lib.is_modtype_strict()), ubinders, None
     | PrimitiveEntry e ->
-      ConstantEntry (Entries.PrimitiveEntry e), false
+      let typ, ubinders, ctx = match e.prim_entry_type with
+      | None -> None, UnivNames.empty_binders, Univ.ContextSet.empty
+      | Some (typ, (univs, ubinders)) ->
+        let univ_entry, ctx = extract_monomorphic univs in
+        Some (typ, univ_entry), ubinders, ctx
+      in
+      let () = DeclareUctx.declare_universe_context ~poly:false ctx in
+      let e = {
+        Entries.prim_entry_type = typ;
+        Entries.prim_entry_content = e.prim_entry_content;
+      } in
+      let ubinders = (UState.Monomorphic_entry ctx, ubinders) in
+      ConstantEntry (Entries.PrimitiveEntry e), false, ubinders, None
   in
   let kn = Global.add_constant ?typing_flags name decl in
+  let () = DeclareUniv.declare_univ_binders (GlobRef.ConstRef kn) ubinders in
   if unsafe || is_unsafe_typing_flags typing_flags then feedback_axiom();
-  kn
+  kn, delayed
 
 let declare_constant ?(local = Locality.ImportDefaultBehavior) ~name ~kind ~typing_flags cd =
   let () = check_exists name in
-  let kn = define_constant ~typing_flags ~name cd in
+  let kn, delayed = declare_constant_core ~typing_flags ~name cd in
   (* Register the libobjects attached to the constants *)
+  let () = match delayed with
+  | None -> ()
+  | Some (body, feedback_id) ->
+    let open Declarations in
+    match (Global.lookup_constant kn).const_body with
+    | OpaqueDef o ->
+      let (_, _, _, i) = Opaqueproof.repr o in
+      Opaques.declare_defined_opaque ?feedback_id i body
+    | Def _ | Undef _ | Primitive _ -> assert false
+  in
   let () = register_constant kn kind local in
   kn
 
 let declare_private_constant ?role ?(local = Locality.ImportDefaultBehavior) ~name ~kind de =
-  let kn, eff =
-    let de =
-      if not de.proof_entry_opaque then
-        DefinitionEff (cast_proof_entry de)
-      else
-        let de = cast_opaque_proof_entry PureEntry de in
-        OpaqueEff de
-    in
-    Global.add_private_constant name de
+  let de, ctx =
+    if not de.proof_entry_opaque then
+      let de, ctx = cast_proof_entry de in
+      DefinitionEff de, ctx
+    else
+      let de, ctx = cast_opaque_proof_entry PureEntry de in
+      OpaqueEff de, ctx
   in
+  let kn, eff = Global.add_private_constant name ctx de in
   let () = register_constant kn kind local in
   let seff_roles = match role with
   | None -> Cmap.empty
@@ -386,15 +473,16 @@ let declare_private_constant ?role ?(local = Locality.ImportDefaultBehavior) ~na
   kn, eff
 
 let inline_private_constants ~uctx env ce =
-  let body, eff = Future.force ce.proof_entry_body in
+  let body, eff = ce.proof_entry_body in
   let cb, ctx = Safe_typing.inline_private_constants env (body, eff.Evd.seff_private) in
   let uctx = UState.merge ~sideff:true Evd.univ_rigid uctx ctx in
   cb, uctx
 
 (** Declaration of section variables and local definitions *)
 type variable_declaration =
-  | SectionLocalDef of Evd.side_effects proof_entry
-  | SectionLocalAssum of { typ:Constr.types; impl:Glob_term.binding_kind; }
+  | SectionLocalDef of proof_entry
+  | SectionLocalAssum of { typ:Constr.types; impl:Glob_term.binding_kind ;
+                           univs:UState.named_universes_entry }
 
 (* This object is only for things which iterate over objects to find
    variables (only Prettyp.print_context AFAICT) *)
@@ -410,31 +498,35 @@ let declare_variable_core ~name ~kind d =
   if Decls.variable_exists name then
     raise (DeclareUniv.AlreadyDeclared (None, name));
 
-  let impl,opaque = match d with (* Fails if not well-typed *)
-    | SectionLocalAssum {typ;impl} ->
+  let impl,opaque,univs = match d with (* Fails if not well-typed *)
+    | SectionLocalAssum {typ;impl;univs} ->
+      let poly, uctx = match fst univs with
+        | UState.Monomorphic_entry uctx -> false, uctx
+        | UState.Polymorphic_entry uctx -> true, Univ.ContextSet.of_context uctx
+      in
+      let () = DeclareUctx.declare_universe_context ~poly uctx in
       let () = Global.push_named_assum (name,typ) in
-      impl, true
+      impl, true, univs
     | SectionLocalDef (de) ->
       (* The body should already have been forced upstream because it is a
          section-local definition, but it's not enforced by typing *)
-      let ((body, body_ui), eff) = Future.force de.proof_entry_body in
+      let ((body, body_uctx), eff) = Future.force de.proof_entry_body in
       let () = export_side_effects eff in
-      let poly, entry_ui = match de.proof_entry_universes with
-        | Entries.Monomorphic_entry uctx -> false, uctx
-        | Entries.Polymorphic_entry (_, uctx) -> true, Univ.ContextSet.of_context uctx
+      let poly, type_uctx = match fst de.proof_entry_universes with
+        | UState.Monomorphic_entry uctx -> false, uctx
+        | UState.Polymorphic_entry uctx -> true, Univ.ContextSet.of_context uctx
       in
-      let univs = Univ.ContextSet.union body_ui entry_ui in
+      let univs = Univ.ContextSet.union body_uctx type_uctx in
       (* We must declare the universe constraints before type-checking the
          term. *)
       let () = DeclareUctx.declare_universe_context ~poly univs in
       let se = {
         Entries.secdef_body = body;
         secdef_secctx = de.proof_entry_secctx;
-        secdef_feedback = de.proof_entry_feedback;
         secdef_type = de.proof_entry_type;
       } in
       let () = Global.push_named_def (name, se) in
-      Glob_term.Explicit, de.proof_entry_opaque
+      Glob_term.Explicit, de.proof_entry_opaque, de.proof_entry_universes
   in
   Nametab.push (Nametab.Until 1) (Libnames.make_path DirPath.empty name) (GlobRef.VarRef name);
   Decls.(add_variable_data name {opaque;kind});
@@ -442,8 +534,8 @@ let declare_variable_core ~name ~kind d =
   Impargs.declare_var_implicits ~impl name;
   Notation.declare_ref_arguments_scope Evd.empty (GlobRef.VarRef name)
 
-let declare_variable ~name ~kind ~typ ~impl =
-  declare_variable_core ~name ~kind (SectionLocalAssum { typ; impl })
+let declare_variable ~name ~kind ~typ ~impl ~univs =
+  declare_variable_core ~name ~kind (SectionLocalAssum { typ; impl; univs })
 
 (* Declaration messages *)
 
@@ -485,6 +577,9 @@ let assumption_message id =
 
 module Internal = struct
 
+  let pmap_entry_body ~f entry =
+    { entry with proof_entry_body = f entry.proof_entry_body }
+
   let map_entry_body ~f entry =
     { entry with proof_entry_body = Future.chain entry.proof_entry_body f }
 
@@ -530,13 +625,11 @@ module Internal = struct
       | None -> assert false
       | Some t -> t
     in
-    (* The body has been forced by the call to [build_constant_by_tactic] *)
-    let () = assert (Future.is_over const.proof_entry_body) in
-    let ((body, uctx), eff) = Future.force const.proof_entry_body in
+    let ((body, uctx), eff) = const.proof_entry_body in
     let (body, typ, ctx) = decompose (List.length sign) body typ [] in
     let (body, typ, args) = shrink ctx sign body typ [] in
     { const with
-      proof_entry_body = Future.from_val ((body, uctx), eff)
+      proof_entry_body = ((body, uctx), eff)
     ; proof_entry_type = Some typ
     }, args
 
@@ -564,7 +657,6 @@ let declare_entry_core ~name ~scope ~kind ~typing_flags ?hook ~obls ~impargs ~uc
     && not (List.is_empty (Global.named_context()))
     && Option.is_empty entry.proof_entry_secctx
   in
-  let ubind = UState.universe_binders uctx in
   let dref = match scope with
   | Locality.Discharge ->
     let () = declare_variable_core ~name ~kind (SectionLocalDef entry) in
@@ -574,7 +666,6 @@ let declare_entry_core ~name ~scope ~kind ~typing_flags ?hook ~obls ~impargs ~uc
     let kn = declare_constant ~name ~local ~kind ~typing_flags (DefinitionEntry entry) in
     let gr = Names.GlobRef.ConstRef kn in
     if should_suggest then Proof_using.suggest_constant (Global.env ()) kn;
-    let () = DeclareUniv.declare_univ_binders gr ubind in
     gr
   in
   let () = Impargs.maybe_declare_manual_implicits false dref impargs in
@@ -622,7 +713,7 @@ let declare_mutually_recursive_core ~info ~cinfo ~opaque ~ntns ~uctx ~rec_declar
   let isfix = Option.has_some possible_indexes in
   let fixnames = List.map (fun { CInfo.name } -> name) cinfo in
   recursive_message isfix indexes fixnames;
-  List.iter (Metasyntax.add_notation_interpretation (Global.env())) ntns;
+  List.iter (Metasyntax.add_notation_interpretation ~local:(scope=Locality.Discharge) (Global.env())) ntns;
   csts
 
 let declare_mutually_recursive = declare_mutually_recursive_core ~restrict_ucontext:true ()
@@ -632,7 +723,7 @@ let warn_let_as_axiom =
     Pp.(fun id -> strbrk "Let definition" ++ spc () ++ Names.Id.print id ++
                   spc () ++ strbrk "declared as an axiom.")
 
-let declare_assumption ~name ~scope ~hook ~impargs ~uctx pe =
+let declare_parameter ~name ~scope ~hook ~impargs ~uctx pe =
   let local = match scope with
     | Locality.Discharge -> warn_let_as_axiom name; Locality.ImportNeedQualified
     | Locality.Global local -> local
@@ -643,14 +734,13 @@ let declare_assumption ~name ~scope ~hook ~impargs ~uctx pe =
   let dref = Names.GlobRef.ConstRef kn in
   let () = Impargs.maybe_declare_manual_implicits false dref impargs in
   let () = assumption_message name in
-  let () = DeclareUniv.declare_univ_binders dref (UState.universe_binders uctx) in
   let () = Hook.(call ?hook { S.uctx; obls = []; scope; dref}) in
   dref
 
 (* Preparing proof entries *)
 let error_unresolved_evars env sigma t evars =
   let pr_unresolved_evar e =
-    hov 2 (str"- " ++ Printer.pr_existential_key sigma e ++  str ": " ++
+    hov 2 (str"- " ++ Printer.pr_existential_key env sigma e ++  str ": " ++
       Himsg.explain_pretype_error env sigma
         (Pretype_errors.UnsolvableImplicit (e,None)))
   in
@@ -710,7 +800,13 @@ let prepare_parameter ~poly ~udecl ~types sigma =
       sigma (fun nf -> nf types)
   in
   let univs = Evd.check_univ_decl ~poly sigma udecl in
-  sigma, (None(*proof using*), (typ, univs), None(*inline*))
+  let pe = {
+      parameter_entry_secctx = None;
+      parameter_entry_type = typ;
+      parameter_entry_universes = univs;
+      parameter_entry_inline_code = None;
+    } in
+  sigma, pe
 
 type progress = Remain of int | Dependent | Defined of GlobRef.t
 
@@ -748,7 +844,7 @@ module ProgramDecl = struct
     ; prg_obligations : obligations
     ; prg_deps : Id.t list
     ; prg_fixkind : fixpoint_kind option
-    ; prg_notations : Vernacexpr.decl_notation list
+    ; prg_notations : Metasyntax.where_decl_notation list
     ; prg_reduce : constr -> constr
     }
 
@@ -886,7 +982,8 @@ let shrink_body c ty =
 let unfold_entry cst = Hints.HintsUnfoldEntry [Tacred.EvalConstRef cst]
 
 let add_hint local prg cst =
-  let locality = if local then Goptions.OptLocal else Goptions.OptExport in
+  (* XXX checking sections here is suspicious but matches historical (unintended?) behaviour *)
+  let locality = if local || Global.sections_are_opened () then Hints.Local else Hints.SuperGlobal in
   Hints.add_hints ~locality [Id.to_string prg.prg_cinfo.CInfo.name] (unfold_entry cst)
 
 let declare_obligation prg obl ~uctx ~types ~body =
@@ -917,10 +1014,10 @@ let declare_obligation prg obl ~uctx ~types ~body =
       add_hint (Locality.make_section_locality None) prg constant;
     definition_message obl.obl_name;
     let body =
-      match univs with
-      | Entries.Polymorphic_entry (_, uctx) ->
+      match fst univs with
+      | UState.Polymorphic_entry uctx ->
         Some (DefinedObl (constant, Univ.UContext.instance uctx))
-      | Entries.Monomorphic_entry _ ->
+      | UState.Monomorphic_entry _ ->
         Some
           (TermObl
              (it_mkLambda_or_LetIn_or_clean
@@ -938,7 +1035,7 @@ let not_transp_msg =
     ++ str "Use 'Defined' instead.")
 
 let err_not_transp () =
-  CErrors.user_err ~hdr:"Program" not_transp_msg
+  CErrors.user_err not_transp_msg
 
 module ProgMap = Id.Map
 
@@ -991,13 +1088,13 @@ let check_solved_obligations ~pm ~what_for : unit =
   if not (ProgMap.is_empty pm) then
     let keys = map_keys pm in
     let have_string = if Int.equal (List.length keys) 1 then " has " else " have " in
-    CErrors.user_err ~hdr:"Program"
+    CErrors.user_err
       Pp.(
         str "Unsolved obligations when closing "
         ++ what_for ++ str ":" ++ spc ()
         ++ prlist_with_sep spc (fun x -> Id.print x) keys
         ++ str have_string
-        ++ str "unsolved obligations" )
+        ++ str "unsolved obligations." )
 
 let map_replace k v m = ProgMap.add k (CEphemeron.create v) (ProgMap.remove k m)
 let progmap_remove pm prg = ProgMap.remove prg.prg_cinfo.CInfo.name pm
@@ -1459,10 +1556,10 @@ let start_mutual_with_initialization ~info ~cinfo ~mutual_info sigma snl =
       | Some init_terms ->
         (* This is the case for hybrid proof mode / definition
            fixpoint, where terms for some constants are given with := *)
-        let tacl = List.map (Option.cata (EConstr.of_constr %> Tactics.exact_no_check) Tacticals.New.tclIDTAC) init_terms in
-        List.map2 (fun tac thm -> Tacticals.New.tclTHEN tac (intro_tac thm)) tacl cinfo
+        let tacl = List.map (Option.cata (EConstr.of_constr %> Tactics.exact_no_check) Tacticals.tclIDTAC) init_terms in
+        List.map2 (fun tac thm -> Tacticals.tclTHEN tac (intro_tac thm)) tacl cinfo
     in
-    Tacticals.New.tclTHENS rec_tac term_tac, guard
+    Tacticals.tclTHENS rec_tac term_tac, guard
   in
   match cinfo with
   | [] -> CErrors.anomaly (Pp.str "No proof to start.")
@@ -1512,7 +1609,7 @@ let get_open_goals ps =
 type proof_object =
   { name : Names.Id.t
   (* [name] only used in the STM *)
-  ; entries : Evd.side_effects proof_entry list
+  ; entries : proof_entry list
   ; uctx: UState.t
   ; pinfo : Proof_info.t
   }
@@ -1564,7 +1661,7 @@ let prepare_proof ~unsafe_typ { proof } =
 
 let make_univs_deferred ~poly ~initial_euctx ~uctx ~udecl
     (used_univs_typ, typ) (used_univs_body, body) =
-  let used_univs = Univ.LSet.union used_univs_body used_univs_typ in
+  let used_univs = Univ.Level.Set.union used_univs_body used_univs_typ in
   let utyp = UState.univ_entry ~poly initial_euctx in
   let uctx = UState.constrain_variables (fst (UState.context_set initial_euctx)) uctx in
   (* For vi2vo compilation proofs are computed now but we need to
@@ -1575,7 +1672,7 @@ let make_univs_deferred ~poly ~initial_euctx ~uctx ~udecl
   utyp, ubody
 
 let make_univs_private_poly ~poly ~uctx ~udecl (used_univs_typ, typ) (used_univs_body, body) =
-  let used_univs = Univ.LSet.union used_univs_body used_univs_typ in
+  let used_univs = Univ.Level.Set.union used_univs_body used_univs_typ in
   let uctx = UState.restrict uctx used_univs in
   let uctx' = UState.restrict uctx used_univs_typ in
   let utyp = UState.check_univ_decl ~poly uctx' udecl in
@@ -1586,7 +1683,7 @@ let make_univs_private_poly ~poly ~uctx ~udecl (used_univs_typ, typ) (used_univs
   utyp, ubody
 
 let make_univs ~poly ~uctx ~udecl (used_univs_typ, typ) (used_univs_body, body) =
-  let used_univs = Univ.LSet.union used_univs_body used_univs_typ in
+  let used_univs = Univ.Level.Set.union used_univs_body used_univs_typ in
   (* Since the proof is computed now, we can simply have 1 set of
      constraints in which we merge the ones for the body and the ones
      for the typ. We recheck the declaration after restricting with
@@ -1635,29 +1732,27 @@ let close_proof_delayed ~feedback_id ps (fpl : closed_proof_output Future.comput
   if poly then
     CErrors.anomaly (Pp.str "Cannot delay universe-polymorphic constants.");
 
-  let fpl, uctx = Future.split2 fpl in
   (* Because of dependent subgoals at the beginning of proofs, we could
      have existential variables in the initial types of goals, we need to
      normalise them for the kernel. *)
-  let subst_evar k = Evd.existential_opt_value0 sigma k in
-  let nf = UnivSubst.nf_evars_and_universes_opt_subst subst_evar (UState.subst initial_euctx) in
+  let nf = Evarutil.nf_evars_universes (Evd.set_universe_context sigma initial_euctx) in
 
   (* We only support opaque proofs, this will be enforced by using
      different entries soon *)
   let opaque = true in
-  let make_entry p (_, types) =
+  let make_entry i (_, types) =
     (* Already checked the univ_decl for the type universes when starting the proof. *)
     let univs = UState.univ_entry ~poly:false initial_euctx in
     let types = nf (EConstr.Unsafe.to_constr types) in
 
-    Future.chain p (fun (pt,eff) ->
+    Future.chain fpl (fun (pf, uctx) ->
+        let (pt, eff) = List.nth pf i in
         (* Deferred proof, we already checked the universe declaration with
              the initial universes, ensure that the final universes respect
              the declaration as well. If the declaration is non-extensible,
              this will prevent the body from adding universes and constraints. *)
-        let uctx = Future.force uctx in
         let uctx = UState.constrain_variables (fst (UState.context_set initial_euctx)) uctx in
-        let used_univs = Univ.LSet.union
+        let used_univs = Univ.Level.Set.union
             (Vars.universes_of_constr types)
             (Vars.universes_of_constr pt)
         in
@@ -1666,7 +1761,7 @@ let close_proof_delayed ~feedback_id ps (fpl : closed_proof_output Future.comput
         (pt,uctx),eff)
     |> delayed_definition_entry ~opaque ~feedback_id ~using ~univs ~types
   in
-  let entries = Future.map2 make_entry fpl (Proofview.initial_goals entry) in
+  let entries = CList.map_i make_entry 0 (Proofview.initial_goals entry) in
   { name; entries; uctx = initial_euctx; pinfo }
 
 let close_future_proof = close_proof_delayed
@@ -1703,6 +1798,7 @@ let build_constant_by_tactic ~name ?(opaque=Vernacexpr.Transparent) ~uctx ~sign
   let { entries; uctx } = close_proof ~opaque ~keep_body_ucst_separate:false pf in
   match entries with
   | [entry] ->
+    let entry = Internal.pmap_entry_body ~f:Future.force entry in
     entry, status, uctx
   | _ ->
     CErrors.anomaly Pp.(str "[build_constant_by_tactic] close_proof returned more than one proof term")
@@ -1715,7 +1811,7 @@ let build_by_tactic ?(side_eff=true) env ~uctx ~poly ~typ tac =
     if side_eff then inline_private_constants ~uctx env ce
     else
       (* GG: side effects won't get reset: no need to treat their universes specially *)
-      let (cb, ctx), _eff = Future.force ce.proof_entry_body in
+      let (cb, ctx), _eff = ce.proof_entry_body in
       cb, UState.merge ~sideff:false Evd.univ_rigid uctx ctx
   in
   cb, ce.proof_entry_type, ce.proof_entry_universes, status, uctx
@@ -1746,9 +1842,9 @@ let declare_abstract ~name ~poly ~kind ~sign ~secsign ~opaque ~solve_tac sigma c
     Exninfo.iraise (e, info)
   in
   let sigma = Evd.set_universe_context sigma uctx in
-  let body, effs = Future.force const.proof_entry_body in
+  let body, effs = const.proof_entry_body in
   (* We drop the side-effects from the entry, they already exist in the ambient environment *)
-  let const = Internal.map_entry_body const ~f:(fun _ -> body, ()) in
+  let const = Internal.pmap_entry_body const ~f:(fun _ -> body, ()) in
   (* EJGA: Hack related to the above call to
      `build_constant_by_tactic` with `~opaque:Transparent`. Even if
      the abstracted term is destined to be opaque, if we trigger the
@@ -1763,13 +1859,13 @@ let declare_abstract ~name ~poly ~kind ~sign ~secsign ~opaque ~solve_tac sigma c
     declare_private_constant ~local:Locality.ImportNeedQualified ~name ~kind const
   in
   let cst, eff = Impargs.with_implicit_protection cst () in
-  let inst = match const.proof_entry_universes with
-  | Entries.Monomorphic_entry _ -> EConstr.EInstance.empty
-  | Entries.Polymorphic_entry (_, ctx) ->
+  let inst = match fst const.proof_entry_universes with
+  | UState.Monomorphic_entry _ -> EConstr.EInstance.empty
+  | UState.Polymorphic_entry ctx ->
     (* We mimic what the kernel does, that is ensuring that no additional
        constraints appear in the body of polymorphic constants. Ideally this
        should be enforced statically. *)
-    let (_, body_uctx), _ = Future.force const.proof_entry_body in
+    let (_, body_uctx), _ = const.proof_entry_body in
     let () = assert (Univ.ContextSet.is_empty body_uctx) in
     EConstr.EInstance.make (Univ.UContext.instance ctx)
   in
@@ -1803,18 +1899,18 @@ let get_current_context pf =
    mutuals previously on this file. *)
 module MutualEntry : sig
 
-  val declare_variable
+  val declare_possibly_mutual_parameters
     : pinfo:Proof_info.t
     -> uctx:UState.t
     -> sec_vars:Id.Set.t option
-    -> univs:Entries.universes_entry
+    -> univs:UState.named_universes_entry
     -> Names.GlobRef.t list
 
   val declare_mutdef
     (* Common to all recthms *)
     : pinfo:Proof_info.t
     -> uctx:UState.t
-    -> entry:Evd.side_effects proof_entry
+    -> entry:proof_entry
     -> Names.GlobRef.t list
 
 end = struct
@@ -1847,7 +1943,9 @@ end = struct
     *)
     let pe, ubind =
       if i > 0 && not (CList.is_empty compute_guard)
-      then Internal.map_entry_type pe ~f:(fun _ -> Some typ), UnivNames.empty_binders
+      then
+        let typ = UState.nf_universes uctx typ in
+        Internal.map_entry_type pe ~f:(fun _ -> Some typ), UnivNames.empty_binders
       else pe, UState.universe_binders uctx
     in
     (* We when compute_guard was [] in the previous step we should not
@@ -1875,13 +1973,17 @@ end = struct
     in
     List.map_i (declare_mutdef ~pinfo ~uctx pe) 0 pinfo.Proof_info.cinfo
 
-  let declare_variable ~pinfo ~uctx ~sec_vars ~univs =
+  let declare_possibly_mutual_parameters ~pinfo ~uctx ~sec_vars ~univs =
     let { Info.scope; hook } = pinfo.Proof_info.info in
     List.map_i (
       fun i { CInfo.name; typ; impargs } ->
-        let typ = Evarutil.nf_evars_universes (Evd.from_ctx uctx) typ in
-        let pe = (sec_vars, (typ, univs), None) in
-        declare_assumption ~name ~scope ~hook ~impargs ~uctx pe
+        let pe = {
+            parameter_entry_secctx = sec_vars;
+            parameter_entry_type = Evarutil.nf_evars_universes (Evd.from_ctx uctx) typ;
+            parameter_entry_universes = univs;
+            parameter_entry_inline_code = None;
+          } in
+        declare_parameter ~name ~scope ~hook ~impargs ~uctx pe
     ) 0 pinfo.Proof_info.cinfo
 
 end
@@ -1911,7 +2013,7 @@ let compute_proof_using_for_admitted proof typ pproofs =
     | _ -> None
 
 let finish_admitted ~pm ~pinfo ~uctx ~sec_vars ~univs =
-  let cst = MutualEntry.declare_variable ~pinfo ~uctx ~sec_vars ~univs in
+  let cst = MutualEntry.declare_possibly_mutual_parameters ~pinfo ~uctx ~sec_vars ~univs in
   (* If the constant was an obligation we need to update the program map *)
   match CEphemeron.default pinfo.Proof_info.proof_ending Proof_ending.Regular with
   | Proof_ending.End_obligation oinfo ->
@@ -1982,9 +2084,11 @@ let finish_proved_equations ~pm ~kind ~hook i proof_obj types sigma0 =
           | Some id -> id
           | None -> let n = !obls in incr obls; Nameops.add_suffix i ("_obligation_" ^ string_of_int n)
         in
+        let entry = Internal.pmap_entry_body ~f:Future.force entry in
         let entry, args = Internal.shrink_entry local_context entry in
+        let entry = Internal.pmap_entry_body ~f:Future.from_val entry in
         let cst = declare_constant ~name:id ~kind ~typing_flags:None (DefinitionEntry entry) in
-        let sigma, app = Evarutil.new_global sigma (GlobRef.ConstRef cst) in
+        let sigma, app = Evd.fresh_global (Global.env ()) sigma (GlobRef.ConstRef cst) in
         let sigma = Evd.define ev (EConstr.applist (app, List.map EConstr.of_constr args)) sigma in
         sigma, cst) sigma0
       types proof_obj.entries
@@ -2006,6 +2110,7 @@ let finalize_proof ~pm proof_obj proof_info =
     pm, MutualEntry.declare_mutdef ~entry ~uctx ~pinfo:proof_info
   | End_obligation oinfo ->
     let entry, uctx = check_single_entry proof_obj "Obligation.save" in
+    let entry = Internal.pmap_entry_body ~f:Future.force entry in
     Obls_.obligation_terminator ~pm ~entry ~uctx ~oinfo
   | End_derive { f ; name } ->
     pm, finish_derived ~f ~name ~entries:proof_obj.entries
@@ -2051,9 +2156,9 @@ let save_lemma_admitted_delayed ~pm ~proof =
   if List.length entries <> 1 then
     CErrors.user_err Pp.(str "Admitted does not support multiple statements");
   let { proof_entry_secctx; proof_entry_type; proof_entry_universes } = List.hd entries in
-  let poly = match proof_entry_universes with
-    | Entries.Monomorphic_entry _ -> false
-    | Entries.Polymorphic_entry (_, _) -> true in
+  let poly = match fst (proof_entry_universes) with
+    | UState.Monomorphic_entry _ -> false
+    | UState.Polymorphic_entry _ -> true in
   let univs = UState.univ_entry ~poly uctx in
   let sec_vars = if get_keep_admitted_vars () then proof_entry_secctx else None in
   finish_admitted ~pm ~uctx ~pinfo ~sec_vars ~univs
@@ -2172,7 +2277,7 @@ let solve_by_tac prg obls i tac =
   | Tacticals.FailError (_, s) as exn ->
     let _ = Exninfo.capture exn in
     let loc = fst obl.obl_location in
-    CErrors.user_err ?loc ~hdr:"solve_obligation" (Lazy.force s)
+    CErrors.user_err ?loc (Lazy.force s)
   (* If the proof is open we absorb the error and leave the obligation open *)
   | Proof_.OpenProof _ ->
     None
