@@ -119,6 +119,8 @@ let libraries_filename_table = ref DPmap.empty
 (* These are the _ordered_ sets of loaded, imported and exported libraries *)
 let libraries_loaded_list = Summary.ref [] ~name:"LIBRARY-LOAD"
 
+(* Opaque proof tables *)
+
 (* various requests to the tables *)
 
 let find_library dir =
@@ -127,8 +129,8 @@ let find_library dir =
 let try_find_library dir =
   try find_library dir
   with Not_found ->
-    user_err ~hdr:"Library.find_library"
-      (str "Unknown library " ++ DirPath.print dir)
+    user_err
+      (str "Unknown library " ++ DirPath.print dir ++ str ".")
 
 let register_library_filename dir f =
   (* Not synchronized: overwrite the previous binding if one existed *)
@@ -169,20 +171,11 @@ let register_loaded_library m =
 
   let loaded_libraries () = !libraries_loaded_list
 
-(************************************************************************)
-(** {6 Tables of opaque proof terms} *)
-
-(** We now store opaque proof terms apart from the rest of the environment.
-    See the [Indirect] constructor in [Lazyconstr.lazy_constr]. This way,
-    we can quickly load a first half of a .vo file without these opaque
-    terms, and access them only when a specific command (e.g. Print or
-    Print Assumptions) needs it. *)
-
 (** Delayed / available tables of opaque terms *)
 
 type table_status =
-  | ToFetch of Opaqueproof.opaque_disk delayed
-  | Fetched of Opaqueproof.opaque_disk
+  | ToFetch of Opaques.opaque_disk delayed
+  | Fetched of Opaques.opaque_disk
 
 let opaque_tables =
   ref (DPmap.empty : table_status DPmap.t)
@@ -199,7 +192,7 @@ let access_table what tables dp i =
       let t =
         try fetch_delayed f
         with Faulty f ->
-          user_err ~hdr:"Library.access_table"
+          user_err
             (str "The file " ++ str f ++ str " (bound to " ++ str dir_path ++
              str ") is corrupted,\ncannot load some " ++
              str what ++ str " in it.\n")
@@ -207,15 +200,26 @@ let access_table what tables dp i =
       tables := DPmap.add dp (Fetched t) !tables;
       t
   in
-  Opaqueproof.get_opaque_disk i t
+  Opaques.get_opaque_disk i t
 
-let access_opaque_table dp i =
-  let what = "opaque proofs" in
-  access_table what opaque_tables dp i
+let access_opaque_table o =
+  let (sub, ci, dp, i) = Opaqueproof.repr o in
+  let ans =
+    if DirPath.equal dp (Global.current_dirpath ()) then
+      Opaques.get_current_opaque i
+    else
+      let what = "opaque proofs" in
+      access_table what opaque_tables dp i
+  in
+  match ans with
+  | None -> None
+  | Some (c, ctx) ->
+    let (c, ctx) = Cooking.cook_constr ci (c, ctx) in
+    let c = Mod_subst.(List.fold_right subst_mps sub c) in
+    Some (c, ctx)
 
 let indirect_accessor = {
-  Opaqueproof.access_proof = access_opaque_table;
-  Opaqueproof.access_discharge = Cooking.cook_constr;
+  Global.access_proof = access_opaque_table;
 }
 
 (************************************************************************)
@@ -225,7 +229,7 @@ type seg_sum = summary_disk
 type seg_lib = library_disk
 type seg_univ = (* true = vivo, false = vi *)
   Univ.ContextSet.t * bool
-type seg_proofs = Opaqueproof.opaque_disk
+type seg_proofs = Opaques.opaque_disk
 
 let mk_library sd md digests univs =
   {
@@ -271,10 +275,10 @@ let rec intern_library ~lib_resolver (needed, contents) (dir, f) from =
   let f = match f with Some f -> f | None -> lib_resolver dir in
   let m = intern_from_file f in
   if not (DirPath.equal dir m.library_name) then
-    user_err ~hdr:"load_physical_library"
+    user_err
       (str "The file " ++ str f ++ str " contains library" ++ spc () ++
        DirPath.print m.library_name ++ spc () ++ str "and not library" ++
-       spc() ++ DirPath.print dir);
+       spc() ++ DirPath.print dir ++ str ".");
   Feedback.feedback (Feedback.FileLoaded(DirPath.to_string dir, f));
   m.library_digests, intern_library_deps ~lib_resolver (needed, contents) dir m f
 
@@ -337,7 +341,7 @@ let load_require _ (_,(needed,modl,_)) =
 
 let open_require i (_,(_,modl,export)) =
   Option.iter (fun export ->
-      let mpl = List.map (fun m -> Unfiltered, MPfile m) modl in
+      let mpl = List.map (fun m -> unfiltered, MPfile m) modl in
       (* TODO support filters in Require *)
       Declaremods.import_modules ~export mpl)
     export
@@ -379,7 +383,7 @@ let require_library_from_dirpath ~lib_resolver modrefl export =
       add_anonymous_leaf (in_require (needed,modrefl,None));
       Option.iter (fun export ->
           (* TODO import filters *)
-          List.iter (fun m -> Declaremods.import_module Unfiltered ~export (MPfile m)) modrefl)
+          List.iter (fun m -> Declaremods.import_module unfiltered ~export (MPfile m)) modrefl)
         export
     end
   else
@@ -389,18 +393,20 @@ let require_library_from_dirpath ~lib_resolver modrefl export =
 (************************************************************************)
 (*s Initializing the compilation of a library. *)
 
+type ('uid, 'doc) tasks = (('uid, 'doc) Stateid.request * bool) list
+
 let load_library_todo f =
   let ch = raw_intern_library f in
   let (s0 : seg_sum), _ = ObjFile.marshal_in_segment ch ~segment:"summary" in
   let (s1 : seg_lib), _ = ObjFile.marshal_in_segment ch ~segment:"library" in
   let (s2 : seg_univ option), _ = ObjFile.marshal_in_segment ch ~segment:"universes" in
-  let tasks, _ = ObjFile.marshal_in_segment ch ~segment:"tasks" in
+  let (tasks : (Opaqueproof.opaque_handle option, 'doc) tasks option), _ = ObjFile.marshal_in_segment ch ~segment:"tasks" in
   let (s4 : seg_proofs), _ = ObjFile.marshal_in_segment ch ~segment:"opaques" in
   ObjFile.close_in ch;
   System.check_caml_version ~caml:s0.md_ocaml ~file:f;
-  if tasks = None then user_err ~hdr:"restart" (str"not a .vio file");
-  if s2 = None then user_err ~hdr:"restart" (str"not a .vio file");
-  if snd (Option.get s2) then user_err ~hdr:"restart" (str"not a .vio file");
+  if tasks = None then user_err (str "Not a .vio file.");
+  if s2 = None then user_err (str "Not a .vio file.");
+  if snd (Option.get s2) then user_err (str "Not a .vio file.");
   s0, s1, Option.get s2, Option.get tasks, s4
 
 (************************************************************************)
@@ -419,6 +425,11 @@ let error_recursively_dependent_library dir =
      strbrk " to save current library because" ++
      strbrk " it already depends on a library of this name.")
 
+type 'doc todo_proofs =
+ | ProofsTodoNone (* for .vo *)
+ | ProofsTodoSomeEmpty of Future.UUIDSet.t (* for .vos *)
+ | ProofsTodoSome of Future.UUIDSet.t * (Future.UUID.t, 'doc) tasks (* for .vio *)
+
 (* We now use two different digests in a .vo file. The first one
    only covers half of the file, without the opaque table. It is
    used for identifying this version of this library : this digest
@@ -436,7 +447,7 @@ let save_library_base f sum lib univs tasks proofs =
     ObjFile.marshal_out_segment ch ~segment:"summary" (sum    : seg_sum);
     ObjFile.marshal_out_segment ch ~segment:"library" (lib    : seg_lib);
     ObjFile.marshal_out_segment ch ~segment:"universes" (univs  : seg_univ option);
-    ObjFile.marshal_out_segment ch ~segment:"tasks" (tasks  : 'tasks option);
+    ObjFile.marshal_out_segment ch ~segment:"tasks" (tasks : (Opaqueproof.opaque_handle option, 'doc) tasks option);
     ObjFile.marshal_out_segment ch ~segment:"opaques" (proofs : seg_proofs);
     ObjFile.close_out ch
   with reraise ->
@@ -446,12 +457,7 @@ let save_library_base f sum lib univs tasks proofs =
     Sys.remove f;
     Exninfo.iraise reraise
 
-type 'document todo_proofs =
- | ProofsTodoNone (* for .vo *)
- | ProofsTodoSomeEmpty of Future.UUIDSet.t (* for .vos *)
- | ProofsTodoSome of Future.UUIDSet.t * ((Future.UUID.t,'document) Stateid.request * bool) list (* for .vio *)
-
-let save_library_to todo_proofs ~output_native_objects dir f otab =
+let save_library_to todo_proofs ~output_native_objects dir f =
   assert(
     let expected_extension = match todo_proofs with
       | ProofsTodoNone -> ".vo"
@@ -464,8 +470,13 @@ let save_library_to todo_proofs ~output_native_objects dir f otab =
     | ProofsTodoSomeEmpty except -> except
     | ProofsTodoSome (except,l) -> except
     in
-  let cenv, seg, ast = Declaremods.end_library ~output_native_objects ~except dir in
-  let opaque_table, f2t_map = Opaqueproof.dump ~except otab in
+  (* Ensure that the call below is performed with all opaques joined *)
+  let () = Opaques.Summary.join ~except () in
+  let opaque_table, f2t_map = Opaques.dump ~except () in
+  let () = assert (not (Future.UUIDSet.is_empty except) ||
+    Safe_typing.is_joined_environment (Global.safe_env ()))
+  in
+  let cenv, seg, ast = Declaremods.end_library ~output_native_objects dir in
   let tasks, utab =
     match todo_proofs with
     | ProofsTodoNone -> None, None
