@@ -125,6 +125,14 @@ let pr_ltac_ref = Libnames.pr_qualid
 
 let pr_module = Libnames.pr_qualid
 
+let pr_import_cats = function
+  | None -> mt()
+  | Some {negative;import_cats=cats} ->
+    (if negative then str "-" else mt()) ++
+    str "(" ++
+    prlist_with_sep (fun () -> str ",") (fun x -> str x.v) cats ++
+    str ")"
+
 let pr_one_import_filter_name (q,etc) =
   Libnames.pr_qualid q ++ if etc then str "(..)" else mt()
 
@@ -464,7 +472,7 @@ let pr_ne_params_list pr_c l =
 
 let pr_thm_token k = keyword (string_of_theorem_kind k)
 
-let pr_syntax_modifier = let open Gramlib.Gramext in function
+let pr_syntax_modifier = let open Gramlib.Gramext in CAst.with_val (function
     | SetItemLevel (l,bko,n) ->
       prlist_with_sep sep_v2 str l ++ spc () ++ pr_at_level n ++
       pr_opt pr_constr_as_binder_kind bko
@@ -478,8 +486,8 @@ let pr_syntax_modifier = let open Gramlib.Gramext in function
     | SetEntryType (x,typ) -> str x ++ spc() ++ pr_set_simple_entry_type typ
     | SetOnlyPrinting -> keyword "only printing"
     | SetOnlyParsing -> keyword "only parsing"
-    | SetFormat("text",s) -> keyword "format " ++ pr_ast qs s
-    | SetFormat(k,s) -> keyword "format " ++ qs k ++ spc() ++ pr_ast qs s
+    | SetFormat (TextFormat s) -> keyword "format " ++ pr_ast qs s
+    | SetFormat (ExtraFormat (k,s)) -> keyword "format " ++ qs k ++ spc() ++ pr_ast qs s)
 
 let pr_syntax_modifiers = function
   | [] -> mt()
@@ -559,8 +567,8 @@ let pr_printable = function
     keyword "Print Custom Grammar" ++ spc() ++ str ent
   | PrintLoadPath dir ->
     keyword "Print LoadPath" ++ pr_opt DirPath.print dir
-  | PrintModules ->
-    keyword "Print Modules"
+  | PrintLibraries ->
+    keyword "Print Libraries"
   | PrintMLLoadPath ->
     keyword "Print ML Path"
   | PrintMLModules ->
@@ -727,12 +735,6 @@ let pr_vernac_expr v =
       if Int.equal i 1 then keyword "Back" else keyword "Back" ++ pr_intarg i
     )
 
-  (* State management *)
-  | VernacWriteState s ->
-    return (keyword "Write State" ++ spc () ++ qs s)
-  | VernacRestoreState s ->
-    return  (keyword "Restore State" ++ spc() ++ qs s)
-
   (* Syntax *)
   | VernacOpenCloseScope (opening,sc) ->
     return (
@@ -757,24 +759,16 @@ let pr_vernac_expr v =
       keyword "Bind Scope" ++ spc () ++ str sc ++
       spc() ++ keyword "with" ++ spc () ++ prlist_with_sep spc pr_class_rawexpr cll
     )
-  | VernacInfix (({v=s},mv),q,sn) -> (* A Verifier *)
+  | VernacNotation (infix,c,({v=s},l),opt) ->
     return (
-      hov 0 (hov 0 (keyword "Infix "
-                    ++ qs s ++ str " :=" ++ pr_constrarg q) ++
-             pr_syntax_modifiers mv ++
-             (match sn with
+      hov 2 (hov 0 (keyword (if infix then "Infix" else "Notation") ++ spc() ++ qs s ++
+             str " :=" ++ Flags.without_option Flags.beautify pr_constrarg c) ++
+             pr_syntax_modifiers l ++
+             (match opt with
               | None -> mt()
               | Some sc -> spc() ++ str":" ++ spc() ++ str sc))
     )
-  | VernacNotation (c,({v=s},l),opt) ->
-    return (
-      hov 2 (keyword "Notation" ++ spc() ++ qs s ++
-             str " :=" ++ Flags.without_option Flags.beautify pr_constrarg c ++ pr_syntax_modifiers l ++
-             (match opt with
-              | None -> mt()
-              | Some sc -> str" :" ++ spc() ++ str sc))
-    )
-  | VernacSyntaxExtension (_, (s, l)) ->
+  | VernacReservedNotation (_, (s, l)) ->
     return (
       keyword "Reserved Notation" ++ spc() ++ pr_ast qs s ++
       pr_syntax_modifiers l
@@ -952,9 +946,9 @@ let pr_vernac_expr v =
         (from ++ keyword "Require" ++ spc() ++ pr_require_token exp ++
          prlist_with_sep sep pr_module l)
     )
-  | VernacImport (f,l) ->
+  | VernacImport (f,cats,l) ->
     return (
-      (if f then keyword "Export" else keyword "Import") ++ spc() ++
+      (if f then keyword "Export" else keyword "Import") ++ pr_import_cats cats ++ spc() ++
       prlist_with_sep sep pr_import_module l
     )
   | VernacCanonical q ->
@@ -1015,7 +1009,7 @@ let pr_vernac_expr v =
     return (
       hov 1 (keyword "Existing" ++ spc () ++
              keyword(String.plural (List.length insts) "Instance") ++
-             spc () ++ prlist_with_sep (fun () -> str", ") pr_inst insts)
+             spc () ++ prlist_with_sep spc pr_inst insts)
     )
 
   | VernacExistingClass id ->
@@ -1056,9 +1050,6 @@ let pr_vernac_expr v =
       hov 2 (keyword "Include" ++ spc() ++
              prlist_with_sep (fun () -> str " <+ ") pr_m mexprs)
     )
-  (* Solving *)
-  | VernacSolveExistential (i,c) ->
-    return (keyword "Existential" ++ spc () ++ int i ++ pr_lconstrarg c)
 
   (* Auxiliary file and library management *)
   | VernacAddLoadPath { implicit; physical_path; logical_path } ->
@@ -1141,8 +1132,8 @@ let pr_vernac_expr v =
                       notation_scope = s;
                       implicit_status = imp } :: tl ->
             let extra, tl = get_arguments_like s imp tl in
-            spc() ++ pr_br imp (extra<>[]) (prlist_with_sep spc pr_one_arg ((id,k)::extra)) ++
-            pr_s s ++ print_arguments tl
+            spc() ++ hov 1 (pr_br imp (extra<>[]) (prlist_with_sep spc pr_one_arg ((id,k)::extra)) ++
+            pr_s s) ++ print_arguments tl
         in
         let rec print_implicits = function
           | [] -> mt ()
@@ -1327,7 +1318,9 @@ let pr_control_flag (p : control_flag) =
     | ControlTime _ -> keyword "Time"
     | ControlRedirect s -> keyword "Redirect" ++ spc() ++ qs s
     | ControlTimeout n -> keyword "Timeout " ++ int n
-    | ControlFail -> keyword "Fail" in
+    | ControlFail -> keyword "Fail"
+    | ControlSucceed -> keyword "Succeed"
+  in
   w ++ spc ()
 
 let pr_vernac_control flags = Pp.prlist pr_control_flag flags
